@@ -5,9 +5,9 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 2003-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2003-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2003-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 2003-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2003-2004, Lars Ellenberg <l.g.e@web.de>.
+        authors.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -64,27 +64,35 @@ STATIC int _drbd_md_sync_page_io(drbd_dev *mdev, struct page *page,
 STATIC int _drbd_md_sync_page_io(drbd_dev *mdev, struct page *page, 
 				 sector_t sector, int rw, int size)
 {
-	struct bio *bio = bio_alloc(GFP_NOIO, 1);
+	struct bio bio;
+	struct bio_vec vec;
 	struct completion event;
 	int ok;
 
-	bio->bi_bdev = mdev->md_bdev;
-	bio->bi_sector = sector;
-	bio_add_page(bio, page, size, 0);
+	bio_init(&bio);
+	bio.bi_io_vec = &vec;
+	vec.bv_page = page;
+	vec.bv_offset = 0;
+	vec.bv_len =
+	bio.bi_size = size;
+	bio.bi_vcnt = 1;
+	bio.bi_idx = 0;
+	bio.bi_bdev = mdev->md_bdev;
+	bio.bi_sector = sector;
 	init_completion(&event);
-	bio->bi_private = &event;
-	bio->bi_end_io = drbd_md_io_complete;
+	bio.bi_private = &event;
+	bio.bi_end_io = drbd_md_io_complete;
 
 #ifdef BIO_RW_SYNC
-	submit_bio(rw | (1 << BIO_RW_SYNC), bio);
+	submit_bio(rw | (1 << BIO_RW_SYNC), &bio);
 #else
-	submit_bio(rw, bio);
+	submit_bio(rw, &bio);
 	drbd_blk_run_queue(bdev_get_queue(mdev->md_bdev));
 #endif
 	wait_for_completion(&event);
 
-	ok = test_bit(BIO_UPTODATE, &bio->bi_flags);
-	bio_put(bio);
+	ok = test_bit(BIO_UPTODATE, &bio.bi_flags);
+
 	return ok;
 }
 #endif
@@ -98,7 +106,6 @@ int drbd_md_sync_page_io(drbd_dev *mdev, sector_t sector, int rw)
 	D_ASSERT(semaphore_is_locked(&mdev->md_io_mutex));
 
 	if (!mdev->md_bdev) {
-		if (test_bit(DISKLESS,&mdev->flags)) return 0;
 		if (DRBD_ratelimit(5*HZ,5)) {
 			ERR("mdev->md_bdev==NULL\n");
 			dump_stack();
@@ -112,7 +119,7 @@ int drbd_md_sync_page_io(drbd_dev *mdev, sector_t sector, int rw)
 	// in case hardsect != 512 [ s390 only? ]
 	if( hardsect != MD_HARDSECT ) {
 		if(!mdev->md_io_tmpp) {
-			struct page *page = alloc_page(GFP_NOIO);
+			struct page *page = alloc_page(GFP_KERNEL);
 			if(!page) return 0;
 
 			WARN("Meta data's bdev hardsect_size != %d\n",
@@ -143,22 +150,22 @@ int drbd_md_sync_page_io(drbd_dev *mdev, sector_t sector, int rw)
 	}
 
 #if DUMP_MD >= 3
-	INFO("%s [%d]:%s(,%llu,%s)\n",
+	INFO("%s [%d]:%s(,%ld,%s)\n",
 	     current->comm, current->pid, __func__,
-	     (unsigned long long)sector, rw ? "WRITE" : "READ");
+	     sector, rw ? "WRITE" : "READ");
 #endif
 
 	if (sector < drbd_md_ss(mdev)  ||
 	    sector > drbd_md_ss(mdev)+MD_BM_OFFSET+BM_SECT_TO_EXT(capacity)) {
-		ALERT("%s [%d]:%s(,%llu,%s) out of range md access!\n",
+		ALERT("%s [%d]:%s(,%ld,%s) out of range md access!\n",
 		     current->comm, current->pid, __func__,
-		     (unsigned long long)sector, rw ? "WRITE" : "READ");
+		     (long)sector, rw ? "WRITE" : "READ");
 	}
 
 	ok = _drbd_md_sync_page_io(mdev,iop,sector,rw,hardsect);
 	if (unlikely(!ok)) {
-		ERR("drbd_md_sync_page_io(,%llu,%s) failed!\n",
-		    (unsigned long long)sector,rw ? "WRITE" : "READ");
+		ERR("drbd_md_sync_page_io(,%lu,%s) failed!\n",
+		    (unsigned long)sector,rw ? "WRITE" : "READ");
 	}
 
 	if( hardsect != MD_HARDSECT && rw == READ ) {
@@ -191,17 +198,8 @@ struct update_odbm_work {
 	unsigned int enr;
 };
 
-struct update_al_work {
-	struct drbd_work w;
-	struct lc_element * al_ext;
-	struct completion event;
-	unsigned int enr;
-	/* if old_enr != LC_FREE, write corresponding bitmap sector, too */
-	unsigned int old_enr;
-};
-
-STATIC int w_al_write_transaction(struct Drbd_Conf *, struct drbd_work *, int);
-
+STATIC void drbd_al_write_transaction(struct Drbd_Conf *,struct lc_element *,
+				      unsigned int );
 static inline
 struct lc_element* _al_get(struct Drbd_Conf *mdev, unsigned int enr)
 {
@@ -238,25 +236,20 @@ void drbd_al_begin_io(struct Drbd_Conf *mdev, sector_t sector)
 {
 	unsigned int enr = (sector >> (AL_EXTENT_SIZE_B-9));
 	struct lc_element *al_ext;
-	struct update_al_work al_work;
 
 	D_ASSERT(atomic_read(&mdev->local_cnt)>0);
 	wait_event(mdev->al_wait, (al_ext = _al_get(mdev,enr)) );
 
 	if (al_ext->lc_number != enr) {
-		/* drbd_al_write_transaction(mdev,al_ext,enr);
-		   generic_make_request() are serialized on the 
-		   current->bio_tail list now. Therefore we have
-		   to deligate writing something to AL to the
-		   worker thread. */
-		init_completion(&al_work.event);
-		al_work.al_ext = al_ext;
-		al_work.enr = enr;
-		al_work.old_enr = al_ext->lc_number;
-		al_work.w.cb = w_al_write_transaction;
-		drbd_queue_work_front(mdev,&mdev->data.work,&al_work.w);
-		wait_for_completion(&al_work.event);
+		// We have to do write an transaction to AL.
+		unsigned int evicted;
 
+		evicted = al_ext->lc_number;
+
+		if(mdev->cstate < Connected && evicted != LC_FREE ) {
+			drbd_bm_write_sect(mdev, evicted/AL_EXT_PER_BM_SECT );
+		}
+		drbd_al_write_transaction(mdev,al_ext,enr);
 		mdev->al_writ_cnt++;
 
 		/*
@@ -293,27 +286,16 @@ void drbd_al_complete_io(struct Drbd_Conf *mdev, sector_t sector)
 	spin_unlock_irqrestore(&mdev->al_lock,flags);
 }
 
-STATIC int
-w_al_write_transaction(struct Drbd_Conf *mdev, struct drbd_work *w, int unused)
+STATIC void
+drbd_al_write_transaction(struct Drbd_Conf *mdev,struct lc_element *updated,
+			  unsigned int new_enr)
 {
-	struct update_al_work *aw = (struct update_al_work*)w;
-	struct lc_element *updated = aw->al_ext;
-	const unsigned int new_enr = aw->enr;
-	const unsigned int evicted = aw->old_enr;
-
-	struct al_transaction* buffer;
-	sector_t sector;
 	int i,n,mx;
 	unsigned int extent_nr;
+	struct al_transaction* buffer;
+	sector_t sector;
 	u32 xor_sum=0;
 
-	/* do we have to do a bitmap write, first?
-	 * TODO reduce maximum latency:
-	 * submit both bios, then wait for both,
-	 * instead of doing two synchronous sector writes. */
-	if (mdev->cstate < Connected && evicted != LC_FREE)
-		drbd_bm_write_sect(mdev, evicted/AL_EXT_PER_BM_SECT);
-
 	down(&mdev->md_io_mutex); // protects md_io_buffer, al_tr_cycle, ...
 	buffer = (struct al_transaction*)page_address(mdev->md_io_page);
 
@@ -365,18 +347,8 @@ w_al_write_transaction(struct Drbd_Conf *mdev, struct drbd_work *w, int unused)
 	mdev->al_tr_number++;
 
 	up(&mdev->md_io_mutex);
-
-	complete(&((struct update_al_work*)w)->event);
-
-	return 1;
 }
 
-/**
- * drbd_al_read_tr: Reads a single transaction record form the 
- * on disk activity log.
- * Returns -1 on IO error, 0 on checksum error and 1 if it is a valid
- * record.
- */
 STATIC int drbd_al_read_tr(struct Drbd_Conf *mdev,
 			   struct al_transaction* b,
 			   int index)
@@ -390,7 +362,7 @@ STATIC int drbd_al_read_tr(struct Drbd_Conf *mdev,
 	if(!drbd_md_sync_page_io(mdev,sector,READ)) {
 		drbd_chk_io_error(mdev, 1);
 		drbd_io_error(mdev);
-		return -1;
+		return 0;
 	}
 
 	rv = ( be32_to_cpu(b->magic) == DRBD_MAGIC );
@@ -403,12 +375,7 @@ STATIC int drbd_al_read_tr(struct Drbd_Conf *mdev,
 	return rv;
 }
 
-/**
- * drbd_al_read_log: Restores the activity log from its on disk
- * representation. Returns 1 on success, returns 0 when 
- * reading the log failed due to IO errors.
- */
-int drbd_al_read_log(struct Drbd_Conf *mdev)
+void drbd_al_read_log(struct Drbd_Conf *mdev)
 {
 	struct al_transaction* buffer;
 	int from=-1,to=-1,i,cnr, overflow=0,rv;
@@ -427,12 +394,7 @@ int drbd_al_read_log(struct Drbd_Conf *mdev)
 
 	// Find the valid transaction in the log
 	for(i=0;i<=mx;i++) {
-		rv = drbd_al_read_tr(mdev,buffer,i);
-		if(rv == 0) continue;
-		if(rv == -1) {
-			up(&mdev->md_io_mutex);
-			return 0;
-		}
+		if(!drbd_al_read_tr(mdev,buffer,i)) continue;
 		cnr = be32_to_cpu(buffer->tr_number);
 		// INFO("index %d valid tnr=%d\n",i,cnr);
 
@@ -452,7 +414,7 @@ int drbd_al_read_log(struct Drbd_Conf *mdev)
 		WARN("No usable activity log found.\n");
 
 		up(&mdev->md_io_mutex);
-		return 1;
+		return;
 	}
 
 	// Read the valid transactions.
@@ -467,21 +429,12 @@ int drbd_al_read_log(struct Drbd_Conf *mdev)
 		unsigned int trn;
 
 		rv = drbd_al_read_tr(mdev,buffer,i);
-		ERR_IF(rv == 0) goto cancel;
-		if(rv == -1) {
-			up(&mdev->md_io_mutex);
-			return 0;
-		}
+		ERR_IF(!rv) goto cancel;
 
 		trn=be32_to_cpu(buffer->tr_number);
 
 		spin_lock_irq(&mdev->al_lock);
-
-		/* This loop runs backwards because in the cyclic 
-		   elements there might be an old version of the
-		   updated element (in slot 0). So the element in slot 0
-		   can overwrite old versions. */
-		for(j=AL_EXTENTS_PT;j>=0;j--) {
+		for(j=0;j<AL_EXTENTS_PT+1;j++) {
 			pos = be32_to_cpu(buffer->updates[j].pos);
 			extent_nr = be32_to_cpu(buffer->updates[j].extent);
 
@@ -512,8 +465,6 @@ int drbd_al_read_log(struct Drbd_Conf *mdev)
 
 	INFO("Found %d transactions (%d active extents) in activity log.\n",
 	     transactions,active_extents);
-
-	return 1;
 }
 
 /**
@@ -570,7 +521,7 @@ void drbd_al_apply_to_bm(struct Drbd_Conf *mdev)
 	wake_up(&mdev->al_wait);
 
 	INFO("Marked additional %s as out-of-sync based on AL.\n",
-	     ppsize(ppb,Bit2KB(add)));
+	     ppsize(ppb,add >> 1));
 }
 
 static inline int _try_lc_del(struct Drbd_Conf *mdev,struct lc_element *al_ext)
@@ -728,7 +679,7 @@ void __drbd_set_in_sync(drbd_dev* mdev, sector_t sector, int size, const char* f
 	unsigned long sbnr,ebnr,lbnr,bnr;
 	unsigned long count = 0;
 	sector_t esector, nr_sectors;
-	int strange_state,wake_up=0;
+	int strange_state;
 
 	strange_state = (mdev->cstate <= Connected) ||
 	                test_bit(DISKLESS,&mdev->flags) ||
@@ -774,26 +725,23 @@ void __drbd_set_in_sync(drbd_dev* mdev, sector_t sector, int size, const char* f
 	 * ok, (capacity & 7) != 0 sometimes, but who cares...
 	 * we count rs_{total,left} in bits, not sectors.
 	 */
-	spin_lock_irq(&mdev->al_lock);
 	for(bnr=sbnr; bnr <= ebnr; bnr++) {
 		if (drbd_bm_clear_bit(mdev,bnr)) count++;
 	}
 	if (count) {
 		// we need the lock for drbd_try_clear_on_disk_bm
+		spin_lock_irq(&mdev->al_lock);
 		if(jiffies - mdev->rs_mark_time > HZ*10) {
 			/* should be roling marks, but we estimate only anyways. */
-			if( mdev->rs_mark_left != drbd_bm_total_weight(mdev)) {
-				mdev->rs_mark_time =jiffies;
-				mdev->rs_mark_left =drbd_bm_total_weight(mdev);
-			}
+			mdev->rs_mark_time = jiffies;
+			mdev->rs_mark_left = drbd_bm_total_weight(mdev);
 		}
 		drbd_try_clear_on_disk_bm(mdev,sector,count);
+		spin_unlock_irq(&mdev->al_lock);
 		/* just wake_up unconditional now,
 		 * various lc_chaged(), lc_put() in drbd_try_clear_on_disk_bm(). */
-		wake_up=1;
+		wake_up(&mdev->al_wait);
 	}
-	spin_unlock_irq(&mdev->al_lock);
-	if(wake_up) wake_up(&mdev->al_wait);
 }
 
 /*
