@@ -5,9 +5,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 2004-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2004-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2004-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 2004, Lars Ellenberg <l.g.e@web.de>.
+	main author.
+
+   Copyright (C) 2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	contributions.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -31,49 +33,6 @@
 #include <linux/drbd.h>
 #include "drbd_int.h"
 
-/* special handling for ppc64 on 2.4 kernel -- find_next_bit is not exported
- * so we include it here (verbatim, from linux 2.4.21 sources) */
-#if defined(__powerpc64__) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-
-unsigned long find_next_bit(unsigned long *addr, unsigned long size, unsigned long offset)
-{
-        unsigned long *p = addr + (offset >> 6);
-        unsigned long result = offset & ~63UL;
-        unsigned long tmp;
-
-        if (offset >= size)
-                return size;
-        size -= result;
-        offset &= 63UL;
-        if (offset) {
-                tmp = *(p++);
-                tmp &= (~0UL << offset);
-                if (size < 64)
-                        goto found_first;
-                if (tmp)
-                        goto found_middle;
-                size -= 64;
-                result += 64;
-        }
-        while (size & ~63UL) {
-                if ((tmp = *(p++)))
-                        goto found_middle;
-                result += 64;
-                size -= 64;
-        }
-        if (!size)
-                return result;
-        tmp = *p;
-
-found_first:
-        tmp &= (~0UL >> (64 - size));
-        if (tmp == 0UL)        /* Are any bits set? */
-                return result + size; /* Nope. */
-found_middle:
-        return result + __ffs(tmp);
-}
-#endif /* NEED_PPC64_WORKAROUND */
-
 /* OPAQUE outside this file!
  * interface defined in drbd_int.h
  *
@@ -271,12 +230,6 @@ int drbd_bm_init(drbd_dev *mdev)
 	return 0;
 }
 
-sector_t drbd_bm_capacity(drbd_dev *mdev)
-{
-	ERR_IF(!mdev->bitmap) return 0;
-	return mdev->bitmap->bm_dev_capacity;
-}
-
 /* called on driver unload. TODO: call when a device is destroyed.
  */
 void drbd_bm_cleanup(drbd_dev *mdev)
@@ -316,33 +269,6 @@ STATIC int bm_clear_surplus(struct drbd_bitmap * b)
 	return cleared;
 }
 
-STATIC void bm_set_surplus(struct drbd_bitmap * b)
-{
-	const unsigned long mask = (1UL << (b->bm_bits & (BITS_PER_LONG-1))) -1;
-	size_t w = b->bm_bits >> LN2_BPL;
-
-	if ( w < b->bm_words ) {
-		b->bm[w++] |= ~mask;
-	}
-
-	if ( w < b->bm_words ) {
-		b->bm[w++] = ~(0UL);
-	}
-}
-
-STATIC unsigned long bm_count_bits(struct drbd_bitmap * b)
-{
-	unsigned long *bm = b->bm;
-	unsigned long *ep = b->bm + b->bm_words;
-	unsigned long bits = 0;
-
-	while ( bm < ep ) {
-		bits += hweight_long(*bm++);
-	}
-
-	return bits;
-}
-
 #define BM_SECTORS_PER_BIT (BM_BLOCK_SIZE/512)
 
 /*
@@ -357,9 +283,9 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long bits, bytes, words, *nbm, *obm = 0;
-	int err = 0, growing;
+	int err = 0;
 
-	ERR_IF(!b) return -ENOMEM;
+	D_BUG_ON(!b);
 	MUST_BE_LOCKED();
 
 	ERR_IF (down_trylock(&b->bm_change)) {
@@ -418,23 +344,19 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		obm = b->bm;
 		// brgs. move several MB within spinlock...
 		if (obm) {
-			bm_set_surplus(b);
 			D_ASSERT(b->bm[b->bm_words] == DRBD_MAGIC);
 			memcpy(nbm,obm,min_t(size_t,b->bm_words,words)*sizeof(long));
 		}
-		growing = words > b->bm_words;
-		if (growing) { // set all newly allocated bits
-			memset( nbm+b->bm_words, -1,
-				(words - b->bm_words) * sizeof(long) );
-			b->bm_set  += bits - b->bm_bits;
+		if (b->bm_words < words) { // set all newly allocated bits
+			memset(nbm,-1,(words - b->bm_words)*sizeof(long));
 		}
 		nbm[words] = DRBD_MAGIC;
 		b->bm = nbm;
+		b->bm_set  += bits - b->bm_bits;
 		b->bm_bits  = bits;
 		b->bm_words = words;
 		b->bm_dev_capacity = capacity;
 		bm_clear_surplus(b);
-		if( !growing ) b->bm_set = bm_count_bits(b);
 		bm_end_info(mdev, __FUNCTION__ );
 		spin_unlock_irq(&b->bm_lock);
 		INFO("resync bitmap: bits=%lu words=%lu\n",bits,words);
@@ -460,7 +382,7 @@ unsigned long drbd_bm_total_weight(drbd_dev *mdev)
 	unsigned long s;
 	unsigned long flags;
 
-	ERR_IF(!b) return 0;
+	D_BUG_ON(!(b && b->bm));
 	// MUST_BE_LOCKED(); well. yes. but ...
 
 	spin_lock_irqsave(&b->bm_lock,flags);
@@ -473,7 +395,7 @@ unsigned long drbd_bm_total_weight(drbd_dev *mdev)
 size_t drbd_bm_words(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	ERR_IF(!b) return 0;
+	D_BUG_ON(!(b && b->bm));
 
 	/* FIXME
 	 * actually yes. really. otherwise it could just change its size ...
@@ -495,8 +417,7 @@ void drbd_bm_merge_lel( drbd_dev *mdev, size_t offset, size_t number,
 	unsigned long word, bits;
 	size_t n = number;
 
-	ERR_IF(!b) return;
-	ERR_IF(!b->bm) return;
+	D_BUG_ON(!(b && b->bm));
 	D_BUG_ON(offset        >= b->bm_words);
 	D_BUG_ON(offset+number >  b->bm_words);
 	D_BUG_ON(number > PAGE_SIZE/sizeof(long));
@@ -535,8 +456,7 @@ void drbd_bm_set_lel( drbd_dev *mdev, size_t offset, size_t number,
 	unsigned long word, bits;
 	size_t n = number;
 
-	ERR_IF(!b) return;
-	ERR_IF(!b->bm) return;
+	D_BUG_ON(!(b && b->bm));
 	D_BUG_ON(offset        >= b->bm_words);
 	D_BUG_ON(offset+number >  b->bm_words);
 	D_BUG_ON(number > PAGE_SIZE/sizeof(long));
@@ -573,8 +493,7 @@ void drbd_bm_get_lel( drbd_dev *mdev, size_t offset, size_t number,
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long *bm;
 
-	ERR_IF(!b) return;
-	ERR_IF(!b->bm) return;
+	D_BUG_ON(!(b && b->bm));
 	if ( (offset        >= b->bm_words) ||
 	     (offset+number >  b->bm_words) ||
 	     (number > PAGE_SIZE/sizeof(long)) ||
@@ -600,8 +519,7 @@ void drbd_bm_get_lel( drbd_dev *mdev, size_t offset, size_t number,
 void drbd_bm_set_all(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	ERR_IF(!b) return;
-	ERR_IF(!b->bm) return;
+	D_BUG_ON(!(b && b->bm));
 
 	MUST_BE_LOCKED();
 
@@ -745,8 +663,7 @@ void drbd_bm_write(struct Drbd_Conf *mdev)
 void drbd_bm_clear_all(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	ERR_IF(!b) return;
-	ERR_IF(!b->bm) return;
+	D_BUG_ON(!(b && b->bm));
 
 	MUST_BE_LOCKED();						\
 
@@ -760,7 +677,7 @@ void drbd_bm_clear_all(drbd_dev *mdev)
 void drbd_bm_reset_find(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	ERR_IF(!b) return;
+	D_BUG_ON(!(b && b->bm));
 
 	MUST_BE_LOCKED();
 
@@ -780,8 +697,7 @@ unsigned long drbd_bm_find_next(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long i = -1UL;
-	ERR_IF(!b) return i;
-	ERR_IF(!b->bm) return i;
+	D_BUG_ON(!(b && b->bm));
 
 	spin_lock_irq(&b->bm_lock);
 	BM_PARANOIA_CHECK();
@@ -813,10 +729,8 @@ int drbd_bm_rs_done(drbd_dev *mdev)
 int drbd_bm_set_bit(drbd_dev *mdev, const unsigned long bitnr)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	unsigned long flags;
 	int i;
-	ERR_IF(!b) return 1;
-	ERR_IF(!b->bm) return 1;
+	D_BUG_ON(!(b && b->bm));
 
 /*
  * only called from drbd_set_out_of_sync.
@@ -829,7 +743,7 @@ int drbd_bm_set_bit(drbd_dev *mdev, const unsigned long bitnr)
 		ERR("%s in drbd_bm_set_bit\n", cstate_to_name(mdev->cstate));
 */
 
-	spin_lock_irqsave(&b->bm_lock,flags);
+	spin_lock_irq(&b->bm_lock);
 	BM_PARANOIA_CHECK();
 	MUST_NOT_BE_LOCKED();
 	ERR_IF (bitnr >= b->bm_bits) {
@@ -839,7 +753,7 @@ int drbd_bm_set_bit(drbd_dev *mdev, const unsigned long bitnr)
 		i = (0 != __test_and_set_bit(bitnr, b->bm));
 		b->bm_set += !i;
 	}
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irq(&b->bm_lock);
 	return i;
 }
 
@@ -849,12 +763,10 @@ int drbd_bm_set_bit(drbd_dev *mdev, const unsigned long bitnr)
 int drbd_bm_clear_bit(drbd_dev *mdev, const unsigned long bitnr)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	unsigned long flags;
 	int i;
-	ERR_IF(!b) return 0;
-	ERR_IF(!b->bm) return 0;
+	D_BUG_ON(!(b && b->bm));
 
-	spin_lock_irqsave(&b->bm_lock,flags);
+	spin_lock_irq(&b->bm_lock);
 	BM_PARANOIA_CHECK();
 	MUST_NOT_BE_LOCKED();
 	ERR_IF (bitnr >= b->bm_bits) {
@@ -864,7 +776,7 @@ int drbd_bm_clear_bit(drbd_dev *mdev, const unsigned long bitnr)
 		i = (0 != __test_and_clear_bit(bitnr, b->bm));
 		b->bm_set -= i;
 	}
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irq(&b->bm_lock);
 
 	/* clearing bits should only take place when sync is in progress!
 	 * this is only called from drbd_set_in_sync.
@@ -885,8 +797,7 @@ int drbd_bm_test_bit(drbd_dev *mdev, const unsigned long bitnr)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	int i;
-	ERR_IF(!b) return 0;
-	ERR_IF(!b->bm) return 0;
+	D_BUG_ON(!(b && b->bm));
 
 	spin_lock_irq(&b->bm_lock);
 	BM_PARANOIA_CHECK();
@@ -920,8 +831,7 @@ int drbd_bm_e_weight(drbd_dev *mdev, unsigned long enr)
 	int count, s, e;
 	unsigned long flags;
 
-	ERR_IF(!b) return 0;
-	ERR_IF(!b->bm) return 0;
+	D_BUG_ON(!(b && b->bm));
 	spin_lock_irqsave(&b->bm_lock,flags);
 	BM_PARANOIA_CHECK();
 
@@ -948,8 +858,7 @@ unsigned long drbd_bm_ALe_set_all(drbd_dev *mdev, unsigned long al_enr)
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long weight;
 	int count, s, e;
-	ERR_IF(!b) return 0;
-	ERR_IF(!b->bm) return 0;
+	D_BUG_ON(!(b && b->bm));
 
 	MUST_BE_LOCKED();
 
