@@ -5,9 +5,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	main author.
+
+   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
+	main contributor.
 
    Copyright (C) 2000, Fbio Oliv Leite <olive@conectiva.com.br>.
 	Some sanity checks in IOCTL_SET_STATE.
@@ -28,7 +30,7 @@
 
  */
 
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/module.h>
 
 #include <asm/uaccess.h>
@@ -66,6 +68,26 @@ STATIC enum { NotMounted=0,MountedRO,MountedRW } drbd_is_mounted(int minor)
 }
 #endif
 
+STATIC int do_determin_dev_size(struct Drbd_Conf* mdev);
+int drbd_determin_dev_size(struct Drbd_Conf* mdev)
+{
+	sector_t pmdss; // previous meta data start sector
+	int rv;
+
+	wait_event(mdev->al_wait, lc_try_lock(mdev->act_log));
+	pmdss = drbd_md_ss(mdev);
+	rv = do_determin_dev_size(mdev);
+	if ( pmdss != drbd_md_ss(mdev) && mdev->md_index == -1 ) {
+		WARN("Moving meta-data.\n");
+		drbd_al_shrink(mdev); // All extents inactive.
+		drbd_bm_write(mdev);  // write bitmap
+		drbd_md_write(mdev);  // Write mdev->la_size to disk.
+	}
+	lc_unlock(mdev->act_log);
+
+	return rv;
+}
+
 char* ppsize(char* buf, size_t size) 
 {
 	// Needs 9 bytes at max.
@@ -80,7 +102,11 @@ char* ppsize(char* buf, size_t size)
 	return buf;
 }
 
-/* Returns -ENOMEM if we could not allocate the bitmap
+
+/* Returns 1 if there is a disk-less node, 0 if both nodes have a disk. */
+/*
+ * THINK do we want the size to be KB or sectors ?
+ * note, *_capacity operates in 512 byte sectors!!
  *
  * currently *_size is in KB.
  *
@@ -91,80 +117,15 @@ char* ppsize(char* buf, size_t size)
  * but again, this is a state change, and thus should be serialized with other
  * state changes on a more general level already.
  */
-int drbd_determin_dev_size(struct Drbd_Conf* mdev)
-{
-	sector_t pmdss; // previous meta data start sector
-	sector_t la_size;
-	sector_t size;
-	char ppb[10];
-
-	int md_moved, la_size_changed;
-	int rv=0;
-
-	wait_event(mdev->al_wait, lc_try_lock(mdev->act_log));
-	pmdss = drbd_md_ss(mdev);
-	la_size = mdev->la_size;
-
-	size = drbd_new_dev_size(mdev);
-
-	if( (drbd_get_capacity(mdev->this_bdev)>>1) != size ) {
-		int err;
-		err = drbd_bm_resize(mdev,size<<1); // wants sectors
-		if (unlikely(err)) {
-			/* currently there is only one error: ENOMEM! */
-			size = drbd_bm_capacity(mdev)>>1;
-			if (size == 0) {
-				ERR("OUT OF MEMORY! Could not allocate bitmap! Set device size => 0\n");
-			} else {
-				/* FIXME this is problematic,
-				 * if we in fact are smaller now! */
-				ERR("BM resizing failed. "
-				    "Leaving size unchanged at size = %lu KB\n", 
-				    (unsigned long)size);
-			}
-			rv = err;
-		}
-		// racy, see comments above.
-		drbd_set_my_capacity(mdev,size<<1);
-		mdev->la_size = size;
-		INFO("size = %s (%lu KB)\n",ppsize(ppb,size),
-		     (unsigned long)size);
-	}
-	if (rv < 0) goto out;
-
-	la_size_changed = (la_size != mdev->la_size);
-	md_moved = pmdss != drbd_md_ss(mdev) /* && mdev->md_index == -1 */;
-
-	if ( md_moved ) {
-		WARN("Moving meta-data.\n");
-		D_ASSERT(mdev->md_index == -1);
-	}
-
-	if ( la_size_changed || md_moved ) {
-		if( inc_local_md_only(mdev)) {
-			drbd_al_shrink(mdev); // All extents inactive.
-			drbd_bm_write(mdev);  // write bitmap
-			// Write mdev->la_size to on disk.
-			drbd_md_write(mdev);
-			dec_local(mdev);
-		}
-	}
-  out:
-	lc_unlock(mdev->act_log);
-
-	return rv;
-}
-
-/*
- * currently *_size is in KB.
- */
-sector_t drbd_new_dev_size(struct Drbd_Conf* mdev)
+STATIC int do_determin_dev_size(struct Drbd_Conf* mdev)
 {
 	sector_t p_size = mdev->p_size;  // partner's disk size.
 	sector_t la_size = mdev->la_size; // last agreed size.
 	sector_t m_size; // my size
 	sector_t u_size = mdev->lo_usize; // size requested by user.
 	sector_t size=0;
+	int rv;
+	char ppb[10];
 
 	m_size = drbd_get_capacity(mdev->backing_bdev)>>1;
 
@@ -174,8 +135,10 @@ sector_t drbd_new_dev_size(struct Drbd_Conf* mdev)
 	}
 
 	if(p_size && m_size) {
+		rv=0;
 		size=min_t(sector_t,p_size,m_size);
 	} else {
+		rv=1;
 		if(la_size) {
 			size=la_size;
 			if(m_size && m_size < size) size=m_size;
@@ -199,7 +162,23 @@ sector_t drbd_new_dev_size(struct Drbd_Conf* mdev)
 		}
 	}
 
-	return size;
+	if( (drbd_get_capacity(mdev->this_bdev)>>1) != size ) {
+		int err;
+		err = drbd_bm_resize(mdev,size<<1); // wants sectors
+		if (unlikely(err)) {
+			ERR("BM resizing failed. "
+			    "Leaving size unchanged at size = %lu KB\n", 
+			    (unsigned long)size);
+		} else {
+			// racy, see comments above.
+			drbd_set_my_capacity(mdev,size<<1);
+			mdev->la_size = size;
+			INFO("size = %s (%lu KB)\n",ppsize(ppb,size),
+			     (unsigned long)size);
+		}
+	}
+
+	return rv;
 }
 
 /* checks that the al lru is of requested size, and if neccessary tries to
@@ -250,18 +229,15 @@ STATIC int drbd_check_al_size(drbd_dev *mdev)
 	} else {
 		if (t) lc_free(t);
 	}
-	drbd_md_write(mdev);
 	return 0;
 }
 
-STATIC int drbd_detach_ioctl(drbd_dev *mdev);
-
 STATIC
 int drbd_ioctl_set_disk(struct Drbd_Conf *mdev,
 			struct ioctl_disk_config * arg)
 {
 	NOT_IN_26(int err;) // unused in 26 ?? cannot believe it ...
-	int i, md_gc_valid, minor, mput=0, apply_al;
+	int i, md_gc_valid, minor, mput=0;
 	enum ret_codes retcode;
 	struct disk_config new_conf;
 	struct file *filp = 0;
@@ -309,9 +285,6 @@ int drbd_ioctl_set_disk(struct Drbd_Conf *mdev,
 		__module_get(THIS_MODULE);
 		mput = 1;
 	} else {
-		/* We currently cannot handle reattach while connected */
-		return -EBUSY;
-
 		/* FIXME allow reattach while connected,
 		 * and allow it in Primary/Diskless state...
 		 * currently there are strange races leading to a distributed
@@ -460,21 +433,13 @@ ONLY_IN_26({
 	request_queue_t * const q = mdev->rq_queue;
 	request_queue_t * const b = bdev->bd_disk->queue;
 
-	q->max_sectors = min_not_zero((typeof(q->max_sectors))(PAGE_SIZE >> 9), b->max_sectors);
+	q->max_sectors = min_not_zero((unsigned short)(PAGE_SIZE >> 9), b->max_sectors);
 	q->max_phys_segments = 1;
 	q->max_hw_segments   = 1;
 	q->max_segment_size  = min((unsigned)PAGE_SIZE,b->max_segment_size);
 	q->hardsect_size     = max((unsigned short)512,b->hardsect_size);
 	q->seg_boundary_mask = PAGE_SIZE-1;
 	D_ASSERT(q->hardsect_size <= PAGE_SIZE); // or we are really screwed ;-)
-
-	if( q->backing_dev_info.ra_pages != b->backing_dev_info.ra_pages) {
-		INFO("Adjusting my ra_pages to backing device's (%lu -> %lu)\n",
-		     q->backing_dev_info.ra_pages,
-		     b->backing_dev_info.ra_pages);
-		q->backing_dev_info.ra_pages = b->backing_dev_info.ra_pages;
-	}
-
 })
 #undef min_not_zero
 
@@ -489,68 +454,11 @@ ONLY_IN_26({
 
 /* FIXME if (md_gc_valid < 0) META DATA IO NOT POSSIBLE! */
 
-	/* If I am currently not Primary,
-	 * but meta data primary indicator is set,
-	 * I just now recover from a hard crash,
-	 * and have been Primary before that crash.
-	 *
-	 * Now, if I had no connection before that crash
-	 * (have been degraded Primary), chances are that
-	 * I won't find my peer now either.
-	 *
-	 * In that case, and _only_ in that case,
-	 * we use the degr-wfc-timeout instead of the default,
-	 * so we can automatically recover from a crash of a
-	 * degraded but active "cluster" after a certain timeout.
-	 */
-	clear_bit(USE_DEGR_WFC_T,&mdev->flags);
-	if ( mdev->state != Primary &&
-	     drbd_md_test_flag(mdev,MDF_PrimaryInd) &&
-	    !drbd_md_test_flag(mdev,MDF_ConnectedInd) ) {
-		set_bit(USE_DEGR_WFC_T,&mdev->flags);
-	}
-
 	drbd_bm_lock(mdev); // racy...
-
-	if(drbd_md_test_flag(mdev,MDF_Consistent) &&
-	   drbd_new_dev_size(mdev) < mdev->la_size ) {
-		D_ASSERT(mdev->cstate == Unconfigured);
-		D_ASSERT(mput == 1);
-		/* Do not attach a too small disk.*/
-		drbd_bm_unlock(mdev);
-		ERR("Lower device smaller than last agreed size!\n");
-		drbd_free_ll_dev(mdev);
-		set_cstate(mdev,Unconfigured);
-		retcode = LDDeviceTooSmall;
-		module_put(THIS_MODULE);
-		if (put_user(retcode, &arg->ret_code)) return -EFAULT;
-		return -EINVAL;
-	}
-
-	apply_al = drbd_md_test_flag(mdev,MDF_PrimaryInd);
-	if(apply_al) set_bit(CRASHED_PRIMARY, &mdev->flags);
-	else       clear_bit(CRASHED_PRIMARY, &mdev->flags);
-	/* All tests on MDF_PrimaryInd and MDF_ConnectedInd must happen before 
-	   this point, because determin_dev_size() might call drbd_md_write(), 
-	   which in turn modifies these flags. Exceptions are where, we want
-	   to test the current state (drbd_md_compare(), drbd_send_param()). */
-
-	if (drbd_determin_dev_size(mdev) < 0) {
-		/* could not allocate bitmap.
-		 * try to undo ... */
-		D_ASSERT(mdev->cstate == Unconfigured);
-		D_ASSERT(mput == 1);
-
-		drbd_bm_unlock(mdev);
-
-		/* from drbd_detach_ioctl */
-		drbd_free_ll_dev(mdev);
-
-		set_cstate(mdev,Unconfigured);
-		drbd_mdev_cleanup(mdev);
-		module_put(THIS_MODULE);
-		return -ENOMEM;
-	}
+	drbd_determin_dev_size(mdev);
+	/* FIXME
+	 * what if we now have la_size == 0 ?? eh?
+	 */
 
 	if (md_gc_valid <= 0) {
 		INFO("Assuming that all blocks are out of sync (aka FullSync)\n");
@@ -565,6 +473,7 @@ ONLY_IN_26({
 
 	i = drbd_check_al_size(mdev);
 	if (i) {
+// FATAL!
 		/* FIXME see the comment above.
 		 * if this fails I need to undo all changes,
 		 * go back into Unconfigured,
@@ -572,23 +481,15 @@ ONLY_IN_26({
 		 */
 		// return i;
 		drbd_panic("Cannot allocate act_log\n");
+		drbd_suicide();
 	}
 
 	if (md_gc_valid > 0) {
-		i = drbd_al_read_log(mdev);
-		if (apply_al && i) {
+		drbd_al_read_log(mdev);
+		if (drbd_md_test_flag(mdev,MDF_PrimaryInd)) {
 			drbd_al_apply_to_bm(mdev);
 			drbd_al_to_on_disk_bm(mdev);
 		}
-		if(!i) {
-			ERR("IO error on meta device while reading AL\n");
-			drbd_free_ll_dev(mdev);
-			set_cstate(mdev,Unconfigured);
-			retcode = MDIOError;
-			module_put(THIS_MODULE);
-			if (put_user(retcode, &arg->ret_code)) return -EFAULT;
-			return -EINVAL;
-		}
 	} /* else {
 	     FIXME wipe out on disk al!
 	} */
@@ -809,8 +710,6 @@ int drbd_set_state(drbd_dev *mdev,Drbd_State newstate)
 		 * become Secondary. */
 		if (bd_claim(mdev->this_bdev,drbd_sec_holder))
 			return -EBUSY;
-		if (disable_bd_claim)
-			bd_release(mdev->this_bdev);
 	}
 #endif
 
@@ -924,7 +823,6 @@ ONLY_IN_26(
 			    mdev->cstate >= Connected ?
 			    ConnectedCnt : ArbitraryCnt);
 		}
-		clear_bit(CRASHED_PRIMARY, &mdev->flags);
 	} else {
 		NOT_IN_26( set_device_ro(MKDEV(MAJOR_NR, minor), TRUE ); )
 		ONLY_IN_26( set_disk_ro(mdev->vdisk, TRUE ); )
@@ -954,11 +852,12 @@ static int drbd_get_wait_time(long *tp, struct Drbd_Conf *mdev,
 		return -EFAULT;
 	}
 
-	if ( test_bit(USE_DEGR_WFC_T,&mdev->flags) ) {
-		time=p.degr_wfc_timeout;
-		if (time) WARN("using degr_wfc_timeout=%ld seconds\n", time);
-	} else {
+	if( drbd_md_test_flag(mdev,MDF_ConnectedInd) ) {
 		time=p.wfc_timeout;
+		//ERR("using wfc_timeout.\n");
+	} else {
+		time=p.degr_wfc_timeout;
+		//ERR("using degr_wfc_timeout.\n");
 	}
 
 	time=time*HZ;
@@ -1065,18 +964,6 @@ STATIC int drbd_detach_ioctl(drbd_dev *mdev)
 	return 0;
 }
 
-#ifdef CONFIG_COMPAT
-long drbd_compat_ioctl(struct file *f, unsigned cmd, unsigned long arg)
-{
-	int ret;
-	ret = drbd_ioctl(f->f_dentry->d_inode, f, cmd, arg);
-	/* need to map "unknown" to ENOIOCTLCMD
-	 * to get the generic fallback path going */
-	if (ret == -ENOTTY) ret = -ENOIOCTLCMD;
-	return ret;
-}
-#endif
-
 int drbd_ioctl(struct inode *inode, struct file *file,
 			   unsigned int cmd, unsigned long arg)
 {
@@ -1124,9 +1011,7 @@ ONLY_IN_26(
 		switch (cmd) {
 		default:
 			/* oops, unknown IOCTL ?? */
-			/* Sorry, 2.4 kernel will get "no such ioctl" for BLK* ioctls
-			 * on an unconfigured DRBD.  I don't care right now. */
-			err = -ENOTTY;
+			err = -EINVAL;
 			goto out_unlocked;
 
 		case DRBD_IOCTL_GET_CONFIG:
@@ -1179,10 +1064,6 @@ ONLY_IN_26(
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 /* see how sys_ioctl and blkdev_ioctl handle it in 2.6 .
  * If I understand correctly, only "private" ioctl end up here.
- *
- * unless, of course, we are called by the compat_ioctl, in which case our
- * drbd_compat_ioctl wrapper will convert ENOTTY into ENOIOCTLCMD, which in
- * turn will cause the fallback path in the generic compat_sys_ioctl.
  */
 	case BLKGETSIZE:
 		err = put_user(drbd_get_capacity(mdev->this_bdev),(long *)arg);
@@ -1224,26 +1105,10 @@ ONLY_IN_26(
 		} else {
 			clear_bit(ON_PRI_INC_HUMAN,&mdev->flags);
 			clear_bit(ON_PRI_INC_TIMEOUTEX,&mdev->flags);
-			if (arg == 0) break;
 
-			// XXX reduce race: don't set it,
-			// if we have a connection.
-			// this does not avoid the race completely, though.
-			if (mdev->cstate > WFConnection) {
-				WARN("race avoidance: did not set "
-				     "the state flags (%s), cstate=%s\n",
-				        arg == (Human|TimeoutExpired)
-				     ?  "Human|TimeoutExpired"
-				     : arg == Human
-				     ? "Human"
-				     : "TimeoutExpired",
-				     cstate_to_name(mdev->cstate));
-				break;
-			}
-
-			if (arg & Human)
+			if (arg & Human ) 
 				set_bit(ON_PRI_INC_HUMAN,&mdev->flags);
-			if (arg & TimeoutExpired)
+			if (arg & TimeoutExpired )
 				set_bit(ON_PRI_INC_TIMEOUTEX,&mdev->flags);
 		}
 		break;
@@ -1257,17 +1122,13 @@ ONLY_IN_26(
 			err = -EBUSY;
 			break;
 		}
-		if ( mdev->state == Secondary && mdev->o_state == Secondary) {
-			err = -EINPROGRESS;
-			break;
-		}
 		err=0;
 		mdev->lo_usize = (unsigned long)arg;
 		drbd_bm_lock(mdev);
 		drbd_determin_dev_size(mdev);
 		drbd_md_write(mdev); // Write mdev->la_size to disk.
 		drbd_bm_unlock(mdev);
-		if (mdev->cstate == Connected) drbd_send_param(mdev,1);
+		if (mdev->cstate == Connected) drbd_send_param(mdev,0);
 		break;
 
 	case DRBD_IOCTL_SET_NET_CONFIG:
@@ -1293,7 +1154,8 @@ ONLY_IN_26(
 			err=-ENODATA;
 			break;
 		}
-
+		/* FIXME what if fsync returns error */
+		drbd_sync_me(mdev);
 		set_bit(DO_NOT_INC_CONCNT,&mdev->flags);
 		set_cstate(mdev,Unconnected);
 		drbd_thread_stop(&mdev->receiver);
@@ -1379,19 +1241,15 @@ ONLY_IN_26(
 		/* disallow "invalidation" of local replica
 		 * when currently in primary state (would be a Bad Idea),
 		 * or during a running sync (won't make any sense) */
-		if( (mdev->state == Primary ||
-		      (mdev->cstate != Connected &&
-		       mdev->cstate != StandAlone)) ||
+		if( mdev->state == Primary ||
+		    mdev->cstate < StandAlone ||
+		    mdev->cstate > Connected ||
 		    test_bit(DISKLESS,&mdev->flags) ||
 		    test_bit(PARTNER_DISKLESS,&mdev->flags) ) {
 			err = -EINPROGRESS;
 			break;
 		}
 
-		drbd_md_set_flag(mdev,MDF_FullSync);
-		drbd_md_clear_flag(mdev,MDF_Consistent);
-		drbd_md_write(mdev);
-
 		if (mdev->cstate == Connected) {
 			/* avoid races with set_in_sync
 			 * for successfull mirrored writes
@@ -1403,6 +1261,10 @@ ONLY_IN_26(
 
 		drbd_bm_lock(mdev); // racy...
 
+		drbd_md_set_flag(mdev,MDF_FullSync);
+		drbd_md_clear_flag(mdev,MDF_Consistent);
+		drbd_md_write(mdev);
+
 		drbd_bm_set_all(mdev);
 		drbd_bm_write(mdev);
 
@@ -1458,10 +1320,11 @@ ONLY_IN_26(
 		break;
 
 	default:
-		err = -ENOTTY;
+		err = -EINVAL;
 	}
  /* out: */
 	up(&mdev->device_mutex);
  out_unlocked:
 	return err;
 }
+
