@@ -4,9 +4,11 @@
 
   This file is part of drbd by Philipp Reisner.
 
-  Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-  Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-  Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+  Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	main author.
+
+  Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
+	main contributor.
 
   drbd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -23,6 +25,7 @@
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
 */
+
 #include <linux/compiler.h>
 #include <linux/types.h>
 #include <linux/version.h>
@@ -57,11 +60,6 @@ static inline void __list_splice(struct list_head *list,
 	last->next = at;
 	at->prev = last;
 }
-static inline void list_splice(struct list_head *list, struct list_head *head)
-{
-	if (!list_empty(list))
-		__list_splice(list, head);
-}
 static inline void list_splice_init(struct list_head *list,
 				    struct list_head *head)
 {
@@ -74,14 +72,12 @@ static inline void list_splice_init(struct list_head *list,
 
 // module parameter, defined in drbd_main.c
 extern int minor_count;
-extern int disable_bd_claim;
+extern int disable_io_hints;
 extern int major_nr;
 extern int use_nbd_major;
 
-#ifndef NO_MORE_DEV_FS
 // use_nbd_major ? "nbd" : "drbd";
 extern char* drbd_devfs_name;
-#endif
 
 #include <linux/major.h>
 #ifdef DRBD_MAJOR
@@ -219,11 +215,11 @@ typedef struct Drbd_Conf drbd_dev;
 
 #ifdef DBG_ASSERTS
 extern void drbd_assert_breakpoint(drbd_dev*, char *, char *, int );
-# define D_ASSERT(exp)  do { if (!(exp)) \
-	 drbd_assert_breakpoint(mdev,#exp,__FILE__,__LINE__); } while (0)
+# define D_ASSERT(exp)  if (!(exp)) \
+	 drbd_assert_breakpoint(mdev,#exp,__FILE__,__LINE__)
 #else
-# define D_ASSERT(exp)  do { if (!(exp)) \
-	 ERR("ASSERT( " #exp " ) in %s:%d\n", __FILE__,__LINE__); } while (0) 
+# define D_ASSERT(exp)  if (!(exp)) \
+	 ERR("ASSERT( " #exp " ) in %s:%d\n", __FILE__,__LINE__)
 #endif
 #define ERR_IF(exp) if (({ \
 	int _b = (exp)!=0; \
@@ -294,6 +290,21 @@ extern void drbd_assert_breakpoint(drbd_dev*, char *, char *, int );
 	( typecheck(struct Drbd_Conf*,x) && \
 	  ((x) ? (((x)->magic ^ DRBD_MAGIC) == (long)(x)):0))
 
+
+/*
+ * GFP_DRBD is used for allocations inside drbd_make_request,
+ * and for the sk->allocation scheme.
+ *
+ * Try to get away with GFP_NOIO, which is
+ * in 2.4.x:	(__GFP_HIGH | __GFP_WAIT) // HIGH == EMERGENCY, not HIGHMEM!
+ * in 2.6.x:	             (__GFP_WAIT)
+ *
+ * As far as i can see we do not allocate from interrupt context...
+ * if we do, we certainly should fix that.
+ * - lge
+ */
+#define GFP_DRBD GFP_NOIO
+
 /* these defines should go into blkdev.h
    (if it will be ever includet into linus' linux) */
 #define RQ_DRBD_NOTHING	  0x0001
@@ -618,14 +629,7 @@ struct drbd_request {
 	int rq_status;
 	struct drbd_barrier *barrier; // The next barrier.
 	drbd_bio_t *master_bio;       // master bio pointer
-	unsigned int size;
-	sector_t sector;
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	drbd_bio_t private_bio;       // private bio struct
-#else
-	struct bio *private_bio;
-	drbd_dev *mdev;
-#endif
 };
 
 struct drbd_barrier {
@@ -689,9 +693,6 @@ enum {
 	SENT_DISK_FAILURE,	// sending it once is enough
 	MD_DIRTY,		// current gen counts and flags not yet on disk
 	SYNC_STARTED,		// Needed to agree on the exact point in time..
-	USE_DEGR_WFC_T,		// Use degr-wfc-timeout instad of wfc-timeout.
-	CRASHED_PRIMARY,        // This node was a crashed primary
-	IO_FROZEN               // IO Frozen.
 };
 
 struct drbd_bitmap; // opaque for Drbd_Conf
@@ -774,6 +775,7 @@ struct Drbd_Conf {
 	atomic_t unacked_cnt;    // Need to send replys for
 	atomic_t local_cnt;      // Waiting for local disk to signal completion
 	spinlock_t req_lock;
+	spinlock_t tl_lock;
 	struct drbd_barrier* newest_barrier;
 	struct drbd_barrier* oldest_barrier;
 	unsigned long flags;
@@ -794,7 +796,7 @@ struct Drbd_Conf {
 	atomic_t resync_locked;   // Number of locked elements in resync LRU
 	int open_cnt;
 	u32 gen_cnt[GEN_CNT_SIZE];
-	atomic_t epoch_size;
+	int epoch_size;
 	spinlock_t ee_lock;
 	struct list_head free_ee;   // available
 	struct list_head active_ee; // IO in progress
@@ -802,14 +804,13 @@ struct Drbd_Conf {
 	struct list_head done_ee;   // send ack
 	struct list_head read_ee;   // IO in progress
 	struct list_head net_ee;    // zero-copy network send in progress
-	drbd_bio_t *first_frozen_bio;
-	drbd_bio_t *last_frozen_bio;
 	spinlock_t pr_lock;
 	struct list_head app_reads;
 	struct list_head resync_reads;
 	int ee_vacant;
 	int ee_in_use;
 	wait_queue_head_t ee_wait;
+	struct list_head busy_blocks;
 	NOT_IN_26(struct tq_struct write_hint_tq;)
 	struct page *md_io_page;      // one page buffer for md_io
 	struct page *md_io_tmpp;     // in case hardsect != 512 [ s390 only? ]
@@ -835,8 +836,7 @@ extern void drbd_free_resources(drbd_dev *mdev);
 extern void tl_release(drbd_dev *mdev,unsigned int barrier_nr,
 		       unsigned int set_size);
 extern void tl_clear(drbd_dev *mdev);
-extern void tl_resend(drbd_dev *mdev);
-extern int tl_dependence(drbd_dev *mdev, drbd_request_t * item, int free_it);
+extern int tl_dependence(drbd_dev *mdev, drbd_request_t * item);
 extern void drbd_free_sock(drbd_dev *mdev);
 extern int drbd_send(drbd_dev *mdev, struct socket *sock,
 		     void* buf, size_t size, unsigned msg_flags);
@@ -844,9 +844,7 @@ extern int drbd_send_param(drbd_dev *mdev, int flags);
 extern int _drbd_send_cmd(drbd_dev *mdev, struct socket *sock,
 			  Drbd_Packet_Cmd cmd, Drbd_Header *h,
 			  size_t size, unsigned msg_flags);
-#define USE_DATA_SOCKET 1
-#define USE_META_SOCKET 0
-extern int drbd_send_cmd(drbd_dev *mdev, int use_data_socket,
+extern int drbd_send_cmd(drbd_dev *mdev, struct socket *sock,
 			  Drbd_Packet_Cmd cmd, Drbd_Header *h, size_t size);
 extern int drbd_send_sync_param(drbd_dev *mdev, struct syncer_config *sc);
 extern int drbd_send_b_ack(drbd_dev *mdev, u32 barrier_nr,
@@ -867,9 +865,6 @@ extern void drbd_free_ll_dev(drbd_dev *mdev);
 extern int drbd_io_error(drbd_dev* mdev);
 extern void drbd_mdev_cleanup(drbd_dev *mdev);
 
-extern int drbd_resend_barrier(drbd_dev *mdev,struct drbd_barrier *b);
-extern int drbd_resend_dblock(drbd_dev *mdev, drbd_request_t *req);
-
 // drbd_meta-data.c (still in drbd_main.c)
 extern void drbd_md_write(drbd_dev *mdev);
 extern int drbd_md_read(drbd_dev *mdev);
@@ -897,7 +892,7 @@ extern int drbd_md_test_flag(drbd_dev *mdev, int flag);
 #define MD_HARDSECT      (1<<MD_HARDSECT_B)
 
 // activity log
-#define AL_EXTENTS_PT    ((MD_HARDSECT-12)/8-1) // 61 ; Extents per 512B sector
+#define AL_EXTENTS_PT    (MD_HARDSECT-12)/8-1 // 61 ; Extents per 512B sector
 #define AL_EXTENT_SIZE_B 22      // One extent represents 4M Storage
 #define AL_EXTENT_SIZE (1<<AL_EXTENT_SIZE_B)
 
@@ -941,7 +936,7 @@ struct bm_extent {
 
 /* thus many _storage_ sectors are described by one bit */
 #define BM_SECT_TO_BIT(x)   ((x)>>(BM_BLOCK_SIZE_B-9))
-#define BM_BIT_TO_SECT(x)   ((sector_t)(x)<<(BM_BLOCK_SIZE_B-9))
+#define BM_BIT_TO_SECT(x)   ((x)<<(BM_BLOCK_SIZE_B-9))
 #define BM_SECT_PER_BIT     BM_BIT_TO_SECT(1)
 
 /* bit to represented kilo byte conversion */
@@ -977,15 +972,11 @@ struct bm_extent {
 #define BM_EXT_PER_SECT	    ( 512 / BM_BYTES_PER_EXTENT )        //   4
  */
 
-#define DRBD_MAX_SECTORS_32 (0xffffffffLU)
-#define DRBD_MAX_SECTORS_BM \
+#if ( !defined(CONFIG_LBD) ) && ( BITS_PER_LONG == 32 )
+# define DRBD_MAX_SECTORS (0xffffffffLU)
+#else 
+# define DRBD_MAX_SECTORS \
           ( (MD_RESERVED_SIZE*2LL - MD_BM_OFFSET) * (1LL<<(BM_EXT_SIZE_B-9)) )
-#if DRBD_MAX_SECTORS_BM < DRBD_MAX_SECTORS_32
-#define DRBD_MAX_SECTORS DRBD_MAX_SECTORS_BM
-#elif ( !defined(CONFIG_LBD) ) && ( BITS_PER_LONG == 32 )
-#define DRBD_MAX_SECTORS DRBD_MAX_SECTORS_32
-#else
-#define DRBD_MAX_SECTORS DRBD_MAX_SECTORS_BM
 #endif
 
 extern int  drbd_bm_init      (drbd_dev *mdev);
@@ -1004,7 +995,6 @@ extern void drbd_bm_read      (drbd_dev *mdev);
 extern void drbd_bm_write     (drbd_dev *mdev);
 extern unsigned long drbd_bm_ALe_set_all (drbd_dev *mdev, unsigned long al_enr);
 extern size_t        drbd_bm_words       (drbd_dev *mdev);
-extern sector_t      drbd_bm_capacity    (drbd_dev *mdev);
 extern unsigned long drbd_bm_find_next   (drbd_dev *mdev);
 extern unsigned long drbd_bm_total_weight(drbd_dev *mdev);
 extern int drbd_bm_rs_done(drbd_dev *mdev);
@@ -1028,13 +1018,13 @@ extern void drbd_bm_unlock    (drbd_dev *mdev);
 // drbd_main.c
 extern drbd_dev *drbd_conf;
 extern int minor_count;
-extern drbd_kmem_cache_t *drbd_request_cache;
-extern drbd_kmem_cache_t *drbd_ee_cache;
+extern kmem_cache_t *drbd_request_cache;
+extern kmem_cache_t *drbd_ee_cache;
 extern mempool_t *drbd_request_mempool;
 
 // drbd_req
-extern void _drbd_end_req(drbd_request_t *, int, int);
-extern void drbd_end_req(drbd_request_t *, int, int);
+#define ERF_NOTLD    2   /* do not call tl_dependence */
+extern void drbd_end_req(drbd_request_t *, int, int, sector_t);
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 extern int drbd_make_request_24(request_queue_t *q, int rw, struct buffer_head *bio);
 #else
@@ -1042,15 +1032,12 @@ extern int drbd_make_request_26(request_queue_t *q, struct bio *bio);
 #endif
 extern int drbd_read_remote(drbd_dev *mdev, drbd_request_t *req);
 
-extern void drbd_thaw_frozen_reqs(drbd_dev *mdev);
 // drbd_fs.c
 extern char* ppsize(char* buf, size_t size);
 extern int drbd_determin_dev_size(drbd_dev*);
-extern sector_t drbd_new_dev_size(struct Drbd_Conf*);
 extern int drbd_set_state(drbd_dev *mdev,Drbd_State newstate);
 extern int drbd_ioctl(struct inode *inode, struct file *file,
 		      unsigned int cmd, unsigned long arg);
-extern long drbd_compat_ioctl(struct file *f, unsigned cmd, unsigned long arg);
 
 // drbd_worker.c
 extern int drbd_worker(struct Drbd_thread *thi);
@@ -1082,7 +1069,7 @@ extern void drbd_wait_ee(drbd_dev *mdev,struct list_head *head);
 
 // drbd_proc.c
 extern struct proc_dir_entry *drbd_proc;
-extern struct file_operations drbd_proc_fops;
+extern int drbd_proc_get_info(char *, char **, off_t, int, int *, void *);
 extern const char* cstate_to_name(Drbd_CState s);
 extern const char* nodestate_to_name(Drbd_State s);
 
@@ -1092,7 +1079,7 @@ extern void drbd_al_complete_io(struct Drbd_Conf *mdev, sector_t sector);
 extern void drbd_rs_complete_io(struct Drbd_Conf *mdev, sector_t sector);
 extern int drbd_rs_begin_io(struct Drbd_Conf *mdev, sector_t sector);
 extern void drbd_rs_cancel_all(drbd_dev* mdev);
-extern int drbd_al_read_log(struct Drbd_Conf *mdev);
+extern void drbd_al_read_log(struct Drbd_Conf *mdev);
 extern void __drbd_set_in_sync(drbd_dev* mdev, sector_t sector, int size, const char* file, const unsigned int line);
 #define drbd_set_in_sync(mdev,sector,size) \
 	__drbd_set_in_sync(mdev,sector,size, __FILE__, __LINE__ )
@@ -1107,6 +1094,19 @@ extern void drbd_al_shrink(struct Drbd_Conf *mdev);
  * event macros
  *************************/
 
+// we use these within spin_lock_irq() ...
+#ifndef wq_write_lock
+#if USE_RW_WAIT_QUEUE_SPINLOCK
+# define wq_write_lock write_lock
+# define wq_write_unlock write_unlock
+# define wq_write_unlock_irq write_unlock_irq
+#else
+# define wq_write_lock spin_lock
+# define wq_write_unlock spin_unlock
+# define wq_write_unlock_irq spin_unlock_irq
+#endif
+#endif
+
 // sched.h does not have it with timeout, so here goes:
 
 #ifndef wait_event_interruptible_timeout
@@ -1148,14 +1148,6 @@ do {									\
 
 #include "drbd_compat_wrappers.h"
 
-static inline int drbd_disk_less_node_present(struct Drbd_Conf* mdev)
-{
-	sector_t p_size = mdev->p_size;
-	sector_t m_size = drbd_get_capacity(mdev->backing_bdev);
-
-	return ! ( p_size && m_size ) ;
-}
-
 static inline void
 drbd_flush_signals(struct task_struct *t)
 {
@@ -1260,9 +1252,8 @@ drbd_queue_work_front(drbd_dev *mdev, struct drbd_work_queue *q,
 	unsigned long flags;
 	spin_lock_irqsave(&mdev->req_lock,flags);
 	list_add(&w->list,&q->q);
-	up(&q->s); /* within the spinlock,
-		      see comment near end of drbd_worker() */
 	spin_unlock_irqrestore(&mdev->req_lock,flags);
+	up(&q->s);
 }
 
 static inline void
@@ -1272,9 +1263,8 @@ drbd_queue_work(drbd_dev *mdev, struct drbd_work_queue *q,
 	unsigned long flags;
 	spin_lock_irqsave(&mdev->req_lock,flags);
 	list_add_tail(&w->list,&q->q);
-	up(&q->s); /* within the spinlock,
-		      see comment near end of drbd_worker() */
 	spin_unlock_irqrestore(&mdev->req_lock,flags);
+	up(&q->s);
 }
 
 static inline void wake_asender(drbd_dev *mdev) {
@@ -1291,19 +1281,19 @@ static inline void request_ping(drbd_dev *mdev) {
 static inline int drbd_send_short_cmd(drbd_dev *mdev, Drbd_Packet_Cmd cmd)
 {
 	Drbd_Header h;
-	return drbd_send_cmd(mdev,USE_DATA_SOCKET,cmd,&h,sizeof(h));
+	return drbd_send_cmd(mdev,mdev->data.socket,cmd,&h,sizeof(h));
 }
 
 static inline int drbd_send_ping(drbd_dev *mdev)
 {
 	Drbd_Header h;
-	return drbd_send_cmd(mdev,USE_META_SOCKET,Ping,&h,sizeof(h));
+	return drbd_send_cmd(mdev,mdev->meta.socket,Ping,&h,sizeof(h));
 }
 
 static inline int drbd_send_ping_ack(drbd_dev *mdev)
 {
 	Drbd_Header h;
-	return drbd_send_cmd(mdev,USE_META_SOCKET,PingAck,&h,sizeof(h));
+	return drbd_send_cmd(mdev,mdev->meta.socket,PingAck,&h,sizeof(h));
 }
 
 static inline void drbd_thread_stop(struct Drbd_thread *thi)
@@ -1332,36 +1322,76 @@ static inline void inc_ap_pending(drbd_dev* mdev)
 		    __func__ , __LINE__ ,			\
 		    atomic_read(&mdev->which))
 
-#define dec_ap_pending(mdev) do {				\
+#define dec_ap_pending(mdev)					\
 	typecheck(drbd_dev*,mdev);				\
 	if(atomic_dec_and_test(&mdev->ap_pending_cnt))		\
 		wake_up(&mdev->cstate_wait);			\
-	ERR_IF_CNT_IS_NEGATIVE(ap_pending_cnt); } while (0)
+	ERR_IF_CNT_IS_NEGATIVE(ap_pending_cnt)
 
 static inline void inc_rs_pending(drbd_dev* mdev)
 {
 	atomic_inc(&mdev->rs_pending_cnt);
 }
 
-#define dec_rs_pending(mdev) do {				\
+#define dec_rs_pending(mdev)					\
 	typecheck(drbd_dev*,mdev);				\
 	atomic_dec(&mdev->rs_pending_cnt);			\
-	ERR_IF_CNT_IS_NEGATIVE(rs_pending_cnt); } while (0)
+	ERR_IF_CNT_IS_NEGATIVE(rs_pending_cnt)
 
 static inline void inc_unacked(drbd_dev* mdev)
 {
 	atomic_inc(&mdev->unacked_cnt);
 }
 
-#define dec_unacked(mdev) do {					\
+#if 0 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+/*
+ * idea was to forcefully push the tcp stack whenever the
+ * currently last pending packet is in the buffer.
+ * should be benchmarked on some real box to see if it has any
+ * effect on overall latency.
+ */
+
+/* this only works with 2.6 kernels because of some conflicting defines
+ * in header files included from net.tcp.h.
+ */
+
+#include <net/tcp.h>
+static inline void drbd_push_msock(drbd_dev* mdev)
+{
+	struct sock    *sk;
+	struct tcp_opt *tp;
+	if (mdev->meta.socket == NULL) return;
+	sk = mdev->meta.socket->sk;
+	tp = tcp_sk(sk);
+	lock_sock(sk);
+	__tcp_push_pending_frames(sk, tp, tcp_current_mss(sk, 1), TCP_NAGLE_PUSH);
+	release_sock(sk);
+}
+
+#define dec_unacked(mdev)					\
+	might_sleep();						\
+	typecheck(drbd_dev*,mdev);				\
+	if (atomic_dec_and_test(&mdev->unacked_cnt))		\
+		drbd_push_msock(mdev);				\
+	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt);
+
+#define sub_unacked(mdev, n)					\
+	might_sleep();						\
+	typecheck(drbd_dev*,mdev);				\
+	if (atomic_sub_and_test(n, &mdev->unacked_cnt))		\
+		drbd_push_msock(mdev);				\
+	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt);
+#else
+#define dec_unacked(mdev)					\
 	typecheck(drbd_dev*,mdev);				\
 	atomic_dec(&mdev->unacked_cnt);				\
-	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt); } while (0)
+	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt)
 
-#define sub_unacked(mdev, n) do {				\
+#define sub_unacked(mdev, n)					\
 	typecheck(drbd_dev*,mdev);				\
 	atomic_sub(n, &mdev->unacked_cnt);			\
-	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt); } while (0)
+	ERR_IF_CNT_IS_NEGATIVE(unacked_cnt)
+#endif
 
 
 /**
@@ -1519,7 +1549,7 @@ static inline unsigned long hweight_long(unsigned long w)
 
 static inline void drbd_suicide(void)
 {
-#ifdef TASK_ZOMBIE
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 	set_current_state(TASK_ZOMBIE);
 #else
 	current->exit_state = EXIT_ZOMBIE;
