@@ -5,9 +5,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	main author.
+
+   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
+	main contributor.
 
    Copyright (C) 2000, Marcelo Tosatti <marcelo@conectiva.com.br>.
 	Early 2.3.x work.
@@ -31,7 +33,7 @@
 
  */
 
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/module.h>
 #include <linux/version.h>
 
@@ -49,9 +51,7 @@
 #include <linux/mm_inline.h>
 #endif
 #include <linux/slab.h>
-#ifndef NO_MORE_DEV_FS
 #include <linux/devfs_fs_kernel.h>
-#endif
 
 #define __KERNEL_SYSCALLS__
 #include <linux/unistd.h>
@@ -76,15 +76,7 @@ extern asmlinkage int sys_ioctl(unsigned int fd, unsigned int cmd, unsigned long
 # endif
 #else
 # ifdef CONFIG_COMPAT
-#  if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
-    /* FIXME on which thing could we test instead of the KERNEL_VERSION
-     * again?  register_ioctl32_conversion was deprecated in 2.6.10, got
-     * "officially" deprecated somewhen in 2.6.12, and removed in 2.6.14.
-     * so lets assume all vendor kernels did the transition.  */
-#    define HAVE_COMPAT_IOCTL_MEMBER
-#  else
-#   include <linux/ioctl32.h>
-#  endif
+#  include <linux/ioctl32.h>
 # endif
 #endif
 
@@ -111,12 +103,13 @@ MODULE_LICENSE("GPL");
 MODULE_PARM_DESC(use_nbd_major, "DEPRECATED! use nbd device major nr (43) "
 		                "instead of the default " __stringify(LANANA_DRBD_MAJOR) );
 MODULE_PARM_DESC(minor_count, "Maximum number of drbd devices (1-255)");
+MODULE_PARM_DESC(disable_io_hints, "Necessary if the loopback network device is used for DRBD" );
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 MODULE_PARM(use_nbd_major,"i");
 MODULE_PARM(minor_count,"i");
+MODULE_PARM(disable_io_hints,"i");
 #else
 #include <linux/moduleparam.h>
-MODULE_PARM_DESC(disable_bd_claim, "DONT USE! disables block device claiming" );
 /*
  * please somebody explain to me what the "perm" of the module_param
  * macro is good for (yes, permission for it in the "driverfs", but what
@@ -133,7 +126,7 @@ MODULE_PARM_DESC(disable_bd_claim, "DONT USE! disables block device claiming" );
  */
 module_param(use_nbd_major,   bool,0);
 module_param(minor_count,      int,0);
-module_param(disable_bd_claim,bool,0);
+module_param(disable_io_hints,bool,0);
 #endif
 
 // module parameter, defined
@@ -144,12 +137,12 @@ int minor_count = 2;
 #else
 int minor_count = 8;
 #endif
-int disable_bd_claim = 0;
+// FIXME disable_io_hints shall die
+int disable_io_hints = 0;
 
-#ifndef NO_MORE_DEV_FS
 // devfs name
 char* drbd_devfs_name = "drbd";
-#endif
+
 
 // global panic flag
 volatile int drbd_did_panic = 0;
@@ -162,8 +155,8 @@ STATIC int *drbd_blocksizes;
 STATIC int *drbd_sizes;
 )
 struct Drbd_Conf *drbd_conf;
-drbd_kmem_cache_t *drbd_request_cache;
-drbd_kmem_cache_t *drbd_ee_cache;
+kmem_cache_t *drbd_request_cache;
+kmem_cache_t *drbd_ee_cache;
 mempool_t *drbd_request_mempool;
 
 STATIC struct block_device_operations drbd_ops = {
@@ -172,10 +165,7 @@ STATIC struct block_device_operations drbd_ops = {
 #endif
 	.open =    drbd_open,
 	.release = drbd_close,
-	.ioctl =   drbd_ioctl,
-#ifdef HAVE_COMPAT_IOCTL_MEMBER
-	.compat_ioctl = drbd_compat_ioctl,
-#endif
+	.ioctl =   drbd_ioctl
 };
 
 #define ARRY_SIZE(A) (sizeof(A)/sizeof(A[0]))
@@ -208,19 +198,34 @@ STATIC void tl_add(drbd_dev *mdev, drbd_request_t * new_item)
 {
 	struct drbd_barrier *b;
 
-	spin_lock_irq(&mdev->req_lock);
+	spin_lock_irq(&mdev->tl_lock);
 
 	b=mdev->newest_barrier;
 
 	new_item->barrier = b;
 	new_item->rq_status |= RQ_DRBD_IN_TL;
-	list_add_tail(&new_item->w.list,&b->requests);
+	list_add(&new_item->w.list,&b->requests);
 
 	if( b->n_req++ > mdev->conf.max_epoch_size ) {
 		set_bit(ISSUE_BARRIER,&mdev->flags);
 	}
 
-	spin_unlock_irq(&mdev->req_lock);
+	spin_unlock_irq(&mdev->tl_lock);
+}
+
+STATIC void tl_cancel(drbd_dev *mdev, drbd_request_t * item)
+{
+	struct drbd_barrier *b;
+
+	spin_lock_irq(&mdev->tl_lock);
+
+	b=item->barrier;
+	b->n_req--;
+
+	list_del(&item->w.list);
+	item->rq_status &= ~RQ_DRBD_IN_TL;
+
+	spin_unlock_irq(&mdev->tl_lock);
 }
 
 STATIC unsigned int tl_add_barrier(drbd_dev *mdev)
@@ -231,6 +236,8 @@ STATIC unsigned int tl_add_barrier(drbd_dev *mdev)
 
 	barrier_nr_issue++;
 
+	// THINK this is called in the IO path with the send_mutex held
+	// and GFP_KERNEL may itself start IO. set it to GFP_NOIO.
 	b=kmalloc(sizeof(struct drbd_barrier),GFP_NOIO);
 	if(!b) {
 		ERR("could not kmalloc() barrier\n");
@@ -241,13 +248,13 @@ STATIC unsigned int tl_add_barrier(drbd_dev *mdev)
 	b->br_number=barrier_nr_issue;
 	b->n_req=0;
 
-	spin_lock_irq(&mdev->req_lock);
+	spin_lock_irq(&mdev->tl_lock);
 
 	bnr = mdev->newest_barrier->br_number;
 	mdev->newest_barrier->next = b;
 	mdev->newest_barrier = b;
 
-	spin_unlock_irq(&mdev->req_lock);
+	spin_unlock_irq(&mdev->tl_lock);
 
 	return bnr;
 }
@@ -256,75 +263,48 @@ void tl_release(drbd_dev *mdev,unsigned int barrier_nr,
 		       unsigned int set_size)
 {
 	struct drbd_barrier *b;
-	struct list_head *le,*tle;
-	drbd_request_t *req;
 
-	spin_lock_irq(&mdev->req_lock);
+	spin_lock_irq(&mdev->tl_lock);
 
 	b = mdev->oldest_barrier;
 	mdev->oldest_barrier = b->next;
 
-	list_for_each_safe(le,tle,&b->requests) {
-		req = list_entry(le, struct drbd_request,w.list);
-		/* master_bio already completed, but protocol != C,
-		 * so we had it still in the tl. need to list_del
-		 * anyways, there may be local io pending! */
-		list_del(&req->w.list);
-		req->rq_status &= ~RQ_DRBD_IN_TL;
-		if( (req->rq_status & RQ_DRBD_DONE) == RQ_DRBD_DONE ) {
-			D_ASSERT(req->master_bio == NULL);
-			INVALIDATE_MAGIC(req);
-			mempool_free(req,drbd_request_mempool);
-		}
-	}
-	spin_unlock_irq(&mdev->req_lock);
+	list_del(&b->requests);
+	/* There could be requests on the list waiting for completion
+	   of the write to the local disk, to avoid corruptions of
+	   slab's data structures we have to remove the lists head */
+
+	spin_unlock_irq(&mdev->tl_lock);
 
 	D_ASSERT(b->br_number == barrier_nr);
 	D_ASSERT(b->n_req == set_size);
 
-#if 1
-        if(b->br_number != barrier_nr) {
-                DUMPI(b->br_number);
-                DUMPI(barrier_nr);
-        }
-        if(b->n_req != set_size) {
-                DUMPI(b->n_req);
-                DUMPI(set_size);
-        }
-#endif
-
-	list_del(&b->requests);
 	kfree(b);
 }
 
-/* Since IO is frozen, nobody may modifiy the Transfer right log now...
- */
-void tl_resend(drbd_dev *mdev)
+/* tl_dependence reports if this sector was present in the current
+   epoch.
+   As side effect it clears also the pointer to the request if it
+   was present in the transfert log. (Since tl_dependence indicates
+   that IO is complete and that drbd_end_req() should not be called
+   in case tl_clear has to be called due to interruption of the
+   communication)
+*/
+/* bool */
+int tl_dependence(drbd_dev *mdev, drbd_request_t * item)
 {
-	struct drbd_request *req;
-	struct list_head *le;
-	struct drbd_barrier *b;
-	int n_req;
+	unsigned long flags;
+	int r=TRUE;
 
-	D_ASSERT(test_bit(IO_FROZEN,&mdev->flags));
+	spin_lock_irqsave(&mdev->tl_lock,flags);
 
-	b = mdev->oldest_barrier;
-	while(1) {
-		n_req=0;
-		list_for_each(le, &b->requests) {
-			req = list_entry(le, struct drbd_request,w.list);
-			drbd_resend_dblock(mdev,req);
-			n_req++;
-		}
-		b->n_req = n_req; // only necessary for the oldest...
-		if( b == mdev->newest_barrier ) break;
-		drbd_resend_barrier(mdev,b);
-		b = b->next;
-	}
-	D_ASSERT(test_bit(IO_FROZEN,&mdev->flags));
+	r = ( item->barrier == mdev->newest_barrier );
+	list_del(&item->w.list);
+
+	spin_unlock_irqrestore(&mdev->tl_lock,flags);
+	return r;
 }
 
-
 void tl_clear(drbd_dev *mdev)
 {
 	struct list_head *le,*tle;
@@ -333,44 +313,42 @@ void tl_clear(drbd_dev *mdev)
 	sector_t sector;
 	unsigned int size;
 
-	new_first=kmalloc(sizeof(struct drbd_barrier),GFP_NOIO);
+	new_first=kmalloc(sizeof(struct drbd_barrier),GFP_KERNEL);
 	if(!new_first) {
 		ERR("could not kmalloc() barrier\n");
 	}
 
-	/* FIXME if indeed we could not kmalloc, this will Oops!
-	 * can we somehow just recycle one of the existing barriers?
-	 */
 	INIT_LIST_HEAD(&new_first->requests);
 	new_first->next=0;
 	new_first->br_number=4711;
 	new_first->n_req=0;
 
-	spin_lock_irq(&mdev->req_lock);
+	spin_lock_irq(&mdev->tl_lock);
 
 	b=mdev->oldest_barrier;
 	mdev->oldest_barrier = new_first;
 	mdev->newest_barrier = new_first;
 
+	spin_unlock_irq(&mdev->tl_lock);
+
 	inc_ap_pending(mdev); // Since we count the old first as well...
+
 	while ( b ) {
 		list_for_each_safe(le, tle, &b->requests) {
 			r = list_entry(le, struct drbd_request,w.list);
-			// r may be freed within drbd_end_req
-			sector = r->sector;
-			size   = r->size;
-
-			r->rq_status &= ~RQ_DRBD_IN_TL;
-			list_del(&r->w.list);
-
+			// bi_size and bi_sector are modified in bio_endio!
+			sector = drbd_req_get_sector(r);
+			size   = drbd_req_get_size(r);
 			if( !(r->rq_status & RQ_DRBD_SENT) ) {
-				_drbd_end_req(r,RQ_DRBD_SENT,1);
-			} else if ((r->rq_status & RQ_DRBD_DONE) == RQ_DRBD_DONE) {
-				D_ASSERT(r->master_bio == NULL);
-				INVALIDATE_MAGIC(r);
-				mempool_free(r,drbd_request_mempool);
-			} /* else: local io still pending */
-			drbd_set_out_of_sync(mdev, sector, size);
+				if(mdev->conf.wire_protocol != DRBD_PROT_A )
+					dec_ap_pending(mdev);
+				drbd_end_req(r,RQ_DRBD_SENT,ERF_NOTLD|1, sector);
+				goto mark;
+			}
+			if(mdev->conf.wire_protocol != DRBD_PROT_C ) {
+			mark:
+				drbd_set_out_of_sync(mdev, sector, size);
+			}
 		}
 		f=b;
 		b=b->next;
@@ -378,7 +356,6 @@ void tl_clear(drbd_dev *mdev)
 		kfree(f);
 		dec_ap_pending(mdev); // for the barrier
 	}
-	spin_unlock_irq(&mdev->req_lock);
 }
 
 /**
@@ -484,6 +461,10 @@ void _set_cstate(drbd_dev* mdev,Drbd_CState ns)
 	smp_mb();
 	wake_up(&mdev->cstate_wait);
 
+	/* THINK.
+	 * was:
+	 * if ( ( os==SyncSource || os==SyncTarget ) && ns <= Connected ) {
+	 */
 	if ( ( os >= SyncSource ) && ns <= Connected ) {
 		clear_bit(SYNC_STARTED,&mdev->flags);
 		set_bit(STOP_SYNC_TIMER,&mdev->flags);
@@ -494,19 +475,6 @@ void _set_cstate(drbd_dev* mdev,Drbd_CState ns)
 // FIXME EXPLAIN
 		clear_bit(MD_IO_ALLOWED,&mdev->flags);
 	}
-
-	if(mdev->conf.on_disconnect == FreezeIO && mdev->state == Primary) {
-		if(os >= Connected && ns < Connected) {
-			set_bit(IO_FROZEN, &mdev->flags);
-		}
-	}
-
-	if( ns <= StandAlone && test_bit(IO_FROZEN, &mdev->flags)) {
-		WARN("Going to thaw IO, setting out of sync %d requests.\n",
-		     atomic_read(&mdev->ap_pending_cnt));
-		tl_clear(mdev);
-		clear_bit(IO_FROZEN, &mdev->flags);
-	}
 }
 
 STATIC int drbd_thread_setup(void* arg)
@@ -518,10 +486,8 @@ STATIC int drbd_thread_setup(void* arg)
 	drbd_daemonize();
 	D_ASSERT(get_t_state(thi) == Running);
 	D_ASSERT(thi->task == NULL);
-	spin_lock(&thi->t_lock);
 	thi->task = current;
 	smp_mb();
-	spin_unlock(&thi->t_lock);
 	complete(&thi->startstop); // notify: thi->task is set.
 
 	retval = thi->function(thi);
@@ -544,6 +510,7 @@ STATIC void drbd_thread_init(drbd_dev *mdev, struct Drbd_thread *thi,
 	thi->t_lock  = SPIN_LOCK_UNLOCKED;
 	thi->task    = NULL;
 	thi->t_state = None;
+	init_completion(&thi->startstop);
 
 	thi->function = func;
 	thi->mdev = mdev;
@@ -564,11 +531,10 @@ void drbd_thread_start(struct Drbd_thread *thi)
 	     thi->t_state); */
 
 	if (thi->t_state == None) {
-		init_completion(&thi->startstop);
 		D_ASSERT(thi->task == NULL);
 		thi->t_state = Running;
 		spin_unlock(&thi->t_lock);
-		drbd_flush_signals(current); // otherw. may get -ERESTARTNOINTR
+
 		pid = kernel_thread(drbd_thread_setup, (void *) thi, CLONE_FS);
 		if (pid < 0) {
 			ERR("Couldn't start thread (%d)\n", pid);
@@ -590,15 +556,10 @@ void _drbd_thread_stop(struct Drbd_thread *thi, int restart,int wait)
 
 	spin_lock(&thi->t_lock);
 
-	/*
-	INFO("%s [%d]: %s[%s] %d -> %d; wait:%d; x->done:%d\n",
+	/* INFO("%s [%d]: %s %d -> %d; %d\n",
 	     current->comm, current->pid,
-	     thi == &mdev->worker ? "worker" :
-	     thi == &mdev->asender ? "asender" :
-	     thi == &mdev->receiver ? "receiver" : "???",
-	     thi->task ? thi->task->comm : "NULL", thi->t_state, ns, wait,
-	     thi->startstop.done);
-	*/
+	     thi->task ? thi->task->comm : "NULL", thi->t_state, ns, wait); */
+
 
 	if (thi->t_state == None) {
 		spin_unlock(&thi->t_lock);
@@ -619,10 +580,10 @@ void _drbd_thread_stop(struct Drbd_thread *thi, int restart,int wait)
 
 		thi->t_state = ns;
 		smp_mb();
-		if (thi->task != current) {
-			if(wait) init_completion(&thi->startstop);
+		if (thi->task != current)
 			force_sig(DRBD_SIGKILL,thi->task);
-		} else D_ASSERT(!wait);
+		else
+			D_ASSERT(!wait);
 
 	}
 	spin_unlock(&thi->t_lock);
@@ -661,7 +622,6 @@ inline void restore_old_sigset(sigset_t oldset)
 	UNLOCK_SIGMASK(current,flags);
 }
 
-/* the appropriate socket mutex must be held already */
 int _drbd_send_cmd(drbd_dev *mdev, struct socket *sock,
 			  Drbd_Packet_Cmd cmd, Drbd_Header *h,
 			  size_t size, unsigned msg_flags)
@@ -686,36 +646,25 @@ int _drbd_send_cmd(drbd_dev *mdev, struct socket *sock,
 	return ok;
 }
 
-/* don't pass the socket. we may only look at it
- * when we hold the appropriate socket mutex.
- */
-int drbd_send_cmd(drbd_dev *mdev, int use_data_socket,
+int drbd_send_cmd(drbd_dev *mdev, struct socket *sock,
 		  Drbd_Packet_Cmd cmd, Drbd_Header* h, size_t size)
 {
-	int ok = 0;
+	int ok;
 	sigset_t old_blocked;
-	struct socket *sock;
 
-	if (use_data_socket) {
+	if (sock == mdev->data.socket) {
 		down(&mdev->data.mutex);
 		spin_lock(&mdev->send_task_lock);
 		mdev->send_task=current;
 		spin_unlock(&mdev->send_task_lock);
-		sock = mdev->data.socket;
-	} else {
+	} else
 		down(&mdev->meta.mutex);
-		sock = mdev->meta.socket;
-	}
 
-	/* drbd_disconnect() could have called drbd_free_sock()
-	 * while we were waiting in down()... */
-	if (likely(sock != NULL)) {
-		old_blocked = drbd_block_all_signals();
-		ok = _drbd_send_cmd(mdev, sock, cmd, h, size, 0);
-		restore_old_sigset(old_blocked);
-	}
+	old_blocked = drbd_block_all_signals();
+	ok = _drbd_send_cmd(mdev,sock,cmd,h,size,0);
+	restore_old_sigset(old_blocked);
 
-	if (use_data_socket) {
+	if (sock == mdev->data.socket) {
 		spin_lock(&mdev->send_task_lock);
 		mdev->send_task=NULL;
 		spin_unlock(&mdev->send_task_lock);
@@ -735,7 +684,7 @@ int drbd_send_sync_param(drbd_dev *mdev, struct syncer_config *sc)
 	p.skip      = cpu_to_be32(sc->skip);
 	p.group     = cpu_to_be32(sc->group);
 
-	ok = drbd_send_cmd(mdev,USE_DATA_SOCKET,SyncParam,(Drbd_Header*)&p,sizeof(p));
+	ok = drbd_send_cmd(mdev,mdev->data.socket,SyncParam,(Drbd_Header*)&p,sizeof(p));
 	if ( ok
 	    && (mdev->cstate == SkippedSyncS || mdev->cstate == SkippedSyncT)
 	    && !sc->skip )
@@ -777,7 +726,7 @@ int drbd_send_param(drbd_dev *mdev, int flags)
 	p.flags          = cpu_to_be32(flags);
 	p.magic          = BE_DRBD_MAGIC;
 
-	ok = drbd_send_cmd(mdev,USE_DATA_SOCKET,ReportParams,(Drbd_Header*)&p,sizeof(p));
+	ok = drbd_send_cmd(mdev,mdev->data.socket,ReportParams,(Drbd_Header*)&p,sizeof(p));
 	if (have_disk) dec_local(mdev);
 	return ok;
 }
@@ -846,23 +795,14 @@ int _drbd_send_barrier(drbd_dev *mdev)
 	int ok;
 	Drbd_Barrier_Packet p;
 
+	/* printk(KERN_DEBUG DEVICE_NAME": issuing a barrier\n"); */
 	/* tl_add_barrier() must be called with the sock_mutex aquired */
 	p.barrier=tl_add_barrier(mdev);
 
 	inc_ap_pending(mdev);
 	ok = _drbd_send_cmd(mdev,mdev->data.socket,Barrier,(Drbd_Header*)&p,sizeof(p),0);
 
-	return ok;
-}
-
-int drbd_resend_barrier(drbd_dev *mdev,struct drbd_barrier *b)
-{
-	int ok;
-	Drbd_Barrier_Packet p;
-
-	p.barrier=b->br_number;
-	ok=drbd_send_cmd(mdev,USE_DATA_SOCKET, Barrier,(Drbd_Header*)&p,sizeof(p));
-
+//	if (!ok) dec_ap_pending(mdev); // is done in tl_clear()
 	return ok;
 }
 
@@ -874,7 +814,7 @@ int drbd_send_b_ack(drbd_dev *mdev, u32 barrier_nr,u32 set_size)
 	p.barrier  = barrier_nr;
 	p.set_size = cpu_to_be32(set_size);
 
-	ok = drbd_send_cmd(mdev,USE_META_SOCKET,BarrierAck,(Drbd_Header*)&p,sizeof(p));
+	ok = drbd_send_cmd(mdev,mdev->meta.socket,BarrierAck,(Drbd_Header*)&p,sizeof(p));
 	return ok;
 }
 
@@ -889,7 +829,7 @@ int drbd_send_ack(drbd_dev *mdev, Drbd_Packet_Cmd cmd, struct Tl_epoch_entry *e)
 	p.blksize  = cpu_to_be32(drbd_ee_get_size(e));
 
 	if (!mdev->meta.socket || mdev->cstate < Connected) return FALSE;
-	ok = drbd_send_cmd(mdev,USE_META_SOCKET,cmd,(Drbd_Header*)&p,sizeof(p));
+	ok = drbd_send_cmd(mdev,mdev->meta.socket,cmd,(Drbd_Header*)&p,sizeof(p));
 	return ok;
 }
 
@@ -903,7 +843,7 @@ int drbd_send_drequest(drbd_dev *mdev, int cmd,
 	p.block_id = block_id;
 	p.blksize  = cpu_to_be32(size);
 
-	ok = drbd_send_cmd(mdev,USE_DATA_SOCKET,cmd,(Drbd_Header*)&p,sizeof(p));
+	ok = drbd_send_cmd(mdev,mdev->data.socket,cmd,(Drbd_Header*)&p,sizeof(p));
 	return ok;
 }
 
@@ -1081,9 +1021,10 @@ int drbd_send_dblock(drbd_dev *mdev, drbd_request_t *req)
 
 	p.head.magic   = BE_DRBD_MAGIC;
 	p.head.command = cpu_to_be16(Data);
-	p.head.length  = cpu_to_be16(sizeof(p)-sizeof(Drbd_Header)+req->size);
+	p.head.length  = cpu_to_be16( sizeof(p)-sizeof(Drbd_Header)
+				     + drbd_req_get_size(req) );
 
-	p.sector   = cpu_to_be64(req->sector);
+	p.sector   = cpu_to_be64(drbd_req_get_sector(req));
 	p.block_id = (unsigned long)req;
 
 	/* About tl_add():
@@ -1095,7 +1036,8 @@ int drbd_send_dblock(drbd_dev *mdev, drbd_request_t *req)
 	   tl_ datastructure (=> We would want to remove it before it
 	   is there!)
 	3. Q: Why can we add it to tl_ even when drbd_send() might fail ?
-	   A: If drbd_send fails, we will lose the connection. Then
+	      There could be a tl_cancel() to remove it within the semaphore!
+	   A: If drbd_send fails, we will loose the connection. Then
 	      tl_cear() will simulate a RQ_DRBD_SEND and set it out of sync
 	      for everything in the data structure.
 	*/
@@ -1109,107 +1051,42 @@ int drbd_send_dblock(drbd_dev *mdev, drbd_request_t *req)
 
 	old_blocked = drbd_block_all_signals();
 	down(&mdev->data.mutex);
-
-	/* drbd_disconnect may have freed that socket while we were waiting
-	 * in down(). we have to check that, to avoid a race with tl_clear
-	 * cleaning up before we can tl_add */
-	if (unlikely(!mdev->data.socket)) {
-                /* this req is not in the tl, tl_clear cannot find it.
-                 *
-                 * For Freeze-IO,
-                 * we add the req to the tl anyways, possibly creating
-                 * a new epoch, and then do nothing.  It will then still be
-                 * submitted locally, but it cannot possibly signal
-                 * completion to the upper layers before it was
-                 * resent (tl_resend) or canceled (tl_clear).
-                 *
-                 * Otherwise,
-                 * we cannot just tl_add it here, either, because tl_clear
-                 * might be done already.  so we have to mark this request
-                 * "SENT" here, otherwise it won't ever complete.
-                 * FIXME if we are Diskless, we complete a WRITE
-                 * as successful here, that has never been written! */
-		if (test_bit(IO_FROZEN,&mdev->flags)) {
-			if(test_and_clear_bit(ISSUE_BARRIER,&mdev->flags)) {
-				inc_ap_pending(mdev);
-				tl_add_barrier(mdev);
-			}
-			tl_add(mdev,req);
-		} else {
-			drbd_set_out_of_sync(mdev, req->sector, req->size);
-			drbd_end_req(req,RQ_DRBD_SENT,1);
-		}
-		goto out;
-	}
-
 	spin_lock(&mdev->send_task_lock);
 	mdev->send_task=current;
 	spin_unlock(&mdev->send_task_lock);
 
 	if(test_and_clear_bit(ISSUE_BARRIER,&mdev->flags))
 		ok = _drbd_send_barrier(mdev);
-
-	/* unconditional tl_add! */
-	tl_add(mdev,req);
 	if(ok) {
+		tl_add(mdev,req);
 		dump_packet(mdev,mdev->data.socket,0,(void*)&p, __FILE__, __LINE__);
 		set_bit(UNPLUG_REMOTE,&mdev->flags);
 		ok = sizeof(p) == drbd_send(mdev,mdev->data.socket,&p,sizeof(p),MSG_MORE);
 		if(ok) {
 			if(mdev->conf.wire_protocol == DRBD_PROT_A) {
-				ok = _drbd_send_bio(mdev,drbd_req_private_bio(req));
-				/* drbd_end_req has to be within the data.mutex,
-				 * which protects us from a concurrently running tl_clear */
-				if (ok) drbd_end_req(req,RQ_DRBD_SENT,1);
+				ok = _drbd_send_bio(mdev,&req->private_bio);
 			} else {
-				ok = _drbd_send_zc_bio(mdev,drbd_req_private_bio(req));
+				ok = _drbd_send_zc_bio(mdev,&req->private_bio);
 			}
 		}
+		if(!ok) tl_cancel(mdev,req);
+	}
+	if (!ok) {
+		drbd_set_out_of_sync(mdev,
+				     drbd_req_get_sector(req),
+				     drbd_req_get_size(req));
+		drbd_end_req(req,RQ_DRBD_SENT,ERF_NOTLD|1,
+			     drbd_req_get_sector(req));
 	}
-	/* if (!ok) ... cleanup done in tl_clear */
 	spin_lock(&mdev->send_task_lock);
 	mdev->send_task=NULL;
 	spin_unlock(&mdev->send_task_lock);
 
-  out:
 	up(&mdev->data.mutex);
 	restore_old_sigset(old_blocked);
 	return ok;
 }
 
-int drbd_resend_dblock(drbd_dev *mdev, drbd_request_t *req)
-{
-	int ok;
-	Drbd_Data_Packet p;
-
-	p.head.magic   = BE_DRBD_MAGIC;
-	p.head.command = cpu_to_be16(Data);
-	p.head.length  = cpu_to_be16(sizeof(p)-sizeof(Drbd_Header)+req->size);
-
-	p.sector   = cpu_to_be64(req->sector);
-	p.block_id = (unsigned long)req;
-
-	down(&mdev->data.mutex);
-
-	spin_lock(&mdev->send_task_lock);
-	mdev->send_task=current;
-	spin_unlock(&mdev->send_task_lock);
-
-	dump_packet(mdev,mdev->data.socket,0,(void*)&p, __FILE__, __LINE__);
-	ok = sizeof(p) == drbd_send(mdev,mdev->data.socket,&p,sizeof(p),MSG_MORE);
-	if(ok) {
-		ok = _drbd_send_bio(mdev,req->master_bio);
-	}
-
-	spin_lock(&mdev->send_task_lock);
-	mdev->send_task=NULL;
-	spin_unlock(&mdev->send_task_lock);
-
-	up(&mdev->data.mutex);
-	return ok;
-}
-
-
 int drbd_send_block(drbd_dev *mdev, Drbd_Packet_Cmd cmd,
 		    struct Tl_epoch_entry *e)
 {
@@ -1261,7 +1138,7 @@ int drbd_send_block(drbd_dev *mdev, Drbd_Packet_Cmd cmd,
 */
 
 /*
- * you must have down()ed the appropriate [m]sock_mutex elsewhere!
+ * you should have down()ed the appropriate [m]sock_mutex elsewhere!
  */
 int drbd_send(drbd_dev *mdev, struct socket *sock,
 	      void* buf, size_t size, unsigned msg_flags)
@@ -1448,6 +1325,10 @@ void drbd_set_defaults(drbd_dev *mdev)
 {
 	mdev->flags = 1<<DISKLESS;
 
+	/* If the UNPLUG_QUEUED flag is set but it is not
+	   actually queued the functionality is completely disabled */
+	if (disable_io_hints) mdev->flags |= 1<<UNPLUG_QUEUED;
+
 	mdev->sync_conf.rate       = 250;
 	mdev->sync_conf.al_extents = 127; // 512 MB active set
 	mdev->state                = Secondary;
@@ -1480,6 +1361,7 @@ void drbd_init_set_defaults(drbd_dev *mdev)
 	sema_init(&mdev->meta.work.s,0);
 
 	mdev->al_lock        = SPIN_LOCK_UNLOCKED;
+	mdev->tl_lock        = SPIN_LOCK_UNLOCKED;
 	mdev->ee_lock        = SPIN_LOCK_UNLOCKED;
 	mdev->req_lock       = SPIN_LOCK_UNLOCKED;
 	mdev->pr_lock        = SPIN_LOCK_UNLOCKED;
@@ -1491,6 +1373,7 @@ void drbd_init_set_defaults(drbd_dev *mdev)
 	INIT_LIST_HEAD(&mdev->done_ee);
 	INIT_LIST_HEAD(&mdev->read_ee);
 	INIT_LIST_HEAD(&mdev->net_ee);
+	INIT_LIST_HEAD(&mdev->busy_blocks);
 	INIT_LIST_HEAD(&mdev->app_reads);
 	INIT_LIST_HEAD(&mdev->resync_reads);
 	INIT_LIST_HEAD(&mdev->data.work.q);
@@ -1569,9 +1452,9 @@ void drbd_mdev_cleanup(drbd_dev *mdev)
 
 	if (   mdev->ee_in_use  !=  0
 	    || mdev->ee_vacant  != 32 /* EE_MININUM */
-	    || atomic_read(&mdev->epoch_size) !=  0)
+	    || mdev->epoch_size !=  0)
 		ERR("ee_in_use:%d ee_vacant:%d epoch_size:%d\n",
-		    mdev->ee_in_use, mdev->ee_vacant, atomic_read(&mdev->epoch_size));
+		    mdev->ee_in_use, mdev->ee_vacant, mdev->epoch_size);
 #define ZAP(x) memset(&x,0,sizeof(x))
 	ZAP(mdev->conf);
 	ZAP(mdev->sync_conf);
@@ -1594,7 +1477,6 @@ void drbd_mdev_cleanup(drbd_dev *mdev)
 	mdev->rs_mark_time = 0;
 	mdev->send_task    = NULL;
 	drbd_set_my_capacity(mdev,0);
-	drbd_bm_resize(mdev,0);
 
 	// just in case
 	drbd_free_resources(mdev);
@@ -1611,6 +1493,7 @@ void drbd_mdev_cleanup(drbd_dev *mdev)
 	D_ASSERT(list_empty(&mdev->done_ee));
 	D_ASSERT(list_empty(&mdev->read_ee));
 	D_ASSERT(list_empty(&mdev->net_ee));
+	D_ASSERT(list_empty(&mdev->busy_blocks));
 	D_ASSERT(list_empty(&mdev->app_reads));
 	D_ASSERT(list_empty(&mdev->resync_reads));
 	D_ASSERT(list_empty(&mdev->data.work.q));
@@ -1627,8 +1510,13 @@ void drbd_destroy_mempools(void)
 {
 	if (drbd_request_mempool)
 		mempool_destroy(drbd_request_mempool);
-	if (drbd_ee_cache) kmem_cache_destroy(drbd_ee_cache);
-	if (drbd_request_cache) kmem_cache_destroy(drbd_request_cache);
+	if (drbd_ee_cache && kmem_cache_destroy(drbd_ee_cache))
+		printk(KERN_ERR DEVICE_NAME
+		       ": kmem_cache_destroy(drbd_ee_cache) FAILED\n");
+	if (drbd_request_cache && kmem_cache_destroy(drbd_request_cache))
+		printk(KERN_ERR DEVICE_NAME
+		       ": kmem_cache_destroy(drbd_request_cache) FAILED\n");
+	// FIXME what can we do if we fail to destroy them?
 
 	drbd_request_mempool = NULL;
 	drbd_ee_cache        = NULL;
@@ -1647,13 +1535,13 @@ int drbd_create_mempools(void)
 	// caches
 	drbd_request_cache = kmem_cache_create(
 		"drbd_req_cache", sizeof(drbd_request_t),
-		0, 0, NULL, NULL);
+		0, SLAB_NO_REAP, NULL, NULL);
 	if (drbd_request_cache == NULL)
 		goto Enomem;
 
 	drbd_ee_cache = kmem_cache_create(
 		"drbd_ee_cache", sizeof(struct Tl_epoch_entry),
-		0, 0, NULL, NULL);
+		0, SLAB_NO_REAP, NULL, NULL);
 	if (drbd_ee_cache == NULL)
 		goto Enomem;
 
@@ -1761,7 +1649,6 @@ ONLY_IN_26(
 		drbd_destroy_mempools();
 	}
 
-#ifndef HAVE_COMPAT_IOCTL_MEMBER
 #if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
 	lock_kernel();
 	unregister_ioctl32_conversion(DRBD_IOCTL_GET_VERSION);
@@ -1777,10 +1664,8 @@ ONLY_IN_26(
 	unregister_ioctl32_conversion(DRBD_IOCTL_WAIT_CONNECT);
 	unregister_ioctl32_conversion(DRBD_IOCTL_WAIT_SYNC);
 	unregister_ioctl32_conversion(DRBD_IOCTL_UNCONFIG_DISK);
-	unregister_ioctl32_conversion(DRBD_IOCTL_SET_STATE_FLAGS);
 	unlock_kernel();
 #endif
-#endif
 
 NOT_IN_26(
 	blksize_size[MAJOR_NR] = NULL;
@@ -1794,9 +1679,7 @@ NOT_IN_26(
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	devfs_unregister(devfs_handle);
 #else
-#ifndef NO_MORE_DEV_FS
 	devfs_remove(drbd_devfs_name);
-#endif
 #endif
 
 	if (unregister_blkdev(MAJOR_NR, DEVICE_NAME) != 0)
@@ -1811,7 +1694,6 @@ int __init drbd_init(void)
 	int i,err;
 
 #if 0
-#warning "DEBUGGING"
 /* I am too lazy to calculate this by hand	-lge
  */
 #define SZO(x) printk(KERN_ERR "sizeof(" #x ") = %d\n", sizeof(x))
@@ -1845,6 +1727,9 @@ int __init drbd_init(void)
 		       ": never change the size or layout of the HandShake packet.\n");
 		return -EINVAL;
 	}
+	if (sizeof_drbd_structs_sanity_check()) {
+		return -EINVAL;
+	}
 
 	if (use_nbd_major) {
 		major_nr = NBD_MAJOR;
@@ -1870,9 +1755,7 @@ int __init drbd_init(void)
 		return err;
 	}
 
-#ifndef NO_MORE_DEV_FS
 	drbd_devfs_name = (major_nr == NBD_MAJOR) ? "nbd" : "drbd";
-#endif
 
 	/*
 	 * allocate all necessary structs
@@ -1894,9 +1777,7 @@ int __init drbd_init(void)
 	if (unlikely(!drbd_blocksizes)) goto Enomem;
 #else
 
-#ifndef NO_MORE_DEV_FS
 	devfs_mk_dir(drbd_devfs_name);
-#endif
 
 	for (i = 0; i < minor_count; i++) {
 		drbd_dev    *mdev = drbd_conf + i;
@@ -1919,9 +1800,7 @@ int __init drbd_init(void)
 		disk->first_minor = i;
 		disk->fops = &drbd_ops;
 		sprintf(disk->disk_name, DEVICE_NAME "%d", i);
-#ifndef NO_MORE_DEV_FS
 		sprintf(disk->devfs_name, "%s/%d", drbd_devfs_name, i);
-#endif
 		disk->private_data = mdev;
 		add_disk(disk);
 
@@ -1931,8 +1810,6 @@ int __init drbd_init(void)
 		if (bd_claim(mdev->this_bdev,drbd_sec_holder)) {
 			// Initial we are Secondary -> should claim myself.
 			WARN("Could not bd_claim() myself.");
-		} else if (disable_bd_claim) {
-			bd_release(mdev->this_bdev);
 		}
 
 		blk_queue_make_request(q,drbd_make_request_26);
@@ -1977,14 +1854,13 @@ NOT_IN_26(
 	/*
 	 * register with procfs
 	 */
-	drbd_proc = create_proc_entry("drbd",  S_IFREG | S_IRUGO , &proc_root);
-
+	// XXX maybe move to a seq_file interface
+	drbd_proc = create_proc_read_entry("drbd", 0, &proc_root,
+					   drbd_proc_get_info, NULL);
 	if (!drbd_proc)	{
 		printk(KERN_ERR DEVICE_NAME": unable to register proc file\n");
 		goto Enomem;
 	}
-	
-	drbd_proc->proc_fops = &drbd_proc_fops;
 	drbd_proc->owner = THIS_MODULE;
 #else
 # error "Currently drbd depends on the proc file system (CONFIG_PROC_FS)"
@@ -2004,7 +1880,6 @@ NOT_IN_26(
 
 	NOT_IN_26(blk_queue_make_request(BLK_DEFAULT_QUEUE(MAJOR_NR),drbd_make_request_24);)
 
-#ifndef HAVE_COMPAT_IOCTL_MEMBER
 #if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
 	// tell the kernel that we think our ioctls are 64bit clean
 	lock_kernel();
@@ -2021,9 +1896,7 @@ NOT_IN_26(
 	register_ioctl32_conversion(DRBD_IOCTL_WAIT_CONNECT,NULL);
 	register_ioctl32_conversion(DRBD_IOCTL_WAIT_SYNC,NULL);
 	register_ioctl32_conversion(DRBD_IOCTL_UNCONFIG_DISK,NULL);
-	register_ioctl32_conversion(DRBD_IOCTL_SET_STATE_FLAGS,NULL);
 	unlock_kernel();
-#endif
 #endif
 
 	printk(KERN_INFO DEVICE_NAME ": initialised. "
@@ -2103,7 +1976,7 @@ struct meta_data_on_disk {
 	u32 al_offset;         // offset to this block
 	u32 al_nr_extents;     // important for restoring the AL
 	u32 bm_offset;         // offset to the bitmap, from here
-} __attribute((packed));
+};
 
 /*
 
@@ -2124,8 +1997,7 @@ void drbd_md_write(drbd_dev *mdev)
 	memset(buffer,0,512);
 
 	flags = mdev->gen_cnt[Flags] & ~(MDF_PrimaryInd|MDF_ConnectedInd);
-	if (mdev->state  == Primary || 
-	    test_bit(CRASHED_PRIMARY, &mdev->flags))   flags |= MDF_PrimaryInd;
+	if (mdev->state  == Primary)        flags |= MDF_PrimaryInd;
 	if (mdev->cstate >= WFReportParams) flags |= MDF_ConnectedInd;
 	mdev->gen_cnt[Flags] = flags;
 
