@@ -5,9 +5,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	main author.
+
+   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
+	main contributor.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,7 +27,7 @@
 
  */
 
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/module.h>
 
 #include <asm/uaccess.h>
@@ -33,46 +35,12 @@
 #include <linux/file.h>
 #include <linux/slab.h>
 #include <linux/proc_fs.h>
-#include <linux/seq_file.h>
 #include <linux/drbd.h>
 #include "drbd_int.h"
 
-STATIC int drbd_proc_open(struct inode *inode, struct file *file);
-STATIC int drbd_seq_show(struct seq_file *seq, void *v);
-
+int drbd_proc_get_info(char *, char **, off_t, int, int *, void *);
 
 struct proc_dir_entry *drbd_proc;
-struct file_operations drbd_proc_fops = {
-	.owner		= THIS_MODULE,
-	.open		= drbd_proc_open,
-	.read		= seq_read,
-	.llseek		= seq_lseek,
-	.release	= seq_release,
-};
-
-// We ommit single_open and single_release, since that is only available
-// after 2.4.23
-static void *single_start(struct seq_file *p, loff_t *pos)
-{
-	return NULL + (*pos == 0);
-}
-
-static void *single_next(struct seq_file *p, void *v, loff_t *pos)
-{
-	++*pos;
-	return NULL;
-}
-
-static void single_stop(struct seq_file *p, void *v)
-{
-}
-
-struct seq_operations drbd_proc_seq_ops = {
-	.start		= single_start,
-	.next		= single_next,
-	.stop		= single_stop,
-	.show		= drbd_seq_show,
-};
 
 /*lge
  * progress bars shamelessly adapted from driver/md/md.c
@@ -80,8 +48,9 @@ struct seq_operations drbd_proc_seq_ops = {
  *	[=====>..............] 33.5% (23456/123456)
  *	finish: 2:20:20 speed: 6,345 (6,456) K/sec
  */
-STATIC void drbd_syncer_progress(struct Drbd_Conf* mdev, struct seq_file *seq)
+STATIC int drbd_syncer_progress(struct Drbd_Conf* mdev,char *buf)
 {
+	int sz = 0;
 	unsigned long res , db, dt, dbdt, rt, rs_left;
 
 	/* the whole sector_div thingy was wrong (did overflow,
@@ -113,23 +82,23 @@ STATIC void drbd_syncer_progress(struct Drbd_Conf* mdev, struct seq_file *seq)
 	}
 	{
 		int i, y = res/50, x = 20-y;
-		seq_printf(seq, "\t[");
+		sz += sprintf(buf + sz, "\t[");
 		for (i = 1; i < x; i++)
-			seq_printf(seq, "=");
-		seq_printf(seq, ">");
+			sz += sprintf(buf + sz, "=");
+		sz += sprintf(buf + sz, ">");
 		for (i = 0; i < y; i++)
-			seq_printf(seq, ".");
-		seq_printf(seq, "] ");
+			sz += sprintf(buf + sz, ".");
+		sz += sprintf(buf + sz, "] ");
 	}
 	res = 1000L - res;
-	seq_printf(seq,"sync'ed:%3lu.%lu%% ", res / 10, res % 10);
+	sz+=sprintf(buf+sz,"sync'ed:%3lu.%lu%% ", res / 10, res % 10);
 	/* if more than 1 GB display in MB */
 	if (mdev->rs_total > 0x100000L) {
-		seq_printf(seq,"(%lu/%lu)M\n\t",
+		sz+=sprintf(buf+sz,"(%lu/%lu)M\n\t",
 			    (unsigned long) Bit2KB(rs_left) >> 10,
 			    (unsigned long) Bit2KB(mdev->rs_total) >> 10 );
 	} else {
-		seq_printf(seq,"(%lu/%lu)K\n\t",
+		sz+=sprintf(buf+sz,"(%lu/%lu)K\n\t",
 			    (unsigned long) Bit2KB(rs_left),
 			    (unsigned long) Bit2KB(mdev->rs_total) );
 	}
@@ -144,28 +113,20 @@ STATIC void drbd_syncer_progress(struct Drbd_Conf* mdev, struct seq_file *seq)
 	 * rt: remaining time
 	 */
 	dt = (jiffies - mdev->rs_mark_time) / HZ;
-
-	if (dt > 20) {
-		/* if we made no update to rs_mark_time for too long,
-		 * we are stalled. show that. */
-		seq_printf(seq, "stalled\n");
-		return;
-	}
-
 	if (!dt) dt++;
 	db = mdev->rs_mark_left - rs_left;
 	rt = (dt * (rs_left / (db/100+1)))/100; /* seconds */
 
-	seq_printf(seq, "finish: %lu:%02lu:%02lu",
+	sz += sprintf(buf + sz, "finish: %lu:%02lu:%02lu",
 		rt / 3600, (rt % 3600) / 60, rt % 60);
 
 	/* current speed average over (SYNC_MARKS * SYNC_MARK_STEP) jiffies */
 	dbdt = Bit2KB(db/dt);
 	if (dbdt > 1000)
-		seq_printf(seq, " speed: %ld,%03ld",
+		sz += sprintf(buf + sz, " speed: %ld,%03ld",
 			dbdt/1000,dbdt % 1000);
 	else
-		seq_printf(seq, " speed: %ld", dbdt);
+		sz += sprintf(buf + sz, " speed: %ld", dbdt);
 
 	/* mean speed since syncer started
 	 * we do account for PausedSync periods */
@@ -174,12 +135,14 @@ STATIC void drbd_syncer_progress(struct Drbd_Conf* mdev, struct seq_file *seq)
 	db = mdev->rs_total - rs_left;
 	dbdt = Bit2KB(db/dt);
 	if (dbdt > 1000)
-		seq_printf(seq, " (%ld,%03ld)",
+		sz += sprintf(buf + sz, " (%ld,%03ld)",
 			dbdt/1000,dbdt % 1000);
 	else
-		seq_printf(seq, " (%ld)", dbdt);
+		sz += sprintf(buf + sz, " (%ld)", dbdt);
 
-	seq_printf(seq," K/sec\n");
+	sz += sprintf(buf+sz," K/sec\n");
+
+	return sz;
 }
 
 const char* cstate_to_name(Drbd_CState s) {
@@ -220,14 +183,16 @@ const char* nodestate_to_name(Drbd_State s) {
 		              : state_names[s];
 }
 
-
-STATIC int drbd_seq_show(struct seq_file *seq, void *v)
+/* FIXME we should use snprintf, we only have guaranteed room for one page...
+ * we should eventually use seq_file for this */
+int drbd_proc_get_info(char *buf, char **start, off_t offset,
+		       int len, int *unused, void *data)
 {
-	int i;
+	int rlen, i;
 	const char *sn;
 
-	seq_printf(seq, "version: " REL_VERSION " (api:%d/proto:%d)\n%s\n",
-		    API_VERSION,PRO_VERSION, drbd_buildtag());
+	rlen = sprintf(buf, "version: " REL_VERSION " (api:%d/proto:%d)\n%s\n",
+		       API_VERSION,PRO_VERSION, drbd_buildtag());
 
 	/*
 	  cs .. connection state
@@ -251,10 +216,11 @@ STATIC int drbd_seq_show(struct seq_file *seq, void *v)
 				sn = "ServerForDLess";
 		}
 		if ( drbd_conf[i].cstate == Unconfigured )
-			seq_printf( seq, "%2d: cs:Unconfigured\n", i);
+			rlen += sprintf( buf + rlen,
+			   "%2d: cs:Unconfigured\n", i);
 		else
-			seq_printf( seq,
-			   "%2d: cs:%s st:%s/%s ld:%s %c\n"
+			rlen += sprintf( buf + rlen,
+			   "%2d: cs:%s st:%s/%s ld:%s\n"
 			   "    ns:%u nr:%u dw:%u dr:%u al:%u bm:%u "
 			   "lo:%d pe:%d ua:%d ap:%d\n",
 			   i, sn,
@@ -262,7 +228,7 @@ STATIC int drbd_seq_show(struct seq_file *seq, void *v)
 			   nodestate_to_name(drbd_conf[i].o_state),
 			   (drbd_conf[i].gen_cnt[Flags]
 			    & MDF_Consistent) ? "Consistent" : "Inconsistent",
-			   test_bit(IO_FROZEN, &drbd_conf[i].flags)? 'F' : ' ',
+			// FIXME partner consistent?
 			   drbd_conf[i].send_cnt/2,
 			   drbd_conf[i].recv_cnt/2,
 			   drbd_conf[i].writ_cnt/2,
@@ -278,15 +244,12 @@ STATIC int drbd_seq_show(struct seq_file *seq, void *v)
 
 		if ( drbd_conf[i].cstate == SyncSource ||
 		     drbd_conf[i].cstate == SyncTarget )
-			drbd_syncer_progress(drbd_conf+i,seq);
+			rlen += drbd_syncer_progress(drbd_conf+i,buf+rlen);
 	}
 
-	return 0;
-}
+	/* DEBUG & profile stuff end */
 
-STATIC int drbd_proc_open(struct inode *inode, struct file *file)
-{
-	return seq_open(file, &drbd_proc_seq_ops);
+	return rlen;
 }
 
 /* PROC FS stuff end */
