@@ -5,9 +5,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	main author.
+
+   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
+	main contributor.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,7 +27,7 @@
  */
 
 
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/module.h>
 
 #include <asm/uaccess.h>
@@ -36,7 +38,6 @@
 #include <linux/version.h>
 #include <linux/fs.h>
 #include <linux/file.h>
-#include <linux/in.h>
 #include <linux/mm.h>
 #include <linux/drbd_config.h>
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)) || defined(HAVE_MM_INLINE_H)
@@ -55,8 +56,8 @@
 
 #define is_syncer_blk(A,B) ((B)==ID_SYNCER)
 
-#if defined(__arch_um__) && !defined(HAVE_UML_TO_VIRT)
-static inline void *to_virt(unsigned long phys)
+#ifdef __arch_um__
+void *to_virt(unsigned long phys)
 {
 	return((void *) uml_physmem + phys);
 }
@@ -145,6 +146,7 @@ You need to hold the ee_lock:
  _drbd_process_ee()
 
 You must not have the ee_lock:
+ _drbd_alloc_ee()
  drbd_alloc_ee()
  drbd_init_ee()
  drbd_release_ee()
@@ -154,90 +156,80 @@ You must not have the ee_lock:
  drbd_wait_ee()
 */
 
-/*
- * gfp_mask is only used for the page allocation,
- * not for the kmem_cache_alloc, which is hardcoded to
- * basically GFP_NOIO, see below.
- *
- * THIS MUST NOT HOLD A LOCK
- */
-struct Tl_epoch_entry*
-drbd_try_alloc_ee(drbd_dev *mdev, int gfp_mask)
+STATIC int _drbd_alloc_ee(drbd_dev *mdev,struct page* page,int mask)
 {
 	struct Tl_epoch_entry* e;
-	struct page *page = alloc_page(gfp_mask);
 
-	if (!page) return NULL;
-	/* if we got the page, we really want the ee, too,
-	 * even for "GFP_TRY".
-	 * we may wait, but better not cause IO,
-	 * we might be in the IO path (of our peer).
-	 *
-	 * kmem_cache does not like to mix different memory types.
-	 * so even if we alloc'ed the page from HIGHMEM,
-	 * the ee comes from normal memory.
-	 */
-	e = kmem_cache_alloc(drbd_ee_cache, (gfp_mask|GFP_NOIO) & ~(__GFP_HIGHMEM));
-
-	if (e) drbd_ee_init(e,page);
-	else  __free_page(page);
-
-	return e;
-}
-
-/*
- * just free the ee and the associated page.
- * no accounting here, no list movement, nothing.
- */
-void drbd_free_ee(drbd_dev *mdev, struct Tl_epoch_entry *e)
-{
-	struct page *page = drbd_bio_get_page(&e->private_bio);
-ONLY_IN_26(
-	D_ASSERT(page == e->ee_bvec.bv_page);
-	page = e->ee_bvec.bv_page;		/* just in case */
-)
-	__free_page(page);
-	kmem_cache_free(drbd_ee_cache, e);
-}
-
-/* XXX currently only called on module init.
- * actually we do not need to care about spinlocks.
- */
-int drbd_init_ee(drbd_dev *mdev)
-{
-	LIST_HEAD(tmp);
-	struct Tl_epoch_entry *e;
-	int i;
-
-	for (i = 0; i < EE_MININUM; i++) {
-		e = drbd_try_alloc_ee(mdev,GFP_HIGHUSER);
-		if (!e) {
-			ERR("Failed to allocate %d EEs !\n",EE_MININUM);
-			break;
-		}
-		list_add(&e->w.list,&tmp);
-	}
+	e = kmem_cache_alloc(drbd_ee_cache, mask);
+	if( e == NULL ) return FALSE;
 
+	drbd_ee_init(e,page);
 	spin_lock_irq(&mdev->ee_lock);
-	list_splice(&tmp,&mdev->free_ee);
-	mdev->ee_vacant += i;
+	list_add(&e->w.list,&mdev->free_ee);
+	mdev->ee_vacant++;
 	spin_unlock_irq(&mdev->ee_lock);
 
-	return (i == EE_MININUM);
+	return TRUE;
+}
+
+/* bool */
+STATIC int drbd_alloc_ee(drbd_dev *mdev,int mask)
+{
+	struct page *page;
+
+	page=alloc_page(mask);
+	if(!page) return FALSE;
+
+	if(!_drbd_alloc_ee(mdev,page,GFP_KERNEL)) {
+		__free_page(page);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+STATIC struct page* drbd_free_ee(drbd_dev *mdev, struct list_head *list)
+{
+	struct list_head *le;
+	struct Tl_epoch_entry* e;
+	struct page* page;
+
+	MUST_HOLD(&mdev->ee_lock);
+
+	D_ASSERT(!list_empty(list));
+	le = list->next;
+	e = list_entry(le, struct Tl_epoch_entry, w.list);
+	list_del(le);
+
+	page = drbd_bio_get_page(&e->private_bio);
+ONLY_IN_26(
+	D_ASSERT(page == e->ee_bvec.bv_page);
+	page = e->ee_bvec.bv_page;
+)
+	kmem_cache_free(drbd_ee_cache, e);
+	mdev->ee_vacant--;
+
+	return page;
+}
+
+int drbd_init_ee(drbd_dev *mdev)
+{
+	while(mdev->ee_vacant < EE_MININUM ) {
+		if(!drbd_alloc_ee(mdev,GFP_USER)) {
+			ERR("Failed to allocate %d EEs !\n",EE_MININUM);
+			return 0;
+		}
+	}
+	return 1;
 }
 
 int drbd_release_ee(drbd_dev *mdev,struct list_head* list)
 {
 	int count=0;
-	struct Tl_epoch_entry *e;
-	struct list_head *le, *tmp;
 
 	spin_lock_irq(&mdev->ee_lock);
-	list_for_each_safe(le,tmp,list) {
-		list_del(le);
-		e = list_entry(le, struct Tl_epoch_entry, w.list);
-		drbd_free_ee(mdev,e);
-		mdev->ee_vacant--;
+	while(!list_empty(list)) {
+		__free_page(drbd_free_ee(mdev,list));
 		count++;
 	}
 	spin_unlock_irq(&mdev->ee_lock);
@@ -288,22 +280,18 @@ STATIC void finish_wait(wait_queue_head_t *q, wait_queue_t *wait)
 #endif
 
 /**
- * drbd_get_ee:
- * Returns an Tl_epoch_entry; might sleep. Fails only if a signal comes in.
- *
- * We do not do this straight forward, because we don't want to consume
- * too much memory: the upper bound is mdev->conf.max_buffers.
+ * drbd_get_ee: Returns an Tl_epoch_entry; might sleep. Fails only if
+ * a signal comes in.
  */
 struct Tl_epoch_entry* drbd_get_ee(drbd_dev *mdev)
 {
 	struct list_head *le;
-	struct Tl_epoch_entry* e = NULL;
+	struct Tl_epoch_entry* e;
 	DEFINE_WAIT(wait);
 
 	MUST_HOLD(&mdev->ee_lock);
 
-	if(mdev->ee_vacant == EE_MININUM / 2 && 
-	   (mdev->conf.unplug_watermark != mdev->conf.max_buffers) ) {
+	if(mdev->ee_vacant == EE_MININUM / 2) {
 		spin_unlock_irq(&mdev->ee_lock);
 		drbd_kick_lo(mdev);
 		spin_lock_irq(&mdev->ee_lock);
@@ -311,50 +299,43 @@ struct Tl_epoch_entry* drbd_get_ee(drbd_dev *mdev)
 
 	if(list_empty(&mdev->free_ee)) _drbd_process_ee(mdev,1);
 
-	while (list_empty(&mdev->free_ee)) {
-		if( mdev->conf.max_buffers >
-		    mdev->ee_vacant + mdev->ee_in_use )
-		{
+	if(list_empty(&mdev->free_ee)) {
+		for (;;) {
+			prepare_to_wait(&mdev->ee_wait, &wait, 
+					TASK_INTERRUPTIBLE);
+			if(!list_empty(&mdev->free_ee)) break;
 			spin_unlock_irq(&mdev->ee_lock);
-			e = drbd_try_alloc_ee(mdev,GFP_TRY);
+			if( ( mdev->ee_vacant+mdev->ee_in_use) < 
+			      mdev->conf.max_buffers ) {
+				if(drbd_alloc_ee(mdev,GFP_TRY)) {
+					spin_lock_irq(&mdev->ee_lock);
+					break;
+				}
+			}
+			drbd_kick_lo(mdev);
+			schedule();
 			spin_lock_irq(&mdev->ee_lock);
-			if (e) goto got_new_ee;
+			finish_wait(&mdev->ee_wait, &wait);
+			if (signal_pending(current)) {
+				WARN("drbd_get_ee interrupted!\n");
+				return 0;
+			}
+			// finish wait is inside, so that we are TASK_RUNNING 
+			// in _drbd_process_ee (which might sleep by itself.)
+			_drbd_process_ee(mdev,1);
 		}
-
-		prepare_to_wait(&mdev->ee_wait, &wait, TASK_INTERRUPTIBLE);
-		if (!list_empty(&mdev->free_ee)) break;
-
-		spin_unlock_irq(&mdev->ee_lock);
-
-		drbd_kick_lo(mdev);
-		schedule();
-
-		// finish wait is inside, so that we are TASK_RUNNING
-		// in _drbd_process_ee (which might sleep by itself.)
-		finish_wait(&mdev->ee_wait, &wait);
-
-		spin_lock_irq(&mdev->ee_lock);
-
-		if (signal_pending(current)) {
-			WARN("drbd_get_ee interrupted!\n");
-			return NULL;
-		}
-
-		_drbd_process_ee(mdev,1);
+		finish_wait(&mdev->ee_wait, &wait); 
 	}
-	/* no-op when &wait has not been touched */
-	finish_wait(&mdev->ee_wait, &wait);
 
-	le = mdev->free_ee.next;
+	le=mdev->free_ee.next;
 	list_del(le);
 	mdev->ee_vacant--;
+	mdev->ee_in_use++;
 	e=list_entry(le, struct Tl_epoch_entry, w.list);
 ONLY_IN_26(
 	D_ASSERT(e->private_bio.bi_idx == 0);
 	drbd_ee_init(e,e->ee_bvec.bv_page); // reinitialize
 )
-  got_new_ee:
-	mdev->ee_in_use++;
 	e->block_id = !ID_VACANT;
 	SET_MAGIC(e);
 	return e;
@@ -362,30 +343,27 @@ ONLY_IN_26(
 
 void drbd_put_ee(drbd_dev *mdev,struct Tl_epoch_entry *e)
 {
+	struct page* page;
+
 	MUST_HOLD(&mdev->ee_lock);
 
 	D_ASSERT(page_count(drbd_bio_get_page(&e->private_bio)) == 1);
 
 	mdev->ee_in_use--;
+	mdev->ee_vacant++;
 	e->block_id = ID_VACANT;
 	INVALIDATE_MAGIC(e);
+	list_add_tail(&e->w.list,&mdev->free_ee);
 
-	if ( (mdev->ee_vacant * 2 > mdev->ee_in_use) &&
-	     (mdev->ee_vacant + mdev->ee_in_use > EE_MININUM) ) {
-		drbd_free_ee(mdev,e);
-	} else {
-		list_add_tail(&e->w.list,&mdev->free_ee);
-		mdev->ee_vacant++;
+	if((mdev->ee_vacant * 2 > mdev->ee_in_use ) &&
+	   ( mdev->ee_vacant + mdev->ee_in_use > EE_MININUM) ) {
+		// FIXME cleanup: never returns NULL anymore
+		page=drbd_free_ee(mdev,&mdev->free_ee);
+		if( page ) __free_page(page);
 	}
-	if (mdev->ee_in_use == 0) {
-		struct Tl_epoch_entry *e;
-		struct list_head *le, *tmp;
-		list_for_each_safe(le,tmp,&mdev->free_ee) {
-			if (mdev->ee_vacant <= EE_MININUM) break;
-			list_del(le);
-			e = list_entry(le, struct Tl_epoch_entry, w.list);
-			drbd_free_ee(mdev,e);
-			mdev->ee_vacant--;
+	if(mdev->ee_in_use == 0) {
+		while( mdev->ee_vacant > EE_MININUM ) {
+			__free_page(drbd_free_ee(mdev,&mdev->free_ee));
 		}
 	}
 
@@ -429,6 +407,7 @@ STATIC int _drbd_process_ee(drbd_dev *mdev, int be_sleepy)
 
 	if( test_and_set_bit(PROCESS_EE_RUNNING,&mdev->flags) ) {
 		if(!be_sleepy) {
+			clear_bit(PROCESS_EE_RUNNING,&mdev->flags);
 			return 3;
 		}
 		spin_unlock_irq(&mdev->ee_lock);
@@ -636,37 +615,15 @@ STATIC struct socket *drbd_try_connect(drbd_dev *mdev)
 {
 	int err;
 	struct socket *sock;
-	struct sockaddr_in src_in;
 
 	err = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
 	if (err) {
 		ERR("sock_creat(..)=%d\n", err);
-		return NULL;
 	}
 
 	sock->sk->SK_(rcvtimeo) =
 	sock->sk->SK_(sndtimeo) =  mdev->conf.try_connect_int*HZ;
 
-       /* explicitly bind to the configured IP as source IP 
-	   for the outgoing connections.
-	   This is needed for multihomed hosts and to be 
-	   able to use lo: interfaces for drbd.
-          Make sure to use 0 as portnumber, so linux selects
-	   a free one dynamically.
-	*/
-	memcpy (&src_in, &(mdev->conf.my_addr), sizeof(struct sockaddr_in));
-	src_in.sin_port = 0; 
-
-	err = sock->ops->bind(sock,
-			      (struct sockaddr * ) &src_in,
-			      sizeof (struct sockaddr_in));
-	if (err) {
-		ERR("Unable to bind source sock (%d)\n", err);
-		sock_release(sock);
-		sock = NULL;
-		return sock;
-	}
-
 	err = sock->ops->connect(sock,
 				 (struct sockaddr *) mdev->conf.other_addr,
 				 mdev->conf.other_addr_len, 0);
@@ -686,7 +643,7 @@ STATIC struct socket *drbd_wait_for_connect(drbd_dev *mdev)
 	err = sock_create(AF_INET, SOCK_STREAM, 0, &sock2);
 	if (err) {
 		ERR("sock_creat(..)=%d\n", err);
-		return NULL;
+		// FIXME return NULL ?
 	}
 
 	sock2->sk->SK_(reuse)    = 1; /* SO_REUSEADDR */
@@ -697,7 +654,7 @@ STATIC struct socket *drbd_wait_for_connect(drbd_dev *mdev)
 			      (struct sockaddr *) mdev->conf.my_addr,
 			      mdev->conf.my_addr_len);
 	if (err) {
-		ERR("Unable to bind sock2 (%d)\n", err);
+		ERR("Unable to bind (%d)\n", err);
 		sock_release(sock2);
 		set_cstate(mdev,Unconnected);
 		return 0;
@@ -711,27 +668,15 @@ STATIC struct socket *drbd_wait_for_connect(drbd_dev *mdev)
 
 STATIC int drbd_do_handshake(drbd_dev *mdev);
 
-/*
- * return values:
- *   1 yess, we have a valid connection
- *   0 oops, did not work out, please try again
- *  -1 peer talks different language,
- *     no point in trying again, please go standalone.
- */
 int drbd_connect(drbd_dev *mdev)
 {
 	struct socket *sock,*msock;
-	int h;
 
 	D_ASSERT(mdev->cstate!=Unconfigured);
 	D_ASSERT(!mdev->data.socket);
 
 	set_cstate(mdev,WFConnection);
 
-	/* Break out of unknown connect loops by random wait here. */
-	set_current_state(TASK_INTERRUPTIBLE);
-	schedule_timeout(net_random() % ((mdev->conf.try_connect_int*HZ)/4));
-
 	while(1) {
 		sock=drbd_try_connect(mdev);
 		if(sock) {
@@ -754,12 +699,12 @@ int drbd_connect(drbd_dev *mdev)
 				sock_release(sock);
 			}
 		}
-		if(mdev->cstate==Unconnected) return -1;
+		if(mdev->cstate==Unconnected) return 0;
 		if(signal_pending(current)) {
 			drbd_flush_signals(current);
 			smp_rmb();
 			if (get_t_state(&mdev->receiver) == Exiting)
-				return -1;
+				return 0;
 		}
 	}
 
@@ -768,8 +713,10 @@ int drbd_connect(drbd_dev *mdev)
 	msock->sk->SK_(reuse)=1; /* SO_REUSEADDR */
 	sock->sk->SK_(reuse)=1; /* SO_REUSEADDR */
 
-	sock->sk->SK_(allocation) = GFP_NOIO;
-	msock->sk->SK_(allocation) = GFP_NOIO;
+	/* to prevent oom deadlock... */
+	/* The default allocation priority was GFP_KERNEL */
+	sock->sk->SK_(allocation) = GFP_DRBD;
+	msock->sk->SK_(allocation) = GFP_DRBD;
 
 	sock->sk->SK_(priority)=TC_PRIO_BULK;
 	NOT_IN_26(sock->sk->tp_pinfo.af_tcp.nonagle=0;)
@@ -799,8 +746,9 @@ int drbd_connect(drbd_dev *mdev)
 	set_cstate(mdev,WFReportParams);
 	D_ASSERT(mdev->asender.task == NULL);
 
-	h = drbd_do_handshake(mdev);
-	if (h <= 0) return h;
+	if (!drbd_do_handshake(mdev)) {
+		return 0;
+	}
 
 	clear_bit(ON_PRI_INC_HUMAN,&mdev->flags);
 	clear_bit(ON_PRI_INC_TIMEOUTEX,&mdev->flags);
@@ -810,8 +758,7 @@ int drbd_connect(drbd_dev *mdev)
 
 	drbd_thread_start(&mdev->asender);
 
-	drbd_send_param(mdev,test_bit(IO_FROZEN, &mdev->flags) ? 4 : 0);
-	clear_bit(USE_DEGR_WFC_T,&mdev->flags);
+	drbd_send_param(mdev,0);
 
 	return 1;
 }
@@ -863,8 +810,8 @@ STATIC int receive_Barrier(drbd_dev *mdev, Drbd_Header* h)
 	spin_lock_irq(&mdev->ee_lock);
 	rv = _drbd_process_ee(mdev,1);
 
-	epoch_size=atomic_read(&mdev->epoch_size);
-	atomic_set(&mdev->epoch_size,0);
+	epoch_size=mdev->epoch_size;
+	mdev->epoch_size=0;
 	spin_unlock_irq(&mdev->ee_lock);
 
 	rv &= drbd_send_b_ack(mdev, p->barrier, epoch_size);
@@ -909,8 +856,11 @@ STATIC void receive_data_tail(drbd_dev *mdev,int data_size)
 	/* kick lower level device, if we have more than (arbitrary number)
 	 * reference counts on it, which typically are locally submitted io
 	 * requests.  don't use unacked_cnt, so we speed up proto A and B, too.
+	 *
+	 * XXX maybe: make that arbitrary number configurable.
+	 * for now, I choose 1/16 of max-epoch-size.
 	 */
-	if (atomic_read(&mdev->local_cnt) >= mdev->conf.unplug_watermark ) {
+	if (atomic_read(&mdev->local_cnt) >= (mdev->conf.max_epoch_size>>4) ) {
 		drbd_kick_lo(mdev);
 	}
 	mdev->writ_cnt+=data_size>>9;
@@ -924,7 +874,7 @@ STATIC int recv_dless_read(drbd_dev *mdev, drbd_request_t *req,
 
 	bio = req->master_bio;
 
-	D_ASSERT( sector == req->sector );
+	D_ASSERT( sector == drbd_req_get_sector(req) );
 
 	rr=drbd_recv(mdev,drbd_bio_kmap(bio),data_size);
 	drbd_bio_kunmap(bio);
@@ -956,7 +906,7 @@ STATIC int e_end_resync_block(drbd_dev *mdev, struct drbd_work *w, int unused)
 			 */
 		}
 		ok = drbd_send_ack(mdev,WriteAck,e);
-		set_bit(SYNC_STARTED,&mdev->flags);
+		__set_bit(SYNC_STARTED,&mdev->flags);
 	} else {
 		ok = drbd_send_ack(mdev,NegAck,e);
 		ok&= drbd_io_error(mdev);
@@ -1073,7 +1023,7 @@ STATIC int e_end_block(drbd_dev *mdev, struct drbd_work *w, int unused)
 	sector_t sector = drbd_ee_get_sector(e);
 	int ok=1;
 
-	atomic_inc(&mdev->epoch_size);
+	mdev->epoch_size++;
 	if(mdev->conf.wire_protocol == DRBD_PROT_C) {
 		if(likely(drbd_bio_uptodate(&e->private_bio))) {
 			ok=drbd_send_ack(mdev,WriteAck,e);
@@ -1203,6 +1153,7 @@ STATIC int receive_DataRequest(drbd_dev *mdev,Drbd_Header *h)
 		return FALSE;
 	}
 	e->block_id = p->block_id; // no meaning on this side, pr* on partner
+	list_add(&e->w.list,&mdev->read_ee);
 	spin_unlock_irq(&mdev->ee_lock);
 
 	if(!inc_local(mdev) || (mdev->gen_cnt[Flags] & MDF_Consistent) == 0) {
@@ -1231,9 +1182,7 @@ STATIC int receive_DataRequest(drbd_dev *mdev,Drbd_Header *h)
 		if (!drbd_rs_begin_io(mdev,sector)) {
 			// we have been interrupted, probably connection lost!
 			D_ASSERT(signal_pending(current));
-			spin_lock_irq(&mdev->ee_lock);
 			drbd_put_ee(mdev,e);
-			spin_unlock_irq(&mdev->ee_lock);
 			return 0;
 		}
 		break;
@@ -1244,14 +1193,7 @@ STATIC int receive_DataRequest(drbd_dev *mdev,Drbd_Header *h)
 
 	mdev->read_cnt += size >> 9;
 	inc_unacked(mdev);
-	spin_lock_irq(&mdev->ee_lock);
-	list_add(&e->w.list,&mdev->read_ee);
-	spin_unlock_irq(&mdev->ee_lock);
 	drbd_generic_make_request(READ,&e->private_bio);
-	if (atomic_read(&mdev->local_cnt) >= (mdev->conf.max_epoch_size>>4) ) {
-		drbd_kick_lo(mdev);
-	}
-
 
 	return TRUE;
 }
@@ -1402,8 +1344,8 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 {
 	Drbd_Parameter_Packet *p = (Drbd_Parameter_Packet*)h;
 	int consider_sync;
-	int oo_state,i;
-	sector_t p_size, p_usize, my_usize;
+	int oo_state;
+	sector_t p_size;
 
 	if (h->length != (sizeof(*p)-sizeof(*h))) {
 		ERR("Incompatible packet size of Parameter packet!\n");
@@ -1480,44 +1422,15 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 
 	set_bit(MD_DIRTY,&mdev->flags); // we are changing state!
 
-	p_usize=be64_to_cpu(p->u_size);
-	/*
-	 * you may get a flip-flop connection established/connection loss, in
-	 * case both really have different usize uppon first connect!
-	 * try to solve it thus:
-	 ***/
-#define min_not_zero(l, r) (l == 0) ? r : ((r == 0) ? l : min(l, r))
-	if (mdev->cstate == WFReportParams) {
-		/* this is first connect, or an otherwise expected param
-		 * exchange.  choose the minimum */
-		p_usize = min_not_zero(mdev->lo_usize, p_usize);
-	} else {
-		/* this was an "unexpected" param packet,
-		 * just do what the peer suggests */
-	}
-#undef min_not_zero
-
-	my_usize = mdev->lo_usize;
-
-	if( mdev->lo_usize > p_usize ) {
-		mdev->lo_usize = p_usize;
-		INFO("Peer sets u_size to %lu KB\n",
-		     (unsigned long)mdev->lo_usize);
-	}
-
-	if( drbd_new_dev_size(mdev) <
-	    (drbd_get_capacity(mdev->this_bdev)>>1) &&
-	    mdev->gen_cnt[Flags] & MDF_Consistent ) {
-		ERR("The peer's disk size is too small!\n");
-		set_cstate(mdev,StandAlone);
-		drbd_thread_stop_nowait(&mdev->receiver);
-		mdev->lo_usize = my_usize;
-		return FALSE;
-	}
+/*lge:
+ * FIXME
+ * please get the order of tests (re)settings for consider_sync
+ * right, and comment them!
+ */
 
 	consider_sync = (mdev->cstate == WFReportParams);
-	drbd_determin_dev_size(mdev);
-	if(drbd_disk_less_node_present(mdev)) consider_sync=0;
+	if(drbd_determin_dev_size(mdev)) consider_sync=0;
+
 	if(test_bit(DISKLESS, &mdev->flags)) consider_sync=0;
 
 	drbd_bm_unlock(mdev); // }
@@ -1528,8 +1441,6 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 	}
 	if(be32_to_cpu(p->flags)&2) consider_sync=1;
 
-	if(be32_to_cpu(p->flags)&4) consider_sync=0;
-
 	// XXX harmless race with ioctl ...
 	mdev->sync_conf.rate  =
 		max_t(int,mdev->sync_conf.rate, be32_to_cpu(p->sync_rate));
@@ -1540,6 +1451,12 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 	mdev->sync_conf.group =
 		min_t(int,mdev->sync_conf.group,be32_to_cpu(p->sync_group));
 
+	if( mdev->lo_usize != be64_to_cpu(p->u_size) ) {
+		mdev->lo_usize = be64_to_cpu(p->u_size);
+		INFO("Peer sets u_size to %lu KB\n",
+		     (unsigned long)mdev->lo_usize);
+	}
+
 	if(!p_size) {
 		/* no point in trying to sync a diskless peer: */
 		consider_sync = 0;
@@ -1585,40 +1502,19 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 
 	if (mdev->cstate == WFReportParams) {
 		INFO("Connection established.\n");
-
-		if(test_bit(IO_FROZEN, &mdev->flags)) {
-			WARN("Going to thaw IO, resending %d requests.\n",
-			     atomic_read(&mdev->ap_pending_cnt));
-			tl_resend(mdev);
-			set_bit(ISSUE_BARRIER,&mdev->flags);
-			if (mdev->cstate == WFReportParams) {
-				consider_sync = 0;
-			} else return FALSE;
-		}
 	}
 
 	if (consider_sync) {
 		if (!drbd_sync_handshake(mdev,p)) return FALSE;
 	}
 
-	if (mdev->cstate == WFReportParams) {
-		set_cstate(mdev,Connected);
-		if(test_bit(IO_FROZEN, &mdev->flags)) {
-			clear_bit(IO_FROZEN, &mdev->flags);
-			drbd_thaw_frozen_reqs(mdev);
-			consider_sync = 0;
-		}
-	}
+	if (mdev->cstate == WFReportParams) set_cstate(mdev,Connected);
 
 	oo_state = mdev->o_state;
 	mdev->o_state = be32_to_cpu(p->state);
 	if(oo_state == Secondary && mdev->o_state == Primary) {
-		/* Secondary has to adopt primary's gen_cnt. */
-		for(i=HumanCnt;i<GEN_CNT_SIZE;i++) {
-			mdev->gen_cnt[i]=be32_to_cpu(p->gen_cnt[i]);
-		}
+		drbd_md_inc(mdev,ConnectedCnt);
 	}
-
 	if (oo_state != mdev->o_state) {
 		INFO( "%s/%s --> %s/%s\n",
 		      nodestate_to_name(mdev->state),
@@ -1668,7 +1564,6 @@ STATIC int receive_bitmap(drbd_dev *mdev, Drbd_Header *h)
 		D_ASSERT(h->command == ReportBitMap);
 	}
 
-	clear_bit(CRASHED_PRIMARY, &mdev->flags); // md_write() is in drbd_start_resync.
 	if (mdev->cstate == WFBitMapS) {
 		drbd_start_resync(mdev,SyncSource);
 	} else if (mdev->cstate == WFBitMapT) {
@@ -1753,11 +1648,12 @@ STATIC int receive_skip(drbd_dev *mdev,Drbd_Header *h)
 STATIC int receive_BecomeSyncTarget(drbd_dev *mdev, Drbd_Header *h)
 {
 	ERR_IF(!mdev->bitmap) return FALSE;
-	ERR_IF(mdev->state != Secondary)
-		return FALSE;
-	ERR_IF(mdev->cstate != Connected)
-		return FALSE;
-	ERR_IF(test_bit(DISKLESS,&mdev->flags))
+
+	/* THINK
+	 * otherwise this does not make much sense, no?
+	 * and some other assertion maybe about cstate...
+	 */
+	ERR_IF(mdev->state != Secondary || mdev->cstate != Connected)
 		return FALSE;
 
 	drbd_bm_lock(mdev);
@@ -1765,24 +1661,18 @@ STATIC int receive_BecomeSyncTarget(drbd_dev *mdev, Drbd_Header *h)
 	drbd_bm_write(mdev);
 	drbd_start_resync(mdev,SyncTarget);
 	drbd_bm_unlock(mdev);
-	return TRUE;
+	return TRUE; // cannot fail ?
 }
 
 STATIC int receive_BecomeSyncSource(drbd_dev *mdev, Drbd_Header *h)
 {
-	ERR_IF(mdev->cstate != Connected)
-		return FALSE;
-	ERR_IF(test_bit(DISKLESS,&mdev->flags))
-		return FALSE;
-	ERR_IF(!drbd_md_test_flag(mdev,MDF_Consistent))
-		return FALSE;
-
+	// FIXME asserts ?
 	drbd_bm_lock(mdev);
 	drbd_bm_set_all(mdev);
 	drbd_bm_write(mdev);
 	drbd_start_resync(mdev,SyncSource);
 	drbd_bm_unlock(mdev);
-	return TRUE;
+	return TRUE; // cannot fail ?
 }
 
 STATIC int receive_UnplugRemote(drbd_dev *mdev, Drbd_Header *h)
@@ -1857,6 +1747,30 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	D_ASSERT(mdev->cstate < Connected);
 	mdev->o_state = Unknown;
 
+	/* in case we have been syncing, and then we drop the connection,
+	 * we need to "w_resume_next_sg", which we try to achieve by
+	 * setting the STOP_SYNC_TIMER bit, and schedulung the timer for
+	 * immediate execution.
+	 * unfortunately we cannot be sure that the timer already triggered.
+	 *
+	 * so we del_timer_sync here, and check that bit.
+	 * if it is still set, we queue w_resume_next_sg anyways,
+	 * just to be sure.
+	 */
+
+	del_timer_sync(&mdev->resync_timer);
+	spin_lock_irq(&mdev->req_lock);
+	if (test_and_clear_bit(STOP_SYNC_TIMER,&mdev->flags)) {
+		mdev->resync_work.cb = w_resume_next_sg;
+		if (list_empty(&mdev->resync_work.list))
+			_drbd_queue_work(&mdev->data.work,&mdev->resync_work);
+		// else: already queued, we only need to release the lock.
+	} else {
+		D_ASSERT(mdev->resync_work.cb == w_resync_inactive);
+	}
+	spin_unlock_irq(&mdev->req_lock);
+
+
 	drbd_thread_stop_nowait(&mdev->worker);
 	drbd_thread_stop(&mdev->asender);
 
@@ -1888,20 +1802,11 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	drbd_wait_ee(mdev,&mdev->sync_ee);
 	drbd_clear_done_ee(mdev);
 
-	if(test_bit(IO_FROZEN, &mdev->flags)) {
-		WARN("IO frozen with ap_pending_cnt = %d\n",
-		     atomic_read(&mdev->ap_pending_cnt));
-	} else {
-		tl_clear(mdev);
-		clear_bit(ISSUE_BARRIER,&mdev->flags);
-		wait_event( mdev->cstate_wait, atomic_read(&mdev->ap_pending_cnt)==0 );
-		D_ASSERT(mdev->oldest_barrier->n_req == 0);
-
-		if(atomic_read(&mdev->ap_pending_cnt)) {
-			ERR("ap_pending_cnt = %d\n",atomic_read(&mdev->ap_pending_cnt));
-			atomic_set(&mdev->ap_pending_cnt,0);
-		}
-	}
+	// primary
+	tl_clear(mdev);
+	clear_bit(ISSUE_BARRIER,&mdev->flags);
+	wait_event( mdev->cstate_wait, atomic_read(&mdev->ap_pending_cnt)==0 );
+	D_ASSERT(mdev->oldest_barrier->n_req == 0);
 
 	// both
 	clear_bit(PARTNER_CONSISTENT, &mdev->flags);
@@ -1913,7 +1818,7 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	D_ASSERT(list_empty(&mdev->sync_ee)); // done here
 	D_ASSERT(list_empty(&mdev->done_ee)); // done here
 
-	atomic_set(&mdev->epoch_size,0);
+	mdev->epoch_size=0;
 	mdev->rs_total=0;
 
 	if(atomic_read(&mdev->unacked_cnt)) {
@@ -1933,14 +1838,17 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	   on the fly. */
 	atomic_set(&mdev->rs_pending_cnt,0);
 
+	if(atomic_read(&mdev->ap_pending_cnt)) {
+		ERR("ap_pending_cnt = %d\n",atomic_read(&mdev->ap_pending_cnt));
+		atomic_set(&mdev->ap_pending_cnt,0);
+	}
+
 	wake_up(&mdev->cstate_wait);
 
 	if ( mdev->state == Primary &&
 	    ( test_bit(DISKLESS,&mdev->flags)
 	    || !drbd_md_test_flag(mdev,MDF_Consistent) ) ) {
-		drbd_thread_stop_nowait(&mdev->receiver);
 		drbd_panic("Sorry, I have no access to good data anymore.\n");
-		return;
 	}
 
 	if (get_t_state(&mdev->receiver) == Exiting) {
@@ -1964,9 +1872,6 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	}
 	clear_bit(DO_NOT_INC_CONCNT,&mdev->flags);
 
-	/* it may still be set, because some unplug was on the fly */
-	NOT_IN_26(mdev->flags &= ~(1<<UNPLUG_QUEUED);)
-
 	INFO("Connection lost.\n");
 }
 
@@ -1997,13 +1902,6 @@ int drbd_send_handshake(drbd_dev *mdev)
 	return ok;
 }
 
-/*
- * return values:
- *   1 yess, we have a valid connection
- *   0 oops, did not work out, please try again
- *  -1 peer talks different language,
- *     no point in trying again, please go standalone.
- */
 STATIC int drbd_do_handshake(drbd_dev *mdev)
 {
 	// ASSERT current == mdev->receiver ...
@@ -2012,25 +1910,25 @@ STATIC int drbd_do_handshake(drbd_dev *mdev)
 	int rv;
 
 	rv = drbd_send_handshake(mdev);
-	if (!rv) goto break_c_loop;
+	if (!rv) return 0;
 
 	rv = drbd_recv_header(mdev,&p->head);
-	if (!rv) goto break_c_loop;
+	if (!rv) return 0;
 
 	if (p->head.command == ReportParams) {
 		ERR("expected HandShake packet, received ReportParams...\n");
 		ERR("peer probaly runs some incompatible 0.7 -preX version\n");
-		return -1;
+		return 0;
 	} else if (p->head.command != HandShake) {
 		ERR( "expected HandShake packet, received: %s (0x%04x)\n",
 		     cmdname(p->head.command), p->head.command );
-		return -1;
+		return 0;
 	}
 
 	if (p->head.length != expect) {
 		ERR( "expected HandShake length: %u, received: %u\n",
 		     expect, p->head.length );
-		return -1;
+		return 0;
 	}
 
 	rv = drbd_recv(mdev, &p->head.payload, expect);
@@ -2062,53 +1960,29 @@ STATIC int drbd_do_handshake(drbd_dev *mdev)
 		ERR( "incompatible DRBD dialects: "
 		     "I support %u, peer wants %u\n",
 		     PRO_VERSION, p->protocol_version );
-		return -1;
+		return 0;
 	}
 
 	return 1;
-
- break_c_loop:
-	WARN( "Network error during initial handshake. I'll try again.\n");
-	/* In case a tcp connection set-up takes longer than
-	   connect-int, we might get into the situation that this
-	   node's msock gets connected to the peer's sock!
-
-	   To break out of this endless loop behaviour, we need to
-	   wait until the peer's msock connect tries are over. (1 Second)
-
-	   Additionally we wait connect-int/2 to increase the chance for
-	   our next connect() to hit exactly the peer's accept() window. */
-
-	set_current_state(TASK_INTERRUPTIBLE);
-	schedule_timeout(HZ + (mdev->conf.try_connect_int*HZ)/2);
-
-	return 0;
 }
 
 int drbdd_init(struct Drbd_thread *thi)
 {
 	drbd_dev *mdev = thi->mdev;
 	int minor = (int)(mdev-drbd_conf);
-	int h;
 
 	sprintf(current->comm, "drbd%d_receiver", minor);
 
 	/* printk(KERN_INFO DEVICE_NAME ": receiver living/m=%d\n", minor); */
 
 	while (TRUE) {
-		h = drbd_connect(mdev);
-		if (h <= 0) {
+		if (!drbd_connect(mdev)) {
+			WARN("Discarding network configuration.\n");
 			/* FIXME DISKLESS StandAlone
 			 * does not make much sense...
 			 * drbd_disconnect should set cstate properly...
 			 */
 			drbd_disconnect(mdev);
-			if (h == 0) {
-				schedule_timeout(HZ);
-				continue;
-			}
-
-			WARN("Discarding network configuration.\n");
 			set_cstate(mdev,StandAlone);
 			break;
 		}
@@ -2171,19 +2045,13 @@ STATIC int got_BlockAck(drbd_dev *mdev, Drbd_Header* h)
 
 		if( is_syncer_blk(mdev,p->block_id)) {
 			drbd_set_in_sync(mdev,sector,blksize);
-			set_bit(SYNC_STARTED,&mdev->flags);
+			__set_bit(SYNC_STARTED,&mdev->flags);
 		} else {
 			req=(drbd_request_t*)(long)p->block_id;
 
 			ERR_IF (!VALID_POINTER(req)) return FALSE;
 
-			spin_lock_irq(&mdev->req_lock);
-			if (mdev->conf.wire_protocol == DRBD_PROT_C) {
-				req->rq_status &= ~RQ_DRBD_IN_TL;
-				list_del(&req->w.list);
-			}
-			_drbd_end_req(req, RQ_DRBD_SENT, 1);
-			spin_unlock_irq(&mdev->req_lock);
+			drbd_end_req(req, RQ_DRBD_SENT, 1, sector);
 
 			if (test_bit(SYNC_STARTED,&mdev->flags) &&
 			    mdev->conf.wire_protocol == DRBD_PROT_C)
@@ -2195,6 +2063,7 @@ STATIC int got_BlockAck(drbd_dev *mdev, Drbd_Header* h)
 		dec_rs_pending(mdev);
 	} else {
 		D_ASSERT(mdev->conf.wire_protocol != DRBD_PROT_A);
+		dec_ap_pending(mdev);
 	}
 	return TRUE;
 }
@@ -2230,40 +2099,25 @@ STATIC int got_NegAck(drbd_dev *mdev, Drbd_Header* h)
 
 STATIC int got_NegDReply(drbd_dev *mdev, Drbd_Header* h)
 {
-	/* drbd_request_t *req;
-	 * unused now */
+	drbd_request_t *req;
 	Drbd_BlockAck_Packet *p = (Drbd_BlockAck_Packet*)h;
 
-	if (is_syncer_blk(mdev,p->block_id)) {
-		/* no resync data available. don't panic just yet ... */
-		printk(KERN_EMERG DEVICE_NAME "%d: "
-		       "Got NegDReply for resync request. "
-		       "WE ARE LOST. We lost our up-to-date disk.\n",
-			(int)(mdev-drbd_conf));
-		return FALSE;
-	} /* else { */
-
-#if 0
-	/* hey, we panic anyways. so why bother? */
 	req = (drbd_request_t *)(long)p->block_id;
-	if (VALID_POINTER(req)) {
-		D_ASSERT(req->w.cb == w_is_app_read);
+	D_ASSERT(req->w.cb == w_is_app_read);
 
-		spin_lock(&mdev->pr_lock);
-		list_del(&req->w.list);
-		spin_unlock(&mdev->pr_lock);
+	spin_lock(&mdev->pr_lock);
+	list_del(&req->w.list);
+	spin_unlock(&mdev->pr_lock);
 
-		INVALIDATE_MAGIC(req);
-		mempool_free(req,drbd_request_mempool);
-	}
-#endif
+	INVALIDATE_MAGIC(req);
+	mempool_free(req,drbd_request_mempool);
 
 	drbd_panic("Got NegDReply. WE ARE LOST. We lost our up-to-date disk.\n");
 
 	// THINK do we have other options, but panic?
 	//       what about bio_endio, in case we don't panic ??
 
-	return FALSE;
+	return TRUE;
 }
 
 STATIC int got_NegRSDReply(drbd_dev *mdev, Drbd_Header* h)
@@ -2343,12 +2197,12 @@ int drbd_asender(struct Drbd_thread *thi)
 		 * this needs to be fixed properly, I'd vote for a separate
 		 * msock sender thread, but others will frown upon yet an other
 		 * kernel thread...
-		 *      -- lge
+		 *	-- lge
 		 */
 		set_bit(SIGNAL_ASENDER, &mdev->flags);
-		
+
 		if (!drbd_process_ee(mdev,0)) goto err;
-		
+
 		rv = drbd_recv_short(mdev,buf,expect-received);
 		clear_bit(SIGNAL_ASENDER, &mdev->flags);
 
