@@ -5,9 +5,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+	main author.
+
+   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
+	main contributor.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,92 +27,95 @@
 
  */
 
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/module.h>
 
 #include <linux/slab.h>
 #include <linux/drbd.h>
 #include "drbd_int.h"
 
-void _drbd_end_req(drbd_request_t *req, int nextstate, int er_flags)
+void drbd_end_req(drbd_request_t *req, int nextstate, int er_flags,
+		  sector_t rsector)
 {
+	/* This callback will be called in irq context by the IDE drivers,
+	   and in Softirqs/Tasklets/BH context by the SCSI drivers.
+	   This function is called by the receiver in kernel-thread context.
+	   Try to get the locking right :) */
+
 	struct Drbd_Conf* mdev = drbd_req_get_mdev(req);
-	sector_t rsector = req->sector;
+	unsigned long flags=0;
 	int uptodate;
 
+	PARANOIA_BUG_ON(!IS_VALID_MDEV(mdev));
+	PARANOIA_BUG_ON(drbd_req_get_sector(req) != rsector);
+	spin_lock_irqsave(&mdev->req_lock,flags);
+
 	if(req->rq_status & nextstate) {
 		ERR("request state error(%d)\n", req->rq_status);
 	}
 
-	if (nextstate == RQ_DRBD_SENT) dec_ap_pending(mdev);
-
 	req->rq_status |= nextstate;
 	req->rq_status &= er_flags | ~0x0001;
-	if( (req->rq_status & RQ_DRBD_DONE) != RQ_DRBD_DONE )
-		return;
+	if( (req->rq_status & RQ_DRBD_DONE) == RQ_DRBD_DONE ) goto end_it;
 
-	/* complete the master bio now. */
+	spin_unlock_irqrestore(&mdev->req_lock,flags);
 
-	/* if this sector was present in the current epoch, close it.
-	 * FIXME compares reusable pointer addresses,
-	 * possibly artificially reducing epoch size */
-	if (req->barrier == mdev->newest_barrier)
-		set_bit(ISSUE_BARRIER,&mdev->flags);
+	return;
+
+/* We only report uptodate == TRUE if both operations (WRITE && SEND)
+   reported uptodate == TRUE
+ */
+
+	end_it:
+	spin_unlock_irqrestore(&mdev->req_lock,flags);
+
+	if( req->rq_status & RQ_DRBD_IN_TL ) {
+		if( ! ( er_flags & ERF_NOTLD ) ) {
+			/*If this call is from tl_clear() we may not call 
+			  tl_dependene, otherwhise we have a homegrown 
+			  spinlock deadlock.   */
+			if(tl_dependence(mdev,req))
+				set_bit(ISSUE_BARRIER,&mdev->flags);
+		} else {
+			list_del(&req->w.list); // we have the tl_lock...
+		}
+	}
 
 	uptodate = req->rq_status & 0x0001;
 	if( !uptodate && mdev->on_io_error == Detach) {
-		drbd_set_out_of_sync(mdev,rsector, req->size);
+		drbd_set_out_of_sync(mdev,rsector, drbd_req_get_size(req));
 		// It should also be as out of sync on
 		// the other side!  See w_io_error()
 
 		drbd_bio_endio(req->master_bio,1);
-		req->master_bio = NULL;
 		dec_ap_bio(mdev);
 		// The assumption is that we wrote it on the peer.
 
 // FIXME proto A and diskless :)
 
 		req->w.cb = w_io_error;
-		_drbd_queue_work(&mdev->data.work,&req->w);
+		drbd_queue_work(mdev,&mdev->data.work,&req->w);
 
 		goto out;
 
 	}
 
 	drbd_bio_endio(req->master_bio,uptodate);
-	req->master_bio = NULL;
 	dec_ap_bio(mdev);
 
-	/* free the request,
-	 * if it is not/no longer in the transfer log, because
-	 *  it was local only (not connected), or
-	 *  this is protocol C, or
-	 *  the corresponding barrier ack has been received already, or
-	 *  it has been cleared from the transfer log (after connection loss)
-	 */
-	if (!(req->rq_status & RQ_DRBD_IN_TL)) {
-		INVALIDATE_MAGIC(req);
-		mempool_free(req,drbd_request_mempool);
-	}
+	INVALIDATE_MAGIC(req);
+	mempool_free(req,drbd_request_mempool);
 
  out:
 	if (test_bit(ISSUE_BARRIER,&mdev->flags)) {
+		spin_lock_irqsave(&mdev->req_lock,flags);
 		if(list_empty(&mdev->barrier_work.list)) {
 			_drbd_queue_work(&mdev->data.work,&mdev->barrier_work);
 		}
+		spin_unlock_irqrestore(&mdev->req_lock,flags);
 	}
 }
 
-void drbd_end_req(drbd_request_t *req, int nextstate, int er_flags)
-{
-	struct Drbd_Conf* mdev = drbd_req_get_mdev(req);
-	unsigned long flags=0;
-	spin_lock_irqsave(&mdev->req_lock,flags);
-	_drbd_end_req(req,nextstate,er_flags);
-	spin_unlock_irqrestore(&mdev->req_lock,flags);
-}
-
-
 int drbd_read_remote(drbd_dev *mdev, drbd_request_t *req)
 {
 	int rv;
@@ -167,34 +172,13 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 {
 	drbd_request_t *req;
 	int local, remote;
-	unsigned long flags;
+	int target_area_out_of_sync = FALSE; // only relevant for reads
 
-	ONLY_IN_26(
-	/* Currently our BARRIER code is disabled. */
-	if(unlikely(bio_barrier(bio))) {
-		bio_endio(bio, bio->bi_size, -EOPNOTSUPP);
-		return 0;
-	}
-	)
 	if (unlikely(drbd_did_panic == DRBD_MAGIC)) {
 		drbd_bio_IO_error(bio);
 		return 0;
 	}
 
-	if( rw == WRITE && test_bit(IO_FROZEN, &mdev->flags)) {
-		bio->REQ_NEXT = NULL;
-
-		spin_lock_irqsave(&mdev->req_lock,flags);
-		if(mdev->last_frozen_bio == NULL) {
-			mdev->first_frozen_bio = bio;
-			mdev->last_frozen_bio = bio;
-		} else {
-			mdev->last_frozen_bio->REQ_NEXT = bio;
-			mdev->last_frozen_bio = bio;
-		}
-		spin_unlock_irqrestore(&mdev->req_lock,flags);
-		return 0;
-	}
 	/*
 	 * If someone tries to mount on Secondary, and this is a 2.4 kernel,
 	 * it would lead to a readonly mounted, but not cache-coherent,
@@ -209,11 +193,9 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 	 * [ ... until we implement some shared mode, and our users confirm by
 	 * configuration, that they handle cache coherency themselves ... ]
 	 */
-	if (mdev->state != Primary &&
-		( !disable_bd_claim || rw == WRITE ) ) {
+	if (mdev->state != Primary) {
 		if (DRBD_ratelimit(5*HZ,5)) {
-			ERR("Not in Primary state, no %s requests allowed\n",
-					disable_bd_claim ? "WRITE" : "IO");
+			ERR("Not in Primary state, no IO requests allowed\n");
 		}
 		drbd_bio_IO_error(bio);
 		return 0;
@@ -242,7 +224,7 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 
 	/* allocate outside of all locks
 	 */
-	req = mempool_alloc(drbd_request_mempool, GFP_NOIO);
+	req = mempool_alloc(drbd_request_mempool, GFP_DRBD);
 	if (!req) {
 		/* only pass the error to the upper layers.
 		 * if user cannot handle io errors, thats not our business.
@@ -265,16 +247,10 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 	// down_read(mdev->device_lock);
 
 	wait_event( mdev->cstate_wait,
-		    (volatile int)mdev->cstate < WFBitMapS ||
-		    (volatile int) mdev->cstate > WFBitMapT);
+		    (volatile int)(mdev->cstate < WFBitMapS || 
+				   mdev->cstate > WFBitMapT) );
 
 	local = inc_local(mdev);
-	ONLY_IN_26(
-	if (!local) {
-		bio_put(req->private_bio); /* or we get a bio leak */
-		req->private_bio = NULL;
-	}
-	)
 	NOT_IN_26( if (rw == READA) rw=READ );
 	if (rw == READ || rw == READA) {
 		if (local) {
@@ -300,10 +276,6 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 				 * local io stack.
 				 */
 				local = 0;
-				ONLY_IN_26(
-				bio_put(req->private_bio);
-				req->private_bio = NULL;
-				)
 				dec_local(mdev);
 			}
 		}
@@ -360,8 +332,14 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 			if (!drbd_send_dblock(mdev,req)) {
 				if (mdev->cstate >= Connected)
 					set_cstate(mdev,NetworkFailure);
+				dec_ap_pending(mdev);
 				drbd_thread_restart_nowait(&mdev->receiver);
+			} else if(mdev->conf.wire_protocol == DRBD_PROT_A) {
+				dec_ap_pending(mdev);
+				drbd_end_req(req, RQ_DRBD_SENT, 1, sector);
 			}
+		} else if (target_area_out_of_sync) {
+			drbd_read_remote(mdev,req);
 		} else {
 			// this node is diskless ...
 			drbd_read_remote(mdev,req);
@@ -385,7 +363,7 @@ drbd_make_request_common(drbd_dev *mdev, int rw, int size,
 		else            mdev->read_cnt += size>>9;
 
 		// in 2.4.X, READA are submitted as READ.
-		drbd_generic_make_request(rw,drbd_req_private_bio(req));
+		drbd_generic_make_request(rw,&req->private_bio);
 	}
 
 	// up_read(mdev->device_lock);
@@ -443,27 +421,3 @@ int drbd_make_request_26(request_queue_t *q, struct bio *bio)
 					bio->bi_sector,bio);
 }
 #endif
-
-void drbd_thaw_frozen_reqs(drbd_dev *mdev)
-{
-	drbd_bio_t *bio;
-	int reqs=0;
-
-	spin_lock_irq(&mdev->req_lock);
-	bio = mdev->first_frozen_bio;
-	mdev->first_frozen_bio = NULL;
-	mdev->last_frozen_bio = NULL;
-	spin_unlock_irq(&mdev->req_lock);
-	
-	while(bio) {
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-		drbd_make_request_common(mdev,WRITE,bio->b_size,bio->b_rsector,bio);
-#else
-		drbd_make_request_common(mdev,WRITE,bio->bi_size,bio->bi_sector,bio);
-#endif
-		bio = bio->REQ_NEXT;
-		reqs++;
-	}
-	WARN("Continued %d requests (which where issued after IO-freeze).\n",
-	     reqs);
-}
