@@ -5,9 +5,9 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 2003-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2003-2004, Lars Ellenberg <l.g.e@web.de>.
+	authors.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
 
  */
 
-#include <linux/autoconf.h>
+#include <linux/config.h>
 #include <linux/module.h>
 #include <linux/version.h>
 
@@ -139,7 +139,6 @@ void drbd_dio_end(struct buffer_head *bh, int uptodate)
 {
 	struct Drbd_Conf* mdev;
 	drbd_request_t *req;
-	sector_t rsector;
 
 	mdev = bh->b_private;
 	PARANOIA_BUG_ON(!IS_VALID_MDEV(mdev));
@@ -148,10 +147,8 @@ void drbd_dio_end(struct buffer_head *bh, int uptodate)
 	PARANOIA_BUG_ON(!VALID_POINTER(req));
 
 	drbd_chk_io_error(mdev,!uptodate);
-        // req may get freed within drbd_end_req
-	rsector = req->sector;
-	drbd_end_req(req, RQ_DRBD_LOCAL, uptodate);
-	drbd_al_complete_io(mdev,rsector);
+	drbd_end_req(req, RQ_DRBD_LOCAL, uptodate, drbd_req_get_sector(req));
+	drbd_al_complete_io(mdev,drbd_req_get_sector(req));
 	dec_local(mdev);
 }
 
@@ -278,23 +275,25 @@ int drbd_dio_end_sec(struct bio *bio, unsigned int bytes_done, int error)
  */
 int drbd_dio_end(struct bio *bio, unsigned int bytes_done, int error)
 {
-	drbd_request_t *req=bio->bi_private;
-	struct Drbd_Conf* mdev=req->mdev;
+	struct Drbd_Conf* mdev=bio->bi_private;
+	drbd_request_t *req;
 	sector_t rsector;
 
 	// see above
 	ERR_IF (bio->bi_size)
 		return 1;
 
-	if(error) DUMPI(error);
+	PARANOIA_BUG_ON(!IS_VALID_MDEV(mdev));
+
+	req = container_of(bio,struct drbd_request,private_bio);
+	PARANOIA_BUG_ON(!VALID_POINTER(req));
 
 	drbd_chk_io_error(mdev,error);
-        // req may get freed within drbd_end_req
-	rsector = req->sector;
-	drbd_end_req(req, RQ_DRBD_LOCAL, (error == 0));
+	rsector = drbd_req_get_sector(req);
+        // the bi_sector of the bio gets modified somewhere in drbd_end_req()!
+	drbd_end_req(req, RQ_DRBD_LOCAL, (error == 0), rsector);
 	drbd_al_complete_io(mdev,rsector);
 	dec_local(mdev);
-	bio_put(bio);
 	return 0;
 }
 
@@ -302,13 +301,18 @@ int drbd_dio_end(struct bio *bio, unsigned int bytes_done, int error)
  */
 int drbd_read_bi_end_io(struct bio *bio, unsigned int bytes_done, int error)
 {
-	drbd_request_t *req=bio->bi_private;
-	struct Drbd_Conf* mdev=req->mdev;
+	struct Drbd_Conf* mdev = bio->bi_private;
+	drbd_request_t *req;
 
 	// see above
 	ERR_IF (bio->bi_size)
 		return 1;
 
+	PARANOIA_BUG_ON(!IS_VALID_MDEV(mdev));
+
+	req = container_of(bio,struct drbd_request,private_bio);
+	PARANOIA_BUG_ON(!VALID_POINTER(req));
+
 	/* READAs may fail.
 	 * upper layers need to be able to handle that themselves */
 	if (bio_rw(bio) == READA) goto pass_on;
@@ -330,7 +334,6 @@ int drbd_read_bi_end_io(struct bio *bio, unsigned int bytes_done, int error)
 		mempool_free(req,drbd_request_mempool);
 	}
 
-	bio_put(bio);
 	dec_local(mdev);
 	return 0;
 }
@@ -379,10 +382,6 @@ int w_read_retry_remote(drbd_dev* mdev, struct drbd_work* w,int cancel)
 
 	// FIXME: what if partner was SyncTarget, and is out of sync for
 	// this area ?? ... should be handled in the receiver.
-
-	ok = drbd_io_error(mdev);
-	if(unlikely(!ok)) ERR("Sending in w_read_retry_remote() failed\n");
-	
 	inc_ap_pending(mdev);
 	ok = drbd_read_remote(mdev,req);
 	if(unlikely(!ok)) {
@@ -695,10 +694,9 @@ STATIC void _drbd_rs_resume(drbd_dev *mdev)
 	_set_cstate(mdev,ns);
 
 	if(mdev->cstate == SyncTarget) {
-		int stop_syncer_timer_was_set =
-			test_and_clear_bit(STOP_SYNC_TIMER,&mdev->flags);
-		ERR_IF(stop_syncer_timer_was_set) {
+		ERR_IF(test_bit(STOP_SYNC_TIMER,&mdev->flags)) {
 			unsigned long rs_left = drbd_bm_total_weight(mdev);
+			clear_bit(STOP_SYNC_TIMER,&mdev->flags);
 			if (rs_left == 0) {
 				INFO("rs_left==0 in _drbd_rs_resume\n");
 			} else {
@@ -784,8 +782,6 @@ int w_resume_next_sg(drbd_dev* mdev, struct drbd_work* w, int unused)
 {
 	drbd_dev *odev;
 	int i,ng=10000;
-	/* FIXME need to document that 10000 somewhere!
-	 * maybe better: first iterate to find the max sync group number */
 
 	PARANOIA_BUG_ON(w != &mdev->resync_work);
 
@@ -793,9 +789,6 @@ int w_resume_next_sg(drbd_dev* mdev, struct drbd_work* w, int unused)
 
 	for (i=0; i < minor_count; i++) {
 		odev = drbd_conf + i;
-		/* currently we should never reach this while being
-		 * mdev->cstate == Sync*, but just incase: skip ourselves */
-		if (odev == mdev) continue;
 		if ( odev->sync_conf.group <= mdev->sync_conf.group
 		     && ( odev->cstate == SyncSource || 
 			  odev->cstate == SyncTarget ) ) {
@@ -806,7 +799,6 @@ int w_resume_next_sg(drbd_dev* mdev, struct drbd_work* w, int unused)
 
 	for (i=0; i < minor_count; i++) { // find next sync group
 		odev = drbd_conf + i;
-		if (odev == mdev) continue;
 		if ( odev->sync_conf.group > mdev->sync_conf.group
 		     && odev->sync_conf.group < ng && 
 		     (odev->cstate==PausedSyncS || odev->cstate==PausedSyncT)){
@@ -816,7 +808,6 @@ int w_resume_next_sg(drbd_dev* mdev, struct drbd_work* w, int unused)
 
 	for (i=0; i < minor_count; i++) { // resume all devices in next group
 		odev = drbd_conf + i;
-		if (odev == mdev) continue;
 		if ( odev->sync_conf.group == ng &&
 		     (odev->cstate==PausedSyncS || odev->cstate==PausedSyncT)){
 			_drbd_rs_resume(odev);
@@ -824,8 +815,8 @@ int w_resume_next_sg(drbd_dev* mdev, struct drbd_work* w, int unused)
 	}
 
  out:
-	w->cb = w_resync_inactive;
 	drbd_global_unlock();
+	w->cb = w_resync_inactive;
 
 	return 1;
 }
@@ -897,6 +888,7 @@ void drbd_start_resync(drbd_dev *mdev, Drbd_CState side)
 		else
 			ERR("resync_work.cb == %p ???, should be w_resync_inactive\n",
 					mdev->resync_work.cb);
+		return;
 	}
 
 	if ( mdev->rs_total == 0 ) {
@@ -917,16 +909,12 @@ void drbd_start_resync(drbd_dev *mdev, Drbd_CState side)
 	   */
 	drbd_global_unlock();
 
-	/* THINK: for consistency, should this be before the drbd_global_unlock() ? */
-	{
-		int stop_syncer_timer_was_set =
-			test_and_clear_bit(STOP_SYNC_TIMER,&mdev->flags);
-		if (mdev->cstate == SyncTarget) {
-			D_ASSERT(!stop_syncer_timer_was_set);
-			mod_timer(&mdev->resync_timer,jiffies);
-		} else if (mdev->cstate == PausedSyncT) {
-			D_ASSERT(stop_syncer_timer_was_set);
-		}
+	if (mdev->cstate == SyncTarget) {
+		D_ASSERT(!test_bit(STOP_SYNC_TIMER,&mdev->flags));
+		mod_timer(&mdev->resync_timer,jiffies);
+	} else if (mdev->cstate == PausedSyncT) { 
+		D_ASSERT(test_bit(STOP_SYNC_TIMER,&mdev->flags));
+		clear_bit(STOP_SYNC_TIMER,&mdev->flags);
 	}
 }
 
@@ -939,7 +927,7 @@ int drbd_worker(struct Drbd_thread *thi)
 
 	sprintf(current->comm, "drbd%d_worker", (int)(mdev-drbd_conf));
 
-	while (get_t_state(thi) == Running) {
+	for (;;) {
 		intr = down_interruptible(&mdev->data.work.s);
 
 		if (unlikely(drbd_did_panic == DRBD_MAGIC)) {
@@ -961,21 +949,7 @@ int drbd_worker(struct Drbd_thread *thi)
 
 		w = 0;
 		spin_lock_irq(&mdev->req_lock);
-		ERR_IF(list_empty(&mdev->data.work.q)) {
-			/* something terribly wrong in our logic.
-			 * we were able to down() the semaphore,
-			 * but the list is empty... doh.
-			 *
-			 * what is the best thing to do now?
-			 * try again from scratch, restarting the receiver,
-			 * asender, whatnot? could break even more ugly,
-			 * e.g. when we are primary, but no good local data.
-			 *
-			 * I'll try to get away just starting over this loop.
-			 */
-			spin_unlock_irq(&mdev->req_lock);
-			continue;
-		}
+		D_ASSERT(!list_empty(&mdev->data.work.q));
 		w = list_entry(mdev->data.work.q.next,struct drbd_work,list);
 		list_del_init(&w->list);
 		spin_unlock_irq(&mdev->req_lock);
@@ -990,49 +964,8 @@ int drbd_worker(struct Drbd_thread *thi)
 
 	drbd_wait_ee(mdev,&mdev->read_ee);
 
-	/* When we terminate a resync process, either because it finished
-	 * sucessfully, or because (like in this case here) we lost
-	 * communications, we need to "w_resume_next_sg".
-	 * We cannot use del_timer_sync from within _set_cstate, and since the
-	 * resync timer may still be scheduled and would then trigger anyways,
-	 * we set the STOP_SYNC_TIMER bit, and schedule the timer for immediate
-	 * execution from within _set_cstate().
-	 * The timer should then clear that bit and queue w_resume_next_sg.
-	 *
-	 * This is fine for the normal "resync finished" case.
-	 *
-	 * In this case (worker thread beeing stopped), there is a race:
-	 * we cannot be sure that the timer already triggered.
-	 *
-	 * So we del_timer_sync here, and check that "STOP_SYNC_TIMER" bit.
-	 * if it is still set, we queue w_resume_next_sg anyways,
-	 * just to be sure.
-	 */
-
-	del_timer_sync(&mdev->resync_timer);
-	/* possible paranoia check: the STOP_SYNC_TIMER bit should be set
-	 * if and only if del_timer_sync returns true ... */
-
-	spin_lock_irq(&mdev->req_lock);
-	if (test_and_clear_bit(STOP_SYNC_TIMER,&mdev->flags)) {
-		mdev->resync_work.cb = w_resume_next_sg;
-		if (list_empty(&mdev->resync_work.list))
-			_drbd_queue_work(&mdev->data.work,&mdev->resync_work);
-		// else: already queued
-	} else {
-		/* timer already consumed that bit, or it was never set */
-		if (list_empty(&mdev->resync_work.list)) {
-			/* not queued, should be inactive */
-			ERR_IF (mdev->resync_work.cb != w_resync_inactive)
-				mdev->resync_work.cb = w_resync_inactive;
-		} else {
-			/* still queued; should be w_resume_next_sg */
-			ERR_IF (mdev->resync_work.cb != w_resume_next_sg)
-				mdev->resync_work.cb = w_resume_next_sg;
-		}
-	}
-
 	i = 0;
+	spin_lock_irq(&mdev->req_lock);
   again:
 	list_splice_init(&mdev->data.work.q,&work_list);
 	spin_unlock_irq(&mdev->req_lock);
@@ -1041,18 +974,13 @@ int drbd_worker(struct Drbd_thread *thi)
 		w = list_entry(work_list.next, struct drbd_work,list);
 		list_del_init(&w->list);
 		w->cb(mdev,w,1);
-		i++; /* dead debugging code */
+		i++;
 	}
 
 	spin_lock_irq(&mdev->req_lock);
 	ERR_IF(!list_empty(&mdev->data.work.q))
 		goto again;
 	sema_init(&mdev->data.work.s,0);
-	/* DANGEROUS race: if someone did queue his work within the spinlock,
-	 * but up() ed outside the spinlock, we could get an up() on the
-	 * semaphore without corresponding list entry.
-	 * So don't do that.
-	 */
 	spin_unlock_irq(&mdev->req_lock);
 
 	INFO("worker terminated\n");
