@@ -22,8 +22,8 @@
 
 extern const char * drbd_buildtag(void);
 
-#define REL_VERSION "0.7.25"
-#define API_VERSION 79
+#define REL_VERSION "0.7.7"
+#define API_VERSION 77
 #define PRO_VERSION 74
 
 //#define DBG_ALL_SYMBOLS // no static functs, improves quality of OOPS traces
@@ -65,15 +65,4 @@ extern const char * drbd_buildtag(void);
 // for troubles.
 // #define DRBD_DISABLE_SENDPAGE
 
-// older, unpached kernel sources do not have it, so we added it in drbd_receiver.c
-// but more recent kernels define it in arch/um/include/mem.h
-#define HAVE_UML_TO_VIRT
-
-#define NO_MORE_DEV_FS
-
-// 2.6.something has deprecated kmem_cache_t
-// some older still use it.
-// some have it defined as struct kmem_cache_s, some as struct kmem_cache
-//#define USE_KMEM_CACHE_S
-
 #endif
