@@ -22,20 +22,10 @@ test -e ./linux/drbd_config.h || {
 
 test -n "$KDIR"
 KDIR=${KDIR%/}
-if test -z "$O"; then
-	O=$KDIR;
-else
-	O=${O%/}
-fi
 
-# some paranoia: check that all files are where we expect them
-ls > /dev/null \
-$KDIR/{Makefile,include/linux/{sched,list,fs}.h}
-ls > /dev/null \
-$O/{.config,Makefile,include/linux/version.h}
-test -e $KDIR/include/asm/bitops.h ||
-test -e $O/include2/asm/bitops.h   ||
-exit 1
+ls >/dev/null \
+$KDIR/{.config,Makefile,include/{linux/{version,sched,list,fs},asm/bitops}.h}
+
 
 if grep_q "^PATCHLEVEL *= *4" $KDIR/Makefile ; then
   # do we have the threadding stuff in the kernel,
@@ -69,13 +59,7 @@ if grep_q "^PATCHLEVEL *= *4" $KDIR/Makefile ; then
     cat 2>/dev/null $KDIR/include/asm{,/arch}/bitops.h |
     grep_q 'find_next_bit'
   then
-    # on ppc64, it's declared but not exported, so we use our own copy
-    if grep_q '^CONFIG_PPC64=y' $KDIR/.config
-    then
-      have_find_next_bit=0
-    else
-      have_find_next_bit=1
-    fi
+    have_find_next_bit=1
   else
     have_find_next_bit=0
   fi
@@ -90,8 +74,6 @@ if grep_q "^PATCHLEVEL *= *4" $KDIR/Makefile ; then
   else
     have_mm_inline_h=0
   fi
-  no_more_dev_fs=0
-  have_kmem_cache_s=0
 else
     # 2.6. kernel. just leave it alone...
     need_sighand_hack=0
@@ -99,13 +81,6 @@ else
     need_RH_2_4_18_hack=0
     have_find_next_bit=0
     have_mm_inline_h=0
-    no_more_dev_fs=1
-    
-    if grep_q "typedef.*kmem_cache_s" $KDIR/include/linux/slab.h ; then
-      have_kmem_cache_s=1
-    else
-      have_kmem_cache_s=0
-    fi
 fi
 
 test -e ./linux/drbd_config.h.orig || cp ./linux/drbd_config.h{,.orig}
@@ -120,11 +95,7 @@ perl -pe "
  s{.*(#define HAVE_FIND_NEXT_BIT.*)}
   { ( $have_find_next_bit ? '' : '//' ) . \$1}e;
  s{.*(#define HAVE_MM_INLINE_H.*)}
-  { ( $have_mm_inline_h ? '' : '//' ) . \$1}e;
- s{.*(#define NO_MORE_DEV_FS.*)}
-  { ( $no_more_dev_fs ? '' : '//' ) . \$1}e;
- s{.*(#define USE_KMEM_CACHE_S.*)}
-  { ( $have_kmem_cache_s ? '' : '//' ) . \$1}e;" \
+  { ( $have_mm_inline_h ? '' : '//' ) . \$1}e;" \
 	  < ./linux/drbd_config.h \
 	  > ./linux/drbd_config.h.new
 
