@@ -3,9 +3,11 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 2002-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2002-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 2002-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+        Initial author.
+
+   Copyright (C) 2003-2004, Lars Ellenberg <l.g.e@web.de>
+        contributions.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,7 +38,6 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <sys/poll.h>
-#include <fcntl.h>
 #include <unistd.h>
 #include <errno.h>
 #include <getopt.h>
@@ -63,9 +64,8 @@ struct adm_cmd {
   const char* name;
   int (* function)(struct d_resource*,char* );
   char* arg;
-  int show_in_usage     :1;
-  int res_name_required :1;
-  int verify_ips        :1;
+  int show_in_usage;
+  int res_name_required;
 };
 
 struct deferred_cmd
@@ -76,7 +76,6 @@ struct deferred_cmd
 };
 
 extern int yyparse();
-extern int yydebug;
 extern FILE* yyin;
 
 int adm_attach(struct d_resource* ,char* );
@@ -102,14 +101,13 @@ char ss_buffer[255];
 struct utsname nodeinfo;
 int line=1;
 int fline, c_resource_start;
-struct d_globals global_options = { 0, 0, 0, 1 };
+struct d_globals global_options = { 0, 0, 1 };
 char *config_file = NULL;
 struct d_resource* config = NULL;
 int nr_resources;
 int highest_minor;
 int config_valid=1;
 int dry_run;
-int do_verify_ips;
 char* drbdsetup;
 char* setup_opts[10];
 int soi=0;
@@ -170,35 +168,34 @@ struct option admopt[] = {
 };
 
 struct adm_cmd cmds[] = {
-/*   name, function, arg,                     show, needs res, verify_ips */
-  { "attach",            adm_attach,  0                  ,1,1,0 },
-  { "detach",            adm_generic_s,"detach"          ,1,1,0 },
-  { "connect",           adm_connect, 0                  ,1,1,1 },
-  { "disconnect",        adm_generic_s,"disconnect"      ,1,1,0 },
-  { "up",                adm_up,      0                  ,1,1,1 },
-  { "down",              adm_generic_s,"down"            ,1,1,0 },
-  { "primary",           adm_primary, 0                  ,1,1,0 },
-  { "secondary",         adm_generic_s,"secondary"       ,1,1,0 },
-  { "invalidate",        adm_generic_l,"invalidate"      ,1,1,0 },
-  { "invalidate_remote", adm_generic_l,"invalidate_remote",1,1,1 },
-  { "resize",            adm_resize,  0                  ,1,1,1 },
-  { "syncer",            adm_syncer,  0                  ,1,1,1 },
-  { "adjust",            adm_adjust,  0                  ,1,1,1 },
-  { "wait_connect",      adm_wait_c,  0                  ,1,1,1 },
-  { "state",             adm_generic_s,"state"           ,1,1,0 },
-  { "cstate",            adm_generic_s,"cstate"          ,1,1,1 },
-  { "dump",              adm_dump,    0                  ,1,1,1 },
-  { "wait_con_int",      adm_wait_ci, 0                  ,1,0,1 },
-  { "sh-resources",      sh_resources,0                  ,0,0,0 },
-  { "sh-mod-parms",      sh_mod_parms,0                  ,0,0,0 },
-  { "sh-dev",            sh_dev,      0                  ,0,1,0 },
-  { "sh-ll-dev",         sh_ll_dev,   0                  ,0,1,0 },
-  { "sh-md-dev",         sh_md_dev,   0                  ,0,1,0 },
-  { "sh-md-idx",         sh_md_idx,   0                  ,0,1,0 },
+  { "attach",            adm_attach,  0                  ,1,1 },
+  { "detach",            adm_generic_s,"detach"          ,1,1 },
+  { "connect",           adm_connect, 0                  ,1,1 },
+  { "disconnect",        adm_generic_s,"disconnect"      ,1,1 },
+  { "up",                adm_up,      0                  ,1,1 },
+  { "down",              adm_generic_s,"down"            ,1,1 },
+  { "primary",           adm_primary, 0                  ,1,1 },
+  { "secondary",         adm_generic_s,"secondary"       ,1,1 },
+  { "invalidate",        adm_generic_l,"invalidate"      ,1,1 },
+  { "invalidate_remote", adm_generic_l,"invalidate_remote",1,1 },
+  { "resize",            adm_resize,  0                  ,1,1 },
+  { "syncer",            adm_syncer,  0                  ,1,1 },
+  { "adjust",            adm_adjust,  0                  ,1,1 },
+  { "wait_connect",      adm_wait_c,  0                  ,1,1 },
+  { "state",             adm_generic_s,"state"           ,1,1 },
+  { "cstate",            adm_generic_s,"cstate"          ,1,1 },
+  { "dump",              adm_dump,    0                  ,1,1 },
+  { "wait_con_int",      adm_wait_ci, 0                  ,1,0 },
+  { "sh-resources",      sh_resources,0                  ,0,0 },
+  { "sh-mod-parms",      sh_mod_parms,0                  ,0,0 },
+  { "sh-dev",            sh_dev,      0                  ,0,1 },
+  { "sh-ll-dev",         sh_ll_dev,   0                  ,0,1 },
+  { "sh-md-dev",         sh_md_dev,   0                  ,0,1 },
+  { "sh-md-idx",         sh_md_idx,   0                  ,0,1 }
 };
 
 
-#define ARRY_SIZE(A) (int)(sizeof(A)/sizeof(A[0]))
+#define ARRY_SIZE(A) (sizeof(A)/sizeof(A[0]))
 
 /*** These functions are used to the print the config ***/
 
@@ -243,17 +240,17 @@ static void dump_options(char* name,struct d_option* opts)
 
 static void dump_global_info()
 {
-  if (  !global_options.minor_count
-     && !global_options.disable_ip_verification
-     &&  global_options.dialog_refresh == 1 ) return;
-  printI("global {\n"); ++indent;
-  if (global_options.disable_ip_verification)
-    printI("disable-ip-verification;\n");
-  if (global_options.minor_count)
-    printI("minor-count %i;\n", global_options.minor_count);
-  if (global_options.dialog_refresh != 1)
-    printI("dialog-refresh %i;\n", global_options.dialog_refresh);
-  --indent; printI("}\n\n");
+  if (global_options.minor_count || global_options.disable_io_hints)
+    {
+      printI("global {\n"); ++indent;
+      if (global_options.disable_io_hints)
+	printI("disable-io-hints;\n");
+      if (global_options.minor_count)
+	printI("minor-count %i;\n", global_options.minor_count);
+      if (global_options.dialog_refresh != 1)
+	printI("dialog-refresh %i;\n", global_options.dialog_refresh);
+      --indent; printI("}\n\n");
+    }
 }
 
 static void dump_host_info(struct d_host_info* hi)
@@ -294,11 +291,8 @@ static int adm_dump(struct d_resource* res,char* unused)
 static int sh_resources(struct d_resource* ignored,char* unused)
 {
   struct d_resource *res,*t;
-  int first=1;
-
   for_each_resource(res,t,config) {
-    printf(first?"%s":" %s",esc(res->name));
-    first=0;
+    printf(res==config?"%s":" %s",esc(res->name));
   }
   printf("\n");
 
@@ -395,21 +389,16 @@ static void find_drbdsetup(char** pathes)
   struct stat buf;
   char **path;
 
-  if (drbdsetup) {
-    free(drbdsetup);
-    drbdsetup = NULL;
-  }
-
   path=pathes;
   while(*path) {
     if(stat(*path,&buf)==0) {
-      drbdsetup = strdup(*path);
+      drbdsetup=*path;
       return;
     }
     path++;
   }
 
-  fprintf(stderr,"Can not find drbdsetup\n");
+  fprintf(stderr,"Can not find drbdsetup");
   exit(E_exec_error);
 }
 
@@ -445,7 +434,7 @@ pid_t m_system(char** argv,int flags)
     exit(E_exec_error);
   }
   if(pid == 0) {
-    execvp(argv[0],argv);
+    execv(argv[0],argv);
     fprintf(stderr,"Can not exec\n");
     exit(E_exec_error);
   }
@@ -454,13 +443,10 @@ pid_t m_system(char** argv,int flags)
     int timeout;
     sigaction(SIGALRM,&sa,&so);
     alarm_raised=0;
-    switch(flags & 0x0f) {
+    switch(flags) {
     case SLEEPS_SHORT:     timeout = 5; break;
     case SLEEPS_LONG:      timeout = 120; break;
     case SLEEPS_VERY_LONG: timeout = 600; break;
-    default:
-			   fprintf(stderr,"%s:%u: oops... flags=%02x\n", __FILE__ , __LINE__ , flags);
-			   exit(E_thinko);
     }
     alarm(timeout);
   }
@@ -490,7 +476,7 @@ pid_t m_system(char** argv,int flags)
   if( flags & SLEEPS_FINITE ) {
     alarm(0);
     sigaction(SIGALRM,&so,NULL);
-    if(rv && !(flags & DONT_REPORT_FAILED)) {
+    if(rv) {
       fprintf(stderr,"Command '");
       while(*argv) {
 	fprintf(stderr,"%s",*argv++);
@@ -552,16 +538,13 @@ int adm_resize(struct d_resource* res,char* unused)
 {
   char* argv[20];
   struct d_option* opt;
-  int i,argc=0;
+  int argc=0;
 
   argv[argc++]=drbdsetup;
   argv[argc++]=res->me->device;
   argv[argc++]="resize";
   opt=find_opt(res->disk_options,"size");
   if(opt) ssprintf(argv[argc++],"--%s=%s",opt->name,opt->value);
-  for(i=0;i<soi;i++) {
-    argv[argc++]=setup_opts[i];
-  }
   argv[argc++]=0;
 
   return m_system(argv,SLEEPS_SHORT);
@@ -840,26 +823,9 @@ static int adm_wait_ci(struct d_resource* ignored ,char* unused)
   struct d_option* opt;
   int rr,wtime,argc,i=0;
   time_t start;
-  int saved_stdin,saved_stdout,fd;
 
   struct sigaction so,sa;
 
-  saved_stdin = -1;
-  if( isatty(fileno(stdin)) == 0 || isatty(fileno(stdout)) == 0 ) {
-    fprintf(stderr,"WARN: stdin/stdout is not a TTY; using /dev/console");
-    fprintf(stdout,"WARN: stdin/stdout is not a TTY; using /dev/console");
-    saved_stdin  = dup(fileno(stdin));
-    if( saved_stdin == -1) perror("dup(stdin)");
-    saved_stdout = dup(fileno(stdout));
-    if( saved_stdin == -1) perror("dup(stdout)");
-    fd = open( "/dev/console", O_RDONLY);
-    if(fd == -1) perror("open('/dev/console, O_RDONLY)");
-    dup2(fd, fileno(stdin) );
-    fd = open( "/dev/console", O_WRONLY);
-    if(fd == -1) perror("open('/dev/console, O_WRONLY)");
-    dup2(fd, fileno(stdout) );
-  }
-
   sa.sa_handler=chld_sig_hand;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=SA_NOCLDSTOP;
@@ -879,21 +845,15 @@ static int adm_wait_ci(struct d_resource* ignored ,char* unused)
     pids[i++]=m_system(argv,RETURN_PID);
   }
 
-  wtime = global_options.dialog_refresh ?
+  wtime = global_options.dialog_refresh ? 
     global_options.dialog_refresh : -1;
 
   start = time(0);
-  for (i = 0; i < 10; i++) {
-    // no string, but timeout
-    rr = gets_timeout(pids,0,0,1*1000);
-    if (rr < 0) break;
-    putchar('.');
-    fflush(stdout);
-    check_exit_codes(pids);
-  }
+  rr = gets_timeout(pids,0,0,3*1000); // no string, but timeout of 3 seconds.
+  check_exit_codes(pids);
 
   if(rr == 0) {
-    printf("\n***************************************************************\n"
+    printf("***************************************************************\n"
 	   " DRBD's startup script waits for the peer node(s) to appear.\n"
 	   " - In case this node was already a degraded cluster before the\n"
 	   "   reboot the timeout is %s seconds. [degr-wfc-timeout]\n"
@@ -921,15 +881,11 @@ static int adm_wait_ci(struct d_resource* ignored ,char* unused)
 	  printf(" To abort waiting enter 'yes' [ -- ]:");
 	}
       }
+      
     } while( rr != -1 );
     printf("\n");
   }
 
-  if( saved_stdin != -1 ) {
-    dup2(saved_stdin,  fileno(stdin ) );
-    dup2(saved_stdout, fileno(stdout) );
-  }
-
   return 0;
 }
 
@@ -952,7 +908,7 @@ const char* make_optstring(struct option *options)
   return buffer;
 }
 
-void print_usage_and_exit(const char* addinfo)
+void print_usage()
 {
   int i;
   struct option *opt;
@@ -984,9 +940,6 @@ void print_usage_and_exit(const char* addinfo)
   printf("\nVersion: "REL_VERSION" (api:%d)\n%s\n",
 		  API_VERSION, drbd_buildtag());
 
-  if (addinfo)
-      printf("\n%s\n",addinfo);
-
   exit(E_usage);
 }
 
@@ -999,8 +952,7 @@ void verify_ips(struct d_resource* res)
   char *argv[] = { "/bin/bash", "-c", NULL, "drbdadm:verify_ips", NULL };
   int ex;
 
-  if (global_options.disable_ip_verification) return;
-  if (dry_run == 1 || do_verify_ips == 0) return;
+  if (dry_run == 1) return;
 
   if (!(res && res->me   && res->me->address
 	    && res->peer && res->peer->address)) {
@@ -1022,7 +974,7 @@ void verify_ips(struct d_resource* res)
 	"fi >/dev/null",
 	my_ip);
   if (ex < 0) { perror("asprintf"); exit(E_thinko); }
-  ex = m_system(argv,SLEEPS_SHORT|DONT_REPORT_FAILED);
+  ex = m_system(argv,SLEEPS_SHORT);
   free(argv[2]); argv[2] = NULL;
 
   if (ex != 0) {
@@ -1166,45 +1118,17 @@ int main(int argc, char** argv)
   int i,rv;
   struct adm_cmd *cmd;
   struct d_resource *res,*tmp;
-  char *env_drbd_nodename = NULL;
 
+  drbdsetup=NULL;
   dry_run=0;
   yyin=NULL;
   uname(&nodeinfo); /* FIXME maybe fold to lower case ? */
 
-  env_drbd_nodename = getenv("__DRBD_NODE__");
-  if (env_drbd_nodename && *env_drbd_nodename) {
-    strncpy(nodeinfo.nodename,env_drbd_nodename,sizeof(nodeinfo.nodename)-1);
-    nodeinfo.nodename[sizeof(nodeinfo.nodename)-1] = 0;
-    fprintf(stderr, "\n"
-            "   found __DRBD_NODE__ in environment\n"
-            "   PRETENDING that I am >>%s<<\n\n",nodeinfo.nodename);
-  }
-
-  /* in case drbdadm is called with an absolut or relative pathname
-   * look for the drbdsetup binary in the same location,
-   * otherwise, just let execvp sort it out... */
-  if( (progname=strrchr(argv[0],'/')) == 0 ) {
+  if( (progname=strrchr(argv[0],'/')) )
+    argv[0] = ++progname;
+  else
     progname=argv[0];
-    drbdsetup = strdup("drbdsetup");
-  } else {
-    size_t len = strlen(argv[0]) + 1;
-    ++progname;
-    len += strlen("drbdsetup") - strlen(progname);
-    drbdsetup = malloc(len);
-    if (drbdsetup) {
-      strncpy(drbdsetup, argv[0], (progname - argv[0]));
-      strcpy(drbdsetup + (progname - argv[0]), "drbdsetup");
-    }
-    argv[0] = progname;
-  }
-
-  if(argc == 1) print_usage_and_exit("missing arguments"); // arguments missing.
-
-  if (drbdsetup == NULL) {
-    fprintf(stderr,"could not strdup argv[0].\n");
-    exit(E_exec_error);
-  }
+  if(argc == 1) print_usage(); // arguments missing.
 
   opterr=1;
   optind=0;
@@ -1243,19 +1167,18 @@ int main(int argc, char** argv)
 	case '?':
 	  // commented out, since opterr=1
 	  //fprintf(stderr,"Unknown option %s\n",argv[optind-1]);
-	  fprintf(stderr,"try '%s help'\n",progname);
 	  return 20;
 	  break;
 	}
     }
 
-  if ( optind == argc ) print_usage_and_exit(0);
+  if ( optind == argc ) print_usage();
 
   while(argv[optind][0]=='-') {
     setup_opts[soi++]=argv[optind++];
-    if (optind == argc) print_usage_and_exit(0);
+    if (optind == argc) print_usage();
   }
-  if (optind == argc) print_usage_and_exit(0);
+  if (optind == argc) print_usage();
 
   cmd=NULL;
   for(i=0;i<ARRY_SIZE(cmds);i++) {
@@ -1266,12 +1189,10 @@ int main(int argc, char** argv)
   }
 
   if(cmd==NULL) {
-    if (!strncmp("help",argv[optind],5)) print_usage_and_exit(0);
     fprintf(stderr,"Unknown command '%s'.\n",argv[optind]);
     exit(E_usage);
   }
   optind++;
-  do_verify_ips = cmd->verify_ips;
 
   if (!config_file) {
     i=0;
@@ -1311,17 +1232,11 @@ int main(int argc, char** argv)
     exit(E_exec_error);
   };
 
-  //yydebug = 1;
   yyparse();
 
   if(!config_valid) exit(E_config_invalid);
 
-  if (config == NULL) {
-    fprintf(stderr, "no resources defined!\n");
-    exit(0); /* THINK exit here? what code? */
-  }
-
-  { /* block for mc to avoid compiler warnings */
+  {
     int mc=global_options.minor_count;
 
     highest_minor=0;
@@ -1341,14 +1256,17 @@ int main(int argc, char** argv)
     }
   }
 
+  if(drbdsetup == NULL) {
+    find_drbdsetup((char *[]){"./drbdsetup", "/sbin/drbdsetup", 0 });
+  }
+
   if(cmd->res_name_required)
     {
-      int is_dump = (cmd->function == adm_dump);
-      if (optind + 1 > argc && !is_dump)
-        print_usage_and_exit("missing arguments"); // arguments missing.
+      if (optind + 1 > argc && cmd->function != adm_dump)
+        print_usage (argv[0]);	// arguments missing.
 
-      if ( optind==argc || !strcmp(argv[optind],"all") ) {
-        if (is_dump) dump_global_info();
+      if(optind==argc || !strcmp(argv[optind],"all")) {
+        if (cmd->function == adm_dump) dump_global_info();
         for_each_resource(res,tmp,config) {
 	  if( (rv=cmd->function(res,cmd->arg)) ) {
 	    fprintf(stderr,"drbdsetup exited with code %d\n",rv);
