@@ -3,13 +3,15 @@
 
    This file is part of drbd by Philipp Reisner.
 
-   Copyright (C) 1999-2006, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2002-2006, Lars Ellenberg <lars.ellenberg@linbit.com>.
-   Copyright (C) 2001-2006, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
+        Initial author.
 
    Copyright (C) 2000, Fbio Oliv Leite <olive@conectiva.com.br>.
         Added sanity checks before using the device.
 
+   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>
+	main contributor.
+
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
@@ -48,7 +50,7 @@
 
 #include "drbd_limits.h"
 
-#define ARRY_SIZE(A) (int)(sizeof(A)/sizeof(A[0]))
+#define ARRY_SIZE(A) (sizeof(A)/sizeof(A[0]))
 
 /* Default values */
 #define DEF_NET_TIMEOUT             60      //  6 seconds
@@ -68,7 +70,6 @@
 #define DEF_ON_IO_ERROR         PassOn
 #define DEF_KO_COUNT                 0
 #define DEF_ON_DISCONNECT       Reconnect
-#define DEF_UNPLUG_WATERMARK       (DEF_MAX_BUFFERS/16)
 
 #if 0
 # define ioctl(X...) (fprintf(stderr,"ioctl(%s)\n",#X),0);
@@ -156,7 +157,6 @@ struct drbd_cmd commands[] = {
      { "timeout",    required_argument, 0, 't' },
      { "max-epoch-size", required_argument, 0, 'e' },
      { "max-buffers",required_argument, 0, 'b' },
-     { "unplug-watermark",required_argument, 0, 'u' },
      { "connect-int",required_argument, 0, 'c' },
      { "ping-int",   required_argument, 0, 'i' },
      { "sndbuf-size",required_argument, 0, 'S' },
@@ -188,7 +188,7 @@ const char *eh_names[] = {
 const char *dh_names[] = {
   [Reconnect]   = "reconnect",
   [DropNetConf] = "stand_alone",
-  [FreezeIO]    = "freeze_io"
+  // [FreezeIO]    = "freeze_io" // TODO on the kernel side...
 };
 
 unsigned long resolv(const char* name)
@@ -526,7 +526,7 @@ int scan_disk_options(char **argv,
       switch(c)
 	{
 	case 'd':
-	  cn->config.disk_size = m_strtoll_range(optarg,'K', "size",
+	  cn->config.disk_size = m_strtoll_range(optarg,'K', "disk-size",
 			      DRBD_DISK_SIZE_SECT_MIN>>1,
 			      DRBD_DISK_SIZE_SECT_MAX>>1 );
 	  break;
@@ -562,7 +562,6 @@ int scan_net_options(char **argv,
   cn->config.sndbuf_size = DEF_SNDBUF_SIZE ;
   cn->config.on_disconnect = DEF_ON_DISCONNECT;
   cn->config.ko_count = DEF_KO_COUNT;
-  cn->config.unplug_watermark = DEF_UNPLUG_WATERMARK;
 
   if(argc==0) return 0;
 
@@ -589,10 +588,6 @@ int scan_net_options(char **argv,
 	  cn->config.max_buffers = m_strtoll_range(optarg,1, "max-buffers",
 			  DRBD_MAX_BUFFERS_MIN, DRBD_MAX_BUFFERS_MAX);
 	  break;
-	case 'u':
-	  cn->config.unplug_watermark = m_strtoll_range(optarg,1, "unplug-watermark",
-			  DRBD_UNPLUG_WATERMARK_MIN, DRBD_UNPLUG_WATERMARK_MAX);
-	  break;
 	case 'c':
 	  cn->config.try_connect_int = m_strtoll_range(optarg,1, "connect-int",
 			  DRBD_CONNECT_INT_MIN, DRBD_CONNECT_INT_MAX);
@@ -656,7 +651,6 @@ void print_config_ioctl_err(int err_no)
     [LDNoConfig]="You have to use the disk command first.",
     [LDMounted]="Lower device is already mounted.",
     [MDMounted]="Meta device is already mounted.",
-    [MDIOError]="IO error on meta device",
     [LDMDInvalid]="Lower device / meta device / index combination invalid.",
     [LDDeviceTooLarge]="Currently we only support devices up to 3.998TB.\n"
                        "(up to 2TB in case you do not have CONFIG_LBD set)",
@@ -1185,7 +1179,7 @@ int cmd_disk_size(int drbd_fd,char** argv,int argc,struct option *options)
 	  switch(c)
 	    {
 	    case 'd':
-	      u_size=m_strtoll_range(optarg,'K', "size",
+	      u_size=m_strtoll_range(optarg,'K', "disk-size",
 			      DRBD_DISK_SIZE_SECT_MIN>>1,
 			      DRBD_DISK_SIZE_SECT_MAX>>1 );
 	      break;
@@ -1197,16 +1191,11 @@ int cmd_disk_size(int drbd_fd,char** argv,int argc,struct option *options)
 	}
     }
 
+  fprintf(stderr,"err=ioctl(drbd_fd,DRBD_IOCTL_SET_DISK_SIZE,%lu);\n",u_size);
   err=ioctl(drbd_fd,DRBD_IOCTL_SET_DISK_SIZE,u_size);
   if(err)
     {
       PERROR("ioctl(,SET_DISK_SIZE,) failed");
-      if(err==EBUSY) {
-	fprintf(stderr,"Online resizing is not allowed during resync.");
-      }
-      if(err==EINPROGRESS) {
-	fprintf(stderr,"One node must be primary to do online resizing.");
-      }
       return 20;
     }
 
@@ -1220,17 +1209,13 @@ const char* guess_dev_name(const char* dir,int major,int minor)
   struct stat sb;
   static char dev_name[50];
 
-  if(major==0 && minor==0)
-    return NULL;
-
   device_dir=opendir(dir);
 
   if(!device_dir) goto err_out;
 
   while((dde=readdir(device_dir)))
     {
-      if(snprintf(dev_name,50,"%s/%s",dir,dde->d_name)>=50)
-	continue;
+      snprintf(dev_name,50,"%s/%s",dir,dde->d_name);
       if(stat(dev_name,&sb)) continue;
 
       if(S_ISBLK(sb.st_mode))
@@ -1248,8 +1233,7 @@ const char* guess_dev_name(const char* dir,int major,int minor)
 
   while((dde=readdir(device_dir)))
     {
-      if(snprintf(dev_name,50,"%s/%s",dir,dde->d_name)>=50)
-	continue;
+      snprintf(dev_name,50,"%s/%s",dir,dde->d_name);
       if(stat(dev_name,&sb)) continue;
 
       if(!strcmp(dde->d_name,".")) continue;
@@ -1262,8 +1246,11 @@ const char* guess_dev_name(const char* dir,int major,int minor)
 	{
 	  char subdir[50];
 
-	  if(snprintf(subdir,50,"%s/%s",dir,dde->d_name)>=50)
-	    continue;
+	  if(snprintf(subdir,50,"%s/%s",dir,dde->d_name)==49)
+	    { /* recursion is too deep */
+	      strcpy(dev_name,"can not guess name");
+	      return dev_name;
+	    }
 
 	  if(guess_dev_name(subdir,major,minor)) return dev_name;
 	}
@@ -1352,7 +1339,6 @@ int cmd_show(int drbd_fd,char** argv,int argc,struct option *options)
   SHOW_I("ping-int","sec", cn.nconf.ping_int, DEF_NET_PING_I);
   SHOW_I("max-epoch-size","", cn.nconf.max_epoch_size, DEF_MAX_EPOCH_SIZE);
   SHOW_I("max-buffers","", cn.nconf.max_buffers, DEF_MAX_BUFFERS);
-  SHOW_I("unplug-watermark","", cn.nconf.unplug_watermark, DEF_UNPLUG_WATERMARK);
   SHOW_I("sndbuf-size","", cn.nconf.sndbuf_size, DEF_SNDBUF_SIZE);
   SHOW_I("ko-count","", cn.nconf.ko_count, DEF_KO_COUNT);
   if( cn.nconf.on_disconnect != DEF_ON_DISCONNECT) {
